Option Strict On
Option Explicit On 

Public Class Designer
    Inherits System.Windows.Forms.Control

#Region " Windows Form Designer generated code "

    Public Sub New()
        MyBase.New()

        'This call is required by the Windows Form Designer.
        InitializeComponent()

        'Add any initialization after the InitializeComponent() call

        setstyle(ControlStyles.DoubleBuffer, True)
    End Sub

    'UserControl overrides dispose to clean up the component list.
    Protected Overloads Overrides Sub Dispose(ByVal disposing As Boolean)
        If disposing Then
            If Not (components Is Nothing) Then
                components.Dispose()
            End If
        End If
        MyBase.Dispose(disposing)
    End Sub

    'Required by the Windows Form Designer
    Private components As System.ComponentModel.IContainer

    'NOTE: The following procedure is required by the Windows Form Designer
    'It can be modified using the Windows Form Designer.  
    'Do not modify it using the code editor.
    <System.Diagnostics.DebuggerStepThrough()> Private Sub InitializeComponent()
        '
        'Designer
        '
        Me.AllowDrop = True
        Me.BackColor = System.Drawing.SystemColors.AppWorkspace

    End Sub

#End Region

    Dim canc As Boolean
    Dim cells As New ArrayList
    Dim draws As New ArrayList
    Dim hsb As HScrollBar, vsb As VScrollBar
    Dim insertingDataField As String
    Dim insertingMemberType As Type
    Dim lMouse As Boolean
    Dim moving, rubbing, horizing, verting, sizing, inserting As Boolean
    Dim offset As Point
    Dim pos As Avand.Designer.MousePos
    Dim selGrab As Avand.Designer.Grab
    Dim selMember As Model10.Member
    Dim selSection As Model10.Section
    Dim totalHeight As Integer
    Dim oX, oY As Integer
    Dim dX, dY As Integer

    Private mSelectedMember As Model10.Member
    Private mSelectedSection As Model10.Section
    Private mTable As Model10.Table
    Private mUser As Model10.User

    Private Const FRAME_CORRECTION As Byte = 0
    Private Const GRAB_SIZE As Byte = 6
    Private Const HEADER_HEIGHT As Byte = 18
    Private Const MARGIN_SIZE As Byte = 16
    Private Const MIN_SIZE As Byte = 1
    Private Const SIDE_SIZE As Byte = 4

    Public Event InsertingFinished As EventHandler
    Public Event RequestDelete As EventHandler
    Public Event ValueChanged As EventHandler

    Public Event TableDoubleClick As EventHandler
    Public Event SectionDoubleClick As EventHandler
    Public Event MembersDoubleClick As EventHandler

    Public Event TableSelected As EventHandler
    Public Event SectionSelected As EventHandler
    Public Event MembersSelected As EventHandler

    Public Event SelectedTableChanged As EventHandler
    Public Event SelectedSectionChanged As EventHandler
    Public Event SelectedMembersChanged As EventHandler

    Private Enum GrabStatus As Byte
        TopLeft = 0
        Top = 1
        TopRight = 2
        Right = 3
        BottomRight = 4
        Bottom = 5
        BottomLeft = 6
        Left = 7
    End Enum
    Private Enum MousePos As Byte
        None = 0
        Table = 1
        Header = 2
        Section = 3
        Member = 4
        Grab = 5
        ResizeTableWidth = 6
        ResizeSectionHeight = 7
    End Enum

    Private Class Grab
        Public Cursor As Cursor
        Public Status As Avand.Designer.GrabStatus
        Public Rectangle As Rectangle
    End Class
    Private Class Cell
        Public Member As Model10.Member
        Public Grabs As ArrayList
    End Class

    Private Function CheckResize(ByVal s As Model10.Section, ByVal p As Point, ByRef y As Integer) As Boolean
        Dim foo As Integer

        foo = y + s.Height + HEADER_HEIGHT
        If foo - SIDE_SIZE <= p.Y And p.Y < foo + SIDE_SIZE Then
            Cursor = Cursors.SizeNS
            pos = MousePos.ResizeSectionHeight
            selSection = s
            selMember = Nothing
            selGrab = Nothing
            Return True
        End If

        y += s.Height + HEADER_HEIGHT
    End Function
    Private Function CheckSection(ByVal s As Model10.Section, ByVal p As Point, ByRef y As Integer) As Boolean
        If y <= p.Y And p.Y < y + HEADER_HEIGHT Then
            If Not Cursor Is Cursors.Cross Then Cursor = Cursors.Default
            pos = MousePos.Header
            selSection = s
            selMember = Nothing
            selGrab = Nothing
            Return True
        End If

        y += HEADER_HEIGHT

        Dim i As Integer
        Dim r As New Rectangle(offset.X, y, mTable.Width, s.Height)
        If r.Contains(p) Then
            If Not Cursor Is Cursors.Cross Then
                For i = cells.Count - 1 To 0 Step -1
                    Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
                    If cel.Member.Section Is s Then
                        Dim ic As New Rectangle(cel.Member.Location, cel.Member.Size)
                        ic.Offset(offset.X, y)
                        Dim oc As Rectangle = ic
                        oc.Inflate(GRAB_SIZE, GRAB_SIZE)
                        If oc.Contains(p) And Not ic.Contains(p) Then
                            Dim grb As Avand.Designer.Grab
                            For Each grb In cel.Grabs
                                Dim rc As Rectangle = grb.Rectangle
                                rc.Offset(offset.X, y)
                                If rc.Contains(p) Then
                                    Cursor = grb.Cursor
                                    pos = MousePos.Grab
                                    selSection = s
                                    selMember = cel.Member
                                    selGrab = grb
                                    Return True
                                End If
                                rc = Nothing
                            Next : grb = Nothing

                            Cursor = Cursors.SizeAll
                            pos = MousePos.Member
                            selSection = s
                            selMember = cel.Member
                            selGrab = Nothing
                            Return True
                        End If
                        oc = Nothing
                        ic = Nothing
                    End If
                    cel = Nothing
                Next

                For i = s.Members.Count - 1 To 0 Step -1
                    Dim mem As Model10.Member = CType(s.Members(i), Model10.Member)
                    Dim rc As New Rectangle(mem.Location, mem.Size)
                    rc.Offset(offset.X, y)
                    If rc.Contains(p) Then
                        Cursor = Cursors.SizeAll
                        pos = MousePos.Member
                        selSection = s
                        selMember = mem
                        selGrab = Nothing
                        Return True
                    End If
                    rc = Nothing
                    mem = Nothing
                Next
            End If

            If Not Cursor Is Cursors.Cross Then Cursor = Cursors.Default
            pos = MousePos.Section
            selSection = s
            selMember = Nothing
            selGrab = Nothing
            Return True
        End If
        r = Nothing

        y += s.Height
    End Function
    Private Function GetMember() As Model10.Member
        If insertingDataField = "" Then
            Select Case insertingMemberType.ToString
                Case "Model10.Line"
                    GetMember = New Model10.Line
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Line"))
                Case "Model10.Frame"
                    GetMember = New Model10.Frame
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Frame"))
                Case "Model10.Label"
                    GetMember = New Model10.Label
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Label"))
                Case "Model10.TextBox"
                    GetMember = New Model10.TextBox
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.TextBox"))
                Case "Model10.PictureBox"
                    GetMember = New Model10.PictureBox
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.PictureBox"))
                Case "Model10.CheckBox"
                    GetMember = New Model10.CheckBox
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.CheckBox"))
                Case "Model10.Select"
                    GetMember = New Model10.Select
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Select"))
                Case "Model10.Edit"
                    GetMember = New Model10.Edit
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Edit"))
                Case "Model10.Update"
                    GetMember = New Model10.Update
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Update"))
                Case "Model10.Cancel"
                    GetMember = New Model10.Cancel
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Cancel"))
                Case "Model10.Delete"
                    GetMember = New Model10.Delete
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Delete"))
                Case "Model10.Browse"
                    GetMember = New Model10.Browse
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Browse"))
                Case "Model10.Computer"
                    GetMember = New Model10.Computer
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Computer"))
                Case "Model10.Expand"
                    GetMember = New Model10.Expand
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Expand"))
                Case "Model10.Next"
                    GetMember = New Model10.Next
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Next"))
                Case "Model10.MultiSelect"
                    GetMember = New Model10.MultiSelect
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.MultiSelect"))
                Case "Model10.Drilldown"
                    GetMember = New Model10.Drilldown
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Drilldown"))
                Case "Model10.Lookup"
                    GetMember = New Model10.Lookup
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Lookup"))
                Case "Model10.Sub"
                    GetMember = New Model10.Sub
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Sub"))
                Case "Model10.Chart"
                    GetMember = New Model10.Chart
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Chart"))
                Case Else : Exit Function
            End Select
        Else
            Select Case insertingMemberType.ToString
                Case "Model10.Label"
                    Dim lbl As New Model10.Label
                    lbl.DataField = insertingDataField
                    GetMember = lbl : lbl = Nothing
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Label"))
                Case "Model10.TextBox"
                    Dim txt As New Model10.TextBox
                    txt.DataField = insertingDataField
                    GetMember = txt : txt = Nothing
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.TextBox"))
                Case "Model10.PictureBox"
                    Dim pic As New Model10.PictureBox
                    pic.DataField = insertingDataField
                    GetMember = pic : pic = Nothing
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.PictureBox"))
                Case "Model10.CheckBox"
                    Dim chk As New Model10.CheckBox
                    chk.DataField = insertingDataField
                    GetMember = chk : chk = Nothing
                    GetMember.Name = mSelectedSection.GetNewMemberName([Global].GetSystemResource("Member.CheckBox"))
                Case Else : Exit Function
            End Select
        End If
        If Not GetMember Is Nothing Then
            If TypeOf GetMember Is Model10.Label Then
                Dim lbl As Model10.Label = CType(GetMember, Model10.Label)
                lbl.Text = GetMember.Name
                Select Case mTable.Presentation.RightToLeft
                    Case RightToLeft.No : lbl.Alignment = HorizontalAlignment.Left
                    Case RightToLeft.Yes : lbl.Alignment = HorizontalAlignment.Right
                    Case RightToLeft.Inherit
                        Select Case Me.RightToLeft
                            Case RightToLeft.No : lbl.Alignment = HorizontalAlignment.Left
                            Case RightToLeft.Yes : lbl.Alignment = HorizontalAlignment.Right
                            Case RightToLeft.Inherit
                        End Select
                End Select
                lbl = Nothing
            ElseIf TypeOf GetMember Is Model10.Button Then
                Dim btn As Model10.Button = CType(GetMember, Model10.Button)
                btn.Text = GetMember.Name
                btn = Nothing
            End If
        End If
    End Function
    Private Function GetRect() As Rectangle
        GetRect = New Rectangle
        If dX < 0 Then
            GetRect.X = oX + dX
            GetRect.Width = -dX
        Else
            GetRect.X = oX
            GetRect.Width = dX
        End If
        If dY < 0 Then
            GetRect.Y = oY + dY
            GetRect.Height = -dY
        Else
            GetRect.Y = oY
            GetRect.Height = dY
        End If
    End Function
    Private Function GetTop(ByVal s As Model10.Section) As Integer
        If Not mTable.PageHeader Is Nothing Then
            GetTop += HEADER_HEIGHT
            If s Is mTable.PageHeader Then Exit Function
            GetTop += mTable.PageHeader.Height
        End If
        If Not mTable.GroupHeader Is Nothing Then
            GetTop += HEADER_HEIGHT
            If s Is mTable.GroupHeader Then Exit Function
            GetTop += mTable.GroupHeader.Height
        End If
        If Not mTable.Detail Is Nothing Then
            GetTop += HEADER_HEIGHT
            If s Is mTable.Detail Then Exit Function
            GetTop += mTable.Detail.Height
        End If
        If Not mTable.GroupFooter Is Nothing Then
            GetTop += HEADER_HEIGHT
            If s Is mTable.GroupFooter Then Exit Function
            GetTop += mTable.GroupFooter.Height
        End If
        If Not mTable.PageFooter Is Nothing Then
            GetTop += HEADER_HEIGHT
            If s Is mTable.PageFooter Then Exit Function
            GetTop += mTable.PageFooter.Height
        End If
    End Function
    Private Function NormX(ByVal x As Integer) As Integer
        If mTable.SnapToGrid Then
            Return CInt(x / mTable.GridSize.Width) * mTable.GridSize.Width
        Else
            Return x
        End If
    End Function
    Private Function NormY(ByVal y As Integer) As Integer
        If mTable.SnapToGrid Then
            Return CInt(y / mTable.GridSize.Height) * mTable.GridSize.Height
        Else
            Return y
        End If
    End Function

    Private Sub Hsb_ValueChanged(ByVal sender As Object, ByVal e As System.EventArgs)
        Dim sb As ScrollBar = CType(sender, ScrollBar)
        offset.X = -sb.Value
        sb = Nothing

        Invalidate()
    End Sub
    Private Sub Vsb_ValueChanged(ByVal sender As Object, ByVal e As System.EventArgs)
        Dim sb As ScrollBar = CType(sender, ScrollBar)
        offset.Y = -sb.Value
        sb = Nothing

        Invalidate()
    End Sub

    Private Sub Canceling()
        If rubbing Then
            If Not canc Then DrawRubber(GetRect)
            pos = MousePos.None
            Me.Cursor = Cursors.Default
            rubbing = False
        End If
        If moving Then
            If Not canc Then DrawLocation()
            pos = MousePos.None
            Me.Cursor = Cursors.Default
            moving = False
        End If
        If verting Then
            If Not canc Then DrawVertic()
            pos = MousePos.None
            Me.Cursor = Cursors.Default
            verting = False
        End If
        If horizing Then
            If Not canc Then DrawHoriz()
            pos = MousePos.None
            Me.Cursor = Cursors.Default
            horizing = False
        End If
        If sizing Then
            If Not canc Then DrawSize()
            pos = MousePos.None
            Me.Cursor = Cursors.Default
            sizing = False
        End If
        If inserting Then
            If Not canc Then DrawInsert(GetRect)
            pos = MousePos.None
            Me.Cursor = Cursors.Default
            inserting = False
        End If

        If Me.Cursor Is Cursors.Cross Then
            Me.Cursor = Cursors.Default
            OnInsertingFinished()
        End If
    End Sub
    Private Sub ClearCells()
        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
            cells.Remove(cel)
            cel = Nothing
        Next
    End Sub
    Private Sub DrawHoriz()
        Dim y As Integer = dY + GetTop(selSection) + offset.Y
        Dim w As Integer : If Not vsb Is Nothing Then w = 16
        ControlPaint.DrawReversibleLine(PointToScreen(New Point(0, y)), PointToScreen(New Point(Width - w, y)), SystemColors.Window)
    End Sub
    Private Sub DrawLocation()
        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            ControlPaint.DrawReversibleFrame(RectangleToScreen(New Rectangle(offset.X + NormX(cel.Member.Location.X + dX) + FRAME_CORRECTION, offset.Y + GetTop(cel.Member.Section) + NormY(cel.Member.Location.Y + dY) + FRAME_CORRECTION, cel.Member.Size.Width - FRAME_CORRECTION, cel.Member.Size.Height - FRAME_CORRECTION)), cel.Member.BackColor, FrameStyle.Thick)
        Next : cel = Nothing
    End Sub
    Private Sub DrawMembers(ByVal s As Model10.Section)
        If s.Bitmap Is Nothing Then
            s.Bitmap = New Bitmap(mTable.Width, s.Height)
        Else
            If s.Bitmap.Size.Width <> mTable.Width Or s.Bitmap.Size.Height <> s.Height Then
                s.Bitmap.Dispose()
                s.Bitmap = New Bitmap(mTable.Width, s.Height)
            End If
        End If

        Dim g As Graphics = Graphics.FromImage(s.Bitmap)
        Dim b As New SolidBrush(SystemColors.Window)

        Dim rc As New Rectangle(New Point, s.Bitmap.Size)
        g.FillRectangle(b, rc)
        If mTable.DisplayGrid Then ControlPaint.DrawGrid(g, rc, mTable.GridSize, SystemColors.Window)
        rc = Nothing

        Dim mem As Model10.Member
        For Each mem In s.Members
            Dim foo As String = mem.Name
            If mem.GetType Is GetType(Model10.Line) Then
                CType(mem, Model10.Line).Draw(g, 0, 0)
            ElseIf mem.GetType Is GetType(Model10.Frame) Then
                CType(mem, Model10.Frame).Draw(g, 0, 0)
            ElseIf TypeOf mem Is Model10.Label Then
                Dim lbl As Model10.Label = CType(mem, Model10.Label)
                Select Case lbl.Value
                    Case Model10.ValueType.None
                        If lbl.DataField = "" Then
                            If lbl.Text <> "" Then foo = lbl.Text
                        Else
                            foo = "[" & lbl.DataField & "]"
                        End If
                        '------------------
                    Case Model10.ValueType.DateTime
                        foo = Model10.FFormat(Now, lbl.Format)
                    Case Model10.ValueType.Title
                        If mTable.Presentation.Name = "" Then
                            foo = Model10.FFormat("{" & "Title" & "}", lbl.Format)
                        Else
                            foo = Model10.FFormat(mTable.Presentation.Name, lbl.Format)
                        End If
                    Case Model10.ValueType.Path
                        If mTable.Presentation.Path = "" Then
                            foo = Model10.FFormat("{" & "Path" & "}", lbl.Format)
                        Else
                            foo = Model10.FFormat(mTable.Presentation.Path, lbl.Format)
                        End If
                    Case Model10.ValueType.PageNumber : foo = "{" & "Page #" & "}"
                    Case Model10.ValueType.PageCount : foo = "{" & "# of Pages" & "}"
                    Case Model10.ValueType.RecordNumber : foo = "{" & "Record #" & "}"
                    Case Model10.ValueType.RecordCount : foo = "{" & "# of Records" & "}"
                    Case Model10.ValueType.UserId
                        If mUser Is Nothing Then
                            foo = "{" & Model10.FFormat("User Id", lbl.Format) & "}"
                        Else
                            foo = Model10.FFormat(mUser.Id, lbl.Format)
                        End If
                    Case Model10.ValueType.UserName
                        If mUser Is Nothing Then
                            foo = "{" & Model10.FFormat("User Full Name", lbl.Format) & "}"
                        Else
                            foo = Model10.FFormat(mUser.FullName, lbl.Format)
                        End If
                    Case Model10.ValueType.UserFullName
                        If mUser Is Nothing Then
                            foo = "{" & Model10.FFormat("User Name", lbl.Format) & "}"
                        Else
                            foo = Model10.FFormat(mUser.Name, lbl.Format)
                        End If
                End Select
                lbl.Draw(g, foo, 0, 0, Me.RightToLeft, Nothing)
                lbl = Nothing
            ElseIf TypeOf mem Is Model10.PictureBox Then
                Dim pic As Model10.PictureBox = CType(mem, Model10.PictureBox)
                If pic.Image Is Nothing Then
                    If pic.DataField = "" Then
                        pic.Draw(g, foo, Font, 0, 0)
                    Else
                        pic.Draw(g, "[" & pic.DataField & "]", Font, 0, 0)
                    End If
                Else
                    pic.Draw(g, 0, 0)
                End If
                pic = Nothing
            ElseIf TypeOf mem Is Model10.CheckBox Then : CType(mem, Model10.CheckBox).Draw(g, 0, 0)
            ElseIf TypeOf mem Is Model10.Button Then
                Dim btn As Model10.Button = CType(mem, Model10.Button)
                If btn.Text <> "" Then foo = btn.Text
                btn.Draw(g, foo, 0, 0, Me.RightToLeft, Nothing)
                btn = Nothing
            ElseIf TypeOf mem Is Model10.Sub Then
                Dim sb As Model10.Sub = CType(mem, Model10.Sub)
                If Not sb.ChildRelation Is Nothing Then foo = "<" & sb.ChildRelation.Name & ">"
                sb.Draw(g, foo, Font, 0, 0)
                sb = Nothing
            ElseIf TypeOf mem Is Model10.Chart Then
                Dim ch As Model10.Chart = CType(mem, Model10.Chart)
                ch.DrawChart(g, 0, 0)
                ch = Nothing
            End If
        Next : mem = Nothing

        b.Dispose() : b = Nothing

        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            If cel.Member.Section Is s Then
                Dim i As New Rectangle(cel.Member.Location, cel.Member.Size)
                Dim o As Rectangle = i : o.Inflate(GRAB_SIZE, GRAB_SIZE)
                ControlPaint.DrawSelectionFrame(g, False, o, i, SystemColors.Window)
                o = Nothing
                i = Nothing

                Dim grb As Avand.Designer.Grab
                For Each grb In cel.Grabs
                    ControlPaint.DrawGrabHandle(g, grb.Rectangle, Not mSelectedMember Is cel.Member, True)
                Next : grb = Nothing
            End If
        Next : cel = Nothing

        g.Dispose() : g = Nothing
    End Sub
    Private Sub DrawRubber(ByVal r As Rectangle)
        ControlPaint.DrawReversibleFrame(RectangleToScreen(r), SystemColors.Window, FrameStyle.Dashed)
    End Sub
    Private Sub DrawSection(ByVal g As Graphics, ByVal s As Model10.Section, ByVal t As String, ByRef y As Integer)
        Dim r As New RectangleF(0, y, ClientRectangle.Width + 2, HEADER_HEIGHT)
        Dim b As Brush

        ControlPaint.DrawBorder3D(g, Rectangle.Truncate(r), Border3DStyle.Raised)

        If mSelectedSection Is s And cells.Count = 0 Then b = SystemBrushes.Highlight Else b = SystemBrushes.Control
        r.Inflate(-2, -2)
        g.FillRectangle(b, r)

        If mSelectedSection Is s And cells.Count = 0 Then b = SystemBrushes.HighlightText Else b = SystemBrushes.ControlText
        'r.Inflate(0, 0)
        g.DrawString(GetSystemResource("Section." & s.GetName), Font, b, r)

        b = Nothing
        r = Nothing

        y += HEADER_HEIGHT
        If s.Bitmap Is Nothing Then DrawMembers(s)
        g.DrawImage(s.Bitmap, offset.X, y)
        y += s.Bitmap.Height
    End Sub
    Private Sub DrawSize()
        Dim l, t, w, h As Integer
        Select Case selGrab.Status
            Case GrabStatus.TopLeft : l = dX : t = dY : w = -dX : h = -dY
            Case GrabStatus.Top : t = dY : h = -dY
            Case GrabStatus.TopRight : t = dY : w = dX : h = -dY
            Case GrabStatus.Right : w = dX
            Case GrabStatus.BottomRight : w = dX : h = dY
            Case GrabStatus.Bottom : h = dY
            Case GrabStatus.BottomLeft : l = dX : w = -dX : h = dY
            Case GrabStatus.Left : l = dX : w = -dX
        End Select

        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            Dim rc As New Rectangle(cel.Member.Location.X + l, cel.Member.Location.Y + t, cel.Member.Size.Width + w, cel.Member.Size.Height + h)
            NormRect(rc)
            rc.X += (offset.X + FRAME_CORRECTION) : rc.Y += (offset.Y + GetTop(cel.Member.Section) + FRAME_CORRECTION) : rc.Width -= FRAME_CORRECTION : rc.Height -= FRAME_CORRECTION
            ControlPaint.DrawReversibleFrame(RectangleToScreen(rc), cel.Member.BackColor, FrameStyle.Thick)
            rc = Nothing
        Next : cel = Nothing
    End Sub
    Private Sub DrawVertic()
        Dim x As Integer = dX + offset.X
        Dim h As Integer : If Not hsb Is Nothing Then h = 16
        ControlPaint.DrawReversibleLine(PointToScreen(New Point(x, 0)), PointToScreen(New Point(x, Height - h)), SystemColors.Window)
    End Sub
    Private Sub DrawInsert(ByVal r As Rectangle)
        ControlPaint.DrawReversibleFrame(RectangleToScreen(r), SystemColors.Window, FrameStyle.Thick)
    End Sub
    Private Sub Insert(ByVal m As Model10.Member)
        mSelectedSection.Members.Add(m, 0, 0)
        Dim over As Boolean
        Dim mX As Integer = m.Location.X + m.Size.Width
        Dim mY As Integer = m.Location.Y + m.Size.Height
        If mX > mTable.Width Then
            mTable.Width = mX
            Refine()
            over = True
        End If
        If mY > mSelectedSection.Height Then
            totalHeight = totalHeight + mY - mSelectedSection.Height
            mSelectedSection.Height = mY
            over = True
        End If
        Dim emt As Boolean = cells.Count = 0
        ClearCells()
        OneCell(mSelectedSection, m)
        If Not draws.Contains(mSelectedSection) Then draws.Add(mSelectedSection)
        mSelectedMember = m
        'mSelectedSection = mSelectedSection
        If over Then Resizing()
        Redraw()
        If emt Then OnMembersSelected()
        OnSelectedMembersChanged()
        OnValueChanged()
    End Sub
    Private Sub NormRect(ByRef r As Rectangle)
        Dim x, y As Integer
        Select Case selGrab.Status
            Case GrabStatus.TopLeft
                x = NormX(r.X) - r.X : r.X += x : r.Width -= x
                y = NormY(r.Y) - r.Y : r.Y += y : r.Height -= y
            Case GrabStatus.Top
                y = NormY(r.Y) - r.Y : r.Y += y : r.Height -= y
            Case GrabStatus.TopRight
                x = NormX(r.X + r.Width) - (r.X + r.Width) : r.Width += x
                y = NormY(r.Y) - r.Y : r.Y += y : r.Height -= y
            Case GrabStatus.Right
                x = NormX(r.X + r.Width) - (r.X + r.Width) : r.Width += x
            Case GrabStatus.BottomRight
                x = NormX(r.X + r.Width) - (r.X + r.Width) : r.Width += x
                y = NormY(r.Y + r.Height) - (r.Y + r.Height) : r.Height += y
            Case GrabStatus.Bottom
                y = NormY(r.Y + r.Height) - (r.Y + r.Height) : r.Height += y
            Case GrabStatus.BottomLeft
                x = NormX(r.X) - r.X : r.X += x : r.Width -= x
                y = NormY(r.Y + r.Height) - (r.Y + r.Height) : r.Height += y
            Case GrabStatus.Left
                x = NormX(r.X) - r.X : r.X += x : r.Width -= x
        End Select
        If r.Width < MIN_SIZE Then r.Width = MIN_SIZE
        If r.Height < MIN_SIZE Then r.Height = MIN_SIZE
    End Sub
    Private Sub OneCell(ByVal s As Model10.Section, ByVal m As Model10.Member)
        Dim c As New Avand.Designer.Cell
        c.Member = m
        c.Grabs = New ArrayList
        '---------------
        Dim i As Integer
        For i = 0 To 7
            Dim g As New Avand.Designer.Grab
            Select Case i
                Case 0
                    g.Rectangle = New Rectangle(m.Location.X - GRAB_SIZE, m.Location.Y - GRAB_SIZE, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.TopLeft
                    g.Cursor = Cursors.SizeNWSE
                Case 1
                    g.Rectangle = New Rectangle(m.Location.X + m.Size.Width \ 2 - GRAB_SIZE \ 2, m.Location.Y - GRAB_SIZE, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.Top
                    g.Cursor = Cursors.SizeNS
                Case 2
                    g.Rectangle = New Rectangle(m.Location.X + m.Size.Width, m.Location.Y - GRAB_SIZE, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.TopRight
                    g.Cursor = Cursors.SizeNESW
                Case 3
                    g.Rectangle = New Rectangle(m.Location.X + m.Size.Width, m.Location.Y + m.Size.Height \ 2 - GRAB_SIZE \ 2, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.Right
                    g.Cursor = Cursors.SizeWE
                Case 4
                    g.Rectangle = New Rectangle(m.Location.X + m.Size.Width, m.Location.Y + m.Size.Height, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.BottomRight
                    g.Cursor = Cursors.SizeNWSE
                Case 5
                    g.Rectangle = New Rectangle(m.Location.X + m.Size.Width \ 2 - GRAB_SIZE \ 2, m.Location.Y + m.Size.Height, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.Bottom
                    g.Cursor = Cursors.SizeNS
                Case 6
                    g.Rectangle = New Rectangle(m.Location.X - GRAB_SIZE, m.Location.Y + m.Size.Height, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.BottomLeft
                    g.Cursor = Cursors.SizeNESW
                Case 7
                    g.Rectangle = New Rectangle(m.Location.X - GRAB_SIZE, m.Location.Y + m.Size.Height \ 2 - GRAB_SIZE \ 2, GRAB_SIZE, GRAB_SIZE)
                    g.Status = GrabStatus.Left
                    g.Cursor = Cursors.SizeWE
            End Select
            c.Grabs.Add(g)
            g = Nothing
        Next
        '---------------
        cells.Add(c)
        c = Nothing
    End Sub
    Private Sub Redraw()
        Dim i As Integer
        For i = draws.Count - 1 To 0 Step -1
            Dim sec As Model10.Section = CType(draws(i), Model10.Section)
            DrawMembers(sec)
            draws.Remove(sec)
            sec = Nothing
        Next

        Invalidate()
    End Sub
    Private Sub Refine()
        If Not mTable.PageHeader Is Nothing Then
            If Not draws.Contains(mTable.PageHeader) Then draws.Add(mTable.PageHeader)
        End If
        If Not mTable.GroupHeader Is Nothing Then
            If Not draws.Contains(mTable.GroupHeader) Then draws.Add(mTable.GroupHeader)
        End If
        If Not mTable.Detail Is Nothing Then
            If Not draws.Contains(mTable.Detail) Then draws.Add(mTable.Detail)
        End If
        If Not mTable.GroupFooter Is Nothing Then
            If Not draws.Contains(mTable.GroupFooter) Then draws.Add(mTable.GroupFooter)
        End If
        If Not mTable.PageFooter Is Nothing Then
            If Not draws.Contains(mTable.PageFooter) Then draws.Add(mTable.PageFooter)
        End If
    End Sub
    Private Sub Resizing()
        Static pic As PictureBox
        If Not mTable Is Nothing Then
            Dim h, v As Boolean
            Dim d As Integer

            If mTable.Width + MARGIN_SIZE <= Width And totalHeight + MARGIN_SIZE <= Height Then
                h = False : v = False
            ElseIf mTable.Width + MARGIN_SIZE <= Width - 16 And Height < totalHeight + MARGIN_SIZE Then
                h = False : v = True
            ElseIf Width < mTable.Width + MARGIN_SIZE And totalHeight + MARGIN_SIZE <= Height - 16 Then
                h = True : v = False
            Else
                h = True : v = True
            End If

            If h Then
                d = mTable.Width - (Width - CInt(IIf(v, 16, 0))) + MARGIN_SIZE
                If hsb Is Nothing Then
                    hsb = New HScrollBar
                    hsb.Height = 16
                    hsb.RightToLeft = RightToLeft.No
                    hsb.SmallChange = 10
                    hsb.LargeChange = 100
                    AddHandler hsb.ValueChanged, AddressOf Hsb_ValueChanged
                    Controls.Add(hsb)
                    hsb.Cursor = Cursors.Default
                End If
                hsb.Maximum = d + hsb.LargeChange - 1
                hsb.Top = Height - 16
                hsb.Width = Width - CInt(IIf(v, 16, 0))
                If d < -offset.X Then hsb.Value = d
            Else
                If Not hsb Is Nothing Then
                    hsb.Value = 0
                    Controls.Remove(hsb)
                    hsb = Nothing
                End If
            End If

            If v Then
                d = totalHeight - (Height - CInt(IIf(h, 16, 0))) + MARGIN_SIZE
                If vsb Is Nothing Then
                    vsb = New VScrollBar
                    vsb.Width = 16
                    vsb.RightToLeft = RightToLeft.No
                    vsb.SmallChange = 10
                    vsb.LargeChange = 100
                    AddHandler vsb.ValueChanged, AddressOf Vsb_ValueChanged
                    Controls.Add(vsb)
                    vsb.Cursor = Cursors.Default
                End If
                vsb.Height = Height - CInt(IIf(h, 16, 0))
                vsb.Left = Width - 16
                vsb.Maximum = d + vsb.LargeChange - 1
                If d < -offset.Y Then vsb.Value = d
            Else
                If Not vsb Is Nothing Then
                    vsb.Value = 0
                    Controls.Remove(vsb)
                    vsb = Nothing
                End If
            End If

            If Not vsb Is Nothing And Not hsb Is Nothing Then
                If pic Is Nothing Then
                    pic = New PictureBox
                    pic.BackColor = SystemColors.Control
                    Controls.Add(pic)
                    pic.BringToFront()
                End If
                pic.Size = New Size(16, 16)
                pic.Left = vsb.Left
                pic.Top = hsb.Top
            Else
                If Not pic Is Nothing Then
                    Controls.Remove(pic)
                    pic = Nothing
                End If
            End If
        Else
            If Not hsb Is Nothing Then
                hsb.Value = 0
                hsb = Nothing
            End If
            If Not vsb Is Nothing Then
                vsb.Value = 0
                vsb = Nothing
            End If
            offset.X = 0 : offset.Y = 0
            If Not pic Is Nothing Then pic = Nothing
        End If
    End Sub
    Private Sub ReverseMembers(ByVal s As Model10.Section, ByVal r As Rectangle, ByRef y As Integer)
        y += HEADER_HEIGHT

        Dim mem As Model10.Member
        For Each mem In s.Members
            Dim f As New Rectangle(mem.Location, mem.Size)
            f.Offset(offset.X, y)
            If Not r.Intersect(f, r).IsEmpty Then
                Dim foo As Avand.Designer.Cell

                Dim cel As Avand.Designer.Cell
                For Each cel In cells
                    If cel.Member Is mem Then
                        foo = cel
                        Exit For
                    End If
                Next : cel = Nothing

                If (Control.ModifierKeys And Keys.Control) = Keys.Control Then
                    If foo Is Nothing Then
                        OneCell(s, mem)
                        If Not draws.Contains(s) Then draws.Add(s)
                    Else
                        cells.Remove(foo)
                        If Not draws.Contains(s) Then draws.Add(s)
                        If mSelectedMember Is mem Then mSelectedMember = Nothing
                    End If
                Else
                    OneCell(s, mem)
                    If Not draws.Contains(s) Then draws.Add(s)
                End If
            End If
            f = Nothing
        Next : mem = Nothing

        y += s.Height
    End Sub
    Private Sub Selecting(ByVal p As Point, Optional ByVal skip As Boolean = False)
        If p.X < 0 Or p.X > Width Or p.Y < 0 Or p.Y > Height Then
            If Not Cursor Is Cursors.Cross Then Cursor = Cursors.Default
            pos = MousePos.None
            selSection = Nothing
            selMember = Nothing
            selGrab = Nothing
            Return
        End If

        If Not mTable Is Nothing Then
            Dim x, y As Integer

            If Not Me.Cursor Is Cursors.Cross And Not skip Then
                y = offset.Y
                If Not mTable.PageHeader Is Nothing Then If CheckResize(mTable.PageHeader, p, y) Then Exit Sub
                If Not mTable.GroupHeader Is Nothing Then If CheckResize(mTable.GroupHeader, p, y) Then Exit Sub
                If Not mTable.Detail Is Nothing Then If CheckResize(mTable.Detail, p, y) Then Exit Sub
                If Not mTable.GroupFooter Is Nothing Then If CheckResize(mTable.GroupFooter, p, y) Then Exit Sub
                If Not mTable.PageFooter Is Nothing Then If CheckResize(mTable.PageFooter, p, y) Then Exit Sub

                x = offset.X + mTable.Width
                If x - SIDE_SIZE <= p.X And p.X < x + SIDE_SIZE Then
                    Cursor = Cursors.SizeWE
                    pos = MousePos.ResizeTableWidth
                    selSection = Nothing
                    selMember = Nothing
                    selGrab = Nothing
                    Return
                End If
            End If

            y = offset.Y
            If Not mTable.PageHeader Is Nothing Then If CheckSection(mTable.PageHeader, p, y) Then Return
            If Not mTable.GroupHeader Is Nothing Then If CheckSection(mTable.GroupHeader, p, y) Then Return
            If Not mTable.Detail Is Nothing Then If CheckSection(mTable.Detail, p, y) Then Return
            If Not mTable.GroupFooter Is Nothing Then If CheckSection(mTable.GroupFooter, p, y) Then Return
            If Not mTable.PageFooter Is Nothing Then If CheckSection(mTable.PageFooter, p, y) Then Return
        End If

        If Not Cursor Is Cursors.Cross Then Cursor = Cursors.Default
        pos = MousePos.Table
        selSection = Nothing
        selMember = Nothing
        selGrab = Nothing
        Return
    End Sub
    Private Sub SetGrabs(ByVal c As Avand.Designer.Cell)
        Dim grb As Avand.Designer.Grab
        For Each grb In c.Grabs
            Select Case grb.Status
                Case GrabStatus.TopLeft : grb.Rectangle.Location = New Point(c.Member.Location.X - GRAB_SIZE, c.Member.Location.Y - GRAB_SIZE)
                Case GrabStatus.Top : grb.Rectangle.Location = New Point(c.Member.Location.X + c.Member.Size.Width \ 2 - GRAB_SIZE \ 2, c.Member.Location.Y - GRAB_SIZE)
                Case GrabStatus.TopRight : grb.Rectangle.Location = New Point(c.Member.Location.X + c.Member.Size.Width, c.Member.Location.Y - GRAB_SIZE)
                Case GrabStatus.Right : grb.Rectangle.Location = New Point(c.Member.Location.X + c.Member.Size.Width, c.Member.Location.Y + c.Member.Size.Height \ 2 - GRAB_SIZE \ 2)
                Case GrabStatus.BottomRight : grb.Rectangle.Location = New Point(c.Member.Location.X + c.Member.Size.Width, c.Member.Location.Y + c.Member.Size.Height)
                Case GrabStatus.Bottom : grb.Rectangle.Location = New Point(c.Member.Location.X + c.Member.Size.Width \ 2 - GRAB_SIZE \ 2, c.Member.Location.Y + c.Member.Size.Height)
                Case GrabStatus.BottomLeft : grb.Rectangle.Location = New Point(c.Member.Location.X - GRAB_SIZE, c.Member.Location.Y + c.Member.Size.Height)
                Case GrabStatus.Left : grb.Rectangle.Location = New Point(c.Member.Location.X - GRAB_SIZE, c.Member.Location.Y + c.Member.Size.Height \ 2 - GRAB_SIZE \ 2)
            End Select
        Next : grb = Nothing
    End Sub
    Private Sub SetTotalHeight()
        totalHeight = 0
        If Not mTable.PageHeader Is Nothing Then totalHeight += mTable.PageHeader.Height + HEADER_HEIGHT
        If Not mTable.GroupHeader Is Nothing Then totalHeight += mTable.GroupHeader.Height + HEADER_HEIGHT
        If Not mTable.Detail Is Nothing Then totalHeight += mTable.Detail.Height + HEADER_HEIGHT
        If Not mTable.GroupFooter Is Nothing Then totalHeight += mTable.GroupFooter.Height + HEADER_HEIGHT
        If Not mTable.PageFooter Is Nothing Then totalHeight += mTable.PageFooter.Height + HEADER_HEIGHT
    End Sub

    Protected Overridable Sub OnInsertingFinished()
        RaiseEvent InsertingFinished(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnRequestDelete()
        RaiseEvent RequestDelete(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnValueChanged()
        RaiseEvent ValueChanged(Me, New EventArgs)
    End Sub

    Protected Overridable Sub OnTableDoubleClick()
        RaiseEvent TableDoubleClick(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnSectionDoubleClick()
        RaiseEvent SectionDoubleClick(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnMembersDoubleClick()
        RaiseEvent MembersDoubleClick(Me, New EventArgs)
    End Sub

    Protected Overridable Sub OnTableSelected()
        RaiseEvent TableSelected(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnSectionSelected()
        RaiseEvent SectionSelected(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnMembersSelected()
        RaiseEvent MembersSelected(Me, New EventArgs)
    End Sub

    Protected Overridable Sub OnSelectedTableChanged()
        RaiseEvent SelectedTableChanged(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnSelectedSectionChanged()
        RaiseEvent SelectedSectionChanged(Me, New EventArgs)
    End Sub
    Protected Overridable Sub OnSelectedMembersChanged()
        RaiseEvent SelectedMembersChanged(Me, New EventArgs)
    End Sub

    Protected Overrides Sub OnDoubleClick(ByVal e As System.EventArgs)
        MyBase.OnDoubleClick(e)

        If Cursor Is Cursors.Cross Then
            Canceling()
            pos = MousePos.None
            selSection = Nothing
            selMember = Nothing
            selGrab = Nothing
        End If

        Select Case pos
            Case MousePos.None
            Case MousePos.Table : OnTableDoubleClick()
            Case MousePos.Header, MousePos.Section : OnSectionDoubleClick()
            Case MousePos.Member, MousePos.Grab : OnMembersDoubleClick()
            Case MousePos.ResizeTableWidth
                Dim mX As Integer = mTable.GetMinWidth()
                If mX < mTable.Width Then
                    mTable.Width = mX
                    Refine()
                    Resizing()
                    Redraw()
                    OnValueChanged()
                End If
            Case MousePos.ResizeSectionHeight
                Dim mY As Integer = selSection.GetMinHeight()
                If mY < selSection.Height Then
                    totalHeight = totalHeight + mY - selSection.Height
                    selSection.Height = mY
                    If Not draws.Contains(selSection) Then draws.Add(selSection)
                    Resizing()
                    Redraw()
                    OnValueChanged()
                End If
        End Select
    End Sub
    Protected Overrides Sub OnDragEnter(ByVal drgevent As System.Windows.Forms.DragEventArgs)
        MyBase.OnDragEnter(drgevent)

    End Sub
    Protected Overrides Sub OnDragLeave(ByVal e As System.EventArgs)
        MyBase.OnDragLeave(e)

    End Sub
    Protected Overrides Sub OnDragDrop(ByVal drgevent As System.Windows.Forms.DragEventArgs)
        MyBase.OnDragDrop(drgevent)

    End Sub
    Protected Overrides Sub OnDragOver(ByVal drgevent As System.Windows.Forms.DragEventArgs)
        MyBase.OnDragOver(drgevent)

    End Sub
    Protected Overrides Sub OnKeyDown(ByVal e As System.Windows.Forms.KeyEventArgs)
        MyBase.OnKeyDown(e)

        Select Case e.KeyCode
            Case Keys.Delete : OnRequestDelete()
            Case Keys.Escape : Canceling()
        End Select
    End Sub
    Protected Overrides Sub OnLostFocus(ByVal e As System.EventArgs)
        MyBase.OnLostFocus(e)

        Canceling()
    End Sub
    Protected Overrides Sub OnMouseEnter(ByVal e As System.EventArgs)
        MyBase.OnMouseEnter(e)

        'Dim p As Point = Me.PointToClient(Me.MousePosition)
        'oX = p.X : oY = p.Y
    End Sub
    Protected Overrides Sub OnMouseLeave(ByVal e As System.EventArgs)
        MyBase.OnMouseLeave(e)

        If Not Cursor Is Cursors.Cross Then Cursor = Cursors.Default
        pos = MousePos.None
        selSection = Nothing
        selMember = Nothing
        selGrab = Nothing
    End Sub
    Protected Overrides Sub OnQueryContinueDrag(ByVal qcdevent As System.Windows.Forms.QueryContinueDragEventArgs)
        MyBase.OnQueryContinueDrag(qcdevent)

    End Sub

    Protected Overrides Sub OnMouseDown(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseDown(e)
        If e.Button = MouseButtons.Left Then lMouse = True

        Dim cnt As Integer = cells.Count
        Dim sec As Model10.Section = mSelectedSection
        If (e.Button = MouseButtons.Left And (MouseButtons And MouseButtons.Right) <> MouseButtons.Right) Or (e.Button = MouseButtons.Right) Then
            If (e.Button = MouseButtons.Right) Then
                Canceling()
                Selecting(New Point(e.X, e.Y), True)
                Me.Cursor = Cursors.Default
            Else
                Selecting(New Point(e.X, e.Y))
            End If
            oX = e.X : oY = e.Y
            dX = 0 : dY = 0
            Select Case pos
                Case MousePos.None
                Case MousePos.Table
                    If (Control.ModifierKeys And Keys.Control) <> Keys.Control Then
                        ClearCells()
                        mSelectedSection = Nothing
                        mSelectedMember = Nothing
                        Redraw()
                        If cnt > 0 Or Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
                        sec = Nothing
                    End If
                Case MousePos.Header, MousePos.Section
                    If (Control.ModifierKeys And Keys.Control) <> Keys.Control Then
                        Dim foo As Boolean = mSelectedMember Is Nothing
                        ClearCells()
                        mSelectedSection = selSection
                        mSelectedMember = selMember
                        Redraw()
                        If cnt > 0 Or sec Is Nothing Then OnSectionSelected() ': OnSelectedSectionChanged()
                        If Not sec Is selSection Or Not foo Then OnSelectedSectionChanged()
                        sec = Nothing
                    End If
                Case MousePos.Member, MousePos.Grab
                    Dim mem As Model10.Member = mSelectedMember
                    '-----------------------------
                    Dim foo As Avand.Designer.Cell

                    Dim cel As Avand.Designer.Cell
                    For Each cel In cells
                        If cel.Member Is selMember Then
                            foo = cel
                            Exit For
                        End If
                    Next : cel = Nothing

                    If (Control.ModifierKeys And Keys.Control) = Keys.Control Then
                        If foo Is Nothing Then
                            OneCell(selSection, selMember)
                            If Not draws.Contains(selSection) Then draws.Add(selSection)
                            If Not mSelectedSection Is Nothing Then If Not draws.Contains(mSelectedSection) Then draws.Add(mSelectedSection)
                            mSelectedMember = selMember
                        Else
                            cells.Remove(foo)
                            If Not draws.Contains(selSection) Then draws.Add(selSection)
                            If Not mSelectedSection Is Nothing Then If Not draws.Contains(mSelectedSection) Then draws.Add(mSelectedSection)
                            mSelectedMember = Nothing
                        End If
                    Else
                        If foo Is Nothing Then
                            ClearCells()
                            OneCell(selSection, selMember)
                            If Not draws.Contains(selSection) Then draws.Add(selSection)
                            mSelectedMember = selMember
                        Else
                            If Not mSelectedMember Is selMember Then
                                If Not draws.Contains(selSection) Then draws.Add(selSection)
                                If Not mSelectedSection Is Nothing Then If Not draws.Contains(mSelectedSection) Then draws.Add(mSelectedSection)
                                mSelectedMember = selMember
                            End If
                        End If
                    End If
                    mSelectedSection = selSection
                    '-----------------------------
                    Redraw()
                    If cnt > 0 Then
                        If cells.Count = 0 Then
                            OnSectionSelected() : OnSelectedSectionChanged()
                        Else
                            If Not mSelectedMember Is mem Then OnSelectedMembersChanged()
                        End If
                    Else 'cnt = 0
                        If cells.Count > 0 Then OnMembersSelected() : OnSelectedMembersChanged()
                    End If
                    mem = Nothing
                Case MousePos.ResizeTableWidth
                Case MousePos.ResizeSectionHeight
            End Select
        End If
    End Sub
    Protected Overrides Sub OnMouseMove(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseMove(e)

        Dim x, y As Integer
        Static mX, mY, mW, mH As Integer

        If e.Button = MouseButtons.Left Then
            If lMouse Then
                Select Case pos
                    Case MousePos.None
                    Case MousePos.Table
                    Case MousePos.Header
                    Case MousePos.Section
                        If Me.Cursor Is Cursors.Cross Then
                            If Not inserting Then
                                If (Math.Abs(e.X - oX) >= mTable.GridSize.Width Or Math.Abs(e.Y - oY) >= mTable.GridSize.Height) Then
                                    GoTo InsertingDiff
                                End If
                            Else
                                If canc Then
                                    canc = False
                                Else
                                    DrawInsert(GetRect)
                                End If
                                GoTo InsertingDiff
                            End If
                        Else
                            If Not rubbing Then
                                If e.X <> oX Or e.Y <> oY Then
                                    rubbing = True
                                    canc = False
                                    GoTo RubbingDiff
                                End If
                            Else
                                If canc Then
                                    canc = False
                                Else
                                    DrawRubber(GetRect)
                                End If
                                GoTo RubbingDiff
                            End If
                        End If
                    Case MousePos.Member
                        If Not moving Then
                            If (Math.Abs(e.X - oX) >= mTable.GridSize.Width Or Math.Abs(e.Y - oY) >= mTable.GridSize.Height) And mSelectedMember Is selMember Then
                                mX = 10000 : mY = 10000

                                Dim cel As Avand.Designer.Cell
                                For Each cel In cells
                                    If mX > cel.Member.Location.X Then mX = cel.Member.Location.X
                                    If mY > cel.Member.Location.Y Then mY = cel.Member.Location.Y
                                Next : cel = Nothing

                                If e.X < 0 Or e.X > Width Or e.Y < 0 Or e.Y > Height Then
                                    Cursor = Cursors.No
                                    canc = True
                                Else
                                    GoTo MovingDiff
                                End If
                            End If
                        Else
                            If canc Then
                                canc = False
                            Else
                                DrawLocation()
                            End If
                            If e.X < 0 Or e.X > Width Or e.Y < 0 Or e.Y > Height Then
                                Cursor = Cursors.No
                                canc = True
                            Else
                                GoTo MovingDiff
                            End If
                        End If
                    Case MousePos.Grab
                        If Not sizing Then
                            If (Math.Abs(e.X - oX) >= mTable.GridSize.Width Or Math.Abs(e.Y - oY) >= mTable.GridSize.Height) And mSelectedMember Is selMember Then
                                mX = 10000 : mY = 10000
                                mW = 10000 : mH = 10000
                                Dim cel As Avand.Designer.Cell
                                Select Case selGrab.Status
                                    Case GrabStatus.TopLeft
                                        For Each cel In cells
                                            If mX > cel.Member.Location.X Then mX = cel.Member.Location.X
                                            If mY > cel.Member.Location.Y Then mY = cel.Member.Location.Y
                                            If mW > cel.Member.Size.Width Then mW = cel.Member.Size.Width
                                            If mH > cel.Member.Size.Height Then mH = cel.Member.Size.Height
                                        Next : cel = Nothing
                                    Case GrabStatus.Top
                                        For Each cel In cells
                                            If mY > cel.Member.Location.Y Then mY = cel.Member.Location.Y
                                            If mH > cel.Member.Size.Height Then mH = cel.Member.Size.Height
                                        Next : cel = Nothing
                                    Case GrabStatus.TopRight
                                        For Each cel In cells
                                            If mX > cel.Member.Size.Width Then mX = cel.Member.Size.Width
                                            If mY > cel.Member.Location.Y Then mY = cel.Member.Location.Y
                                            If mH > cel.Member.Size.Height Then mH = cel.Member.Size.Height
                                        Next : cel = Nothing
                                    Case GrabStatus.Right
                                        For Each cel In cells
                                            If mX > cel.Member.Size.Width Then mX = cel.Member.Size.Width
                                        Next : cel = Nothing
                                    Case GrabStatus.BottomRight
                                        For Each cel In cells
                                            If mX > cel.Member.Size.Width Then mX = cel.Member.Size.Width
                                            If mY > cel.Member.Size.Height Then mY = cel.Member.Size.Height
                                        Next : cel = Nothing
                                    Case GrabStatus.Bottom
                                        For Each cel In cells
                                            If mY > cel.Member.Size.Height Then mY = cel.Member.Size.Height
                                        Next : cel = Nothing
                                    Case GrabStatus.BottomLeft
                                        For Each cel In cells
                                            If mX > cel.Member.Location.X Then mX = cel.Member.Location.X
                                            If mY > cel.Member.Size.Height Then mY = cel.Member.Size.Height
                                            If mW > cel.Member.Size.Width Then mW = cel.Member.Size.Width
                                        Next : cel = Nothing
                                    Case GrabStatus.Left
                                        For Each cel In cells
                                            If mX > cel.Member.Location.X Then mX = cel.Member.Location.X
                                            If mW > cel.Member.Size.Width Then mW = cel.Member.Size.Width
                                        Next : cel = Nothing
                                End Select
                                sizing = True
                                canc = False
                                GoTo sizingdiff
                            End If
                        Else
                            If canc Then
                                canc = False
                            Else
                                DrawSize()
                            End If
                            GoTo sizingdiff
                        End If
                    Case MousePos.ResizeTableWidth
                        If Not verting Then
                            If Math.Abs(e.X - oX) >= mTable.GridSize.Width Then
                                mX = mTable.GetMinWidth()
                                GoTo VertingDiff
                            End If
                        Else
                            If Not canc Then DrawVertic()
                            GoTo VertingDiff
                        End If
                    Case MousePos.ResizeSectionHeight
                        If Not horizing Then
                            If Math.Abs(e.Y - oY) >= mTable.GridSize.Height Then
                                mY = selSection.GetMinHeight()
                                GoTo HorizingDiff
                            End If
                        Else
                            If Not canc Then DrawHoriz()
                            GoTo HorizingDiff
                        End If
                End Select
            Else
                Me.Cursor = Cursors.Default
            End If
        ElseIf e.Button = MouseButtons.None Then
            Selecting(New Point(e.X, e.Y))
        End If


        Exit Sub
InsertingDiff:
        x = e.X : y = e.Y
        If x < offset.X Then x = offset.X
        Dim foo As Integer = GetTop(selSection) + offset.Y
        If y < foo Then y = foo
        dX = x - oX
        dY = y - oY
        DrawInsert(GetRect)
        inserting = True
        canc = False

        Exit Sub
RubbingDiff:
        x = e.X : y = e.Y
        If x < offset.X Then x = offset.X
        If x > offset.X + mTable.Width Then x = offset.X + mTable.Width
        If y < offset.Y Then y = offset.Y
        If y > offset.Y + totalHeight Then y = offset.Y + totalHeight
        dX = x - oX
        dY = y - oY
        DrawRubber(GetRect)

        Exit Sub
MovingDiff:
        dX = e.X - oX : If dX < -mX Then dX = -mX
        dY = e.Y - oY : If dY < -mY Then dY = -mY
        DrawLocation()
        Cursor = Cursors.Default
        moving = True
        canc = False

        Exit Sub
SizingDiff:
        dX = e.X - oX
        If dX < -mX Then dX = -mX
        If dX > mW Then dX = mW
        dY = e.Y - oY
        If dY < -mY Then dY = -mY
        If dY > mH Then dY = mH
        DrawSize()

        Exit Sub
VertingDiff:
        dX = NormX(e.X - offset.X)
        If dX < mX And mX + offset.X > 0 Then
            If dX < mX Then dX = mX
            DrawVertic()
            verting = True
            canc = False
        ElseIf e.X < 0 Or e.X > Width + CInt(Not vsb Is Nothing) * 16 Then
            If dX < mX Then dX = mX
            canc = True
        Else
            If dX < mX Then dX = mX
            DrawVertic()
            verting = True
            canc = False
        End If

        Exit Sub
HorizingDiff:
        Dim t As Integer = offset.Y + GetTop(selSection)
        dY = NormY(e.Y - t)
        If dY < mY And mY + t > 0 Then
            If dY < mY Then dY = mY
            DrawHoriz()
            horizing = True
            canc = False
        ElseIf e.Y < 0 Or e.Y > Height + CInt(Not hsb Is Nothing) * 16 Then
            If dY < mY Then dY = mY
            canc = True
        Else
            If dY < mY Then dY = mY
            DrawHoriz()
            horizing = True
            canc = False
        End If

    End Sub
    Protected Overrides Sub OnMouseUp(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseUp(e)
        If e.Button = MouseButtons.Left Then lMouse = False

        If (e.Button = MouseButtons.Left) Then
            Select Case pos
                Case MousePos.None
                Case MousePos.Table
                Case MousePos.Header
                Case MousePos.Section
                    If Me.Cursor Is Cursors.Cross Then
                        If inserting Then
                            inserting = False
                            Dim rec As Rectangle = GetRect()
                            If Not canc Then DrawInsert(rec)
                            If rec.Width = 0 Then rec.Width = 1
                            If rec.Height = 0 Then rec.Height = 1
                            Dim mem As Model10.Member = GetMember()
                            mem.Location = New Point(rec.Location.X - offset.X, rec.Location.Y - GetTop(mSelectedSection) - offset.Y)
                            mem.Size = rec.Size
                            Insert(mem)
                        End If
                    Else
                        If rubbing Then
                            rubbing = False
                            Dim r As Rectangle = GetRect()
                            '-----------------------------
                            If Not canc Then DrawRubber(r)
                            Dim y As Integer = offset.Y

                            If Not mTable.PageHeader Is Nothing Then ReverseMembers(mTable.PageHeader, r, y)
                            If Not mTable.GroupHeader Is Nothing Then ReverseMembers(mTable.GroupHeader, r, y)
                            If Not mTable.Detail Is Nothing Then ReverseMembers(mTable.Detail, r, y)
                            If Not mTable.GroupFooter Is Nothing Then ReverseMembers(mTable.GroupFooter, r, y)
                            If Not mTable.PageFooter Is Nothing Then ReverseMembers(mTable.PageFooter, r, y)
                            '-----------------------------
                            r = Nothing
                            Redraw()
                            If cells.Count <> 0 Then
                                mSelectedSection = Nothing '  
                                OnMembersSelected() : OnSelectedMembersChanged()
                            End If
                        End If
                    End If
                Case MousePos.Member
                    If moving Then
                        moving = False
                        If Not canc Then
                            DrawLocation()
                            Dim mX, mY As Integer
                            Dim changed, over As Boolean

                            Dim cel As Avand.Designer.Cell
                            For Each cel In cells
                                Dim x As Integer = NormX(cel.Member.Location.X + dX) - cel.Member.Location.X
                                Dim y As Integer = NormY(cel.Member.Location.Y + dY) - cel.Member.Location.Y
                                Dim l As New Point(cel.Member.Location.X + x, cel.Member.Location.Y + y)
                                If Not cel.Member.Location.Equals(l) Then
                                    cel.Member.Location = l

                                    If mX < cel.Member.Location.X + cel.Member.Size.Width Then mX = cel.Member.Location.X + cel.Member.Size.Width
                                    mY = cel.Member.Location.Y + cel.Member.Size.Height
                                    If mY > cel.Member.Section.Height Then
                                        totalHeight = totalHeight + mY - cel.Member.Section.Height
                                        cel.Member.Section.Height = mY
                                        'If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                                        over = True
                                    End If

                                    SetGrabs(cel)
                                    If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                                    changed = True
                                End If
                            Next : cel = Nothing

                            If changed Then
                                If mX > mTable.Width Then
                                    mTable.Width = mX
                                    Refine()
                                    over = True
                                End If
                                If over Then Resizing()
                                Redraw()
                                OnValueChanged()
                            End If
                        End If
                    End If
                Case MousePos.Grab
                    If sizing Then
                        sizing = False
                        If Not canc Then DrawSize()
                        Dim l, t, w, h As Integer
                        Select Case selGrab.Status
                            Case GrabStatus.TopLeft : l = dX : t = dY : w = -dX : h = -dY
                            Case GrabStatus.Top : t = dY : h = -dY
                            Case GrabStatus.TopRight : t = dY : w = dX : h = -dY
                            Case GrabStatus.Right : w = dX
                            Case GrabStatus.BottomRight : w = dX : h = dY
                            Case GrabStatus.Bottom : h = dY
                            Case GrabStatus.BottomLeft : l = dX : w = -dX : h = dY
                            Case GrabStatus.Left : l = dX : w = -dX
                        End Select
                        Dim mX, mY As Integer
                        Dim changed, over As Boolean

                        Dim cel As Avand.Designer.Cell
                        For Each cel In cells
                            Dim rc As New Rectangle(cel.Member.Location.X + l, cel.Member.Location.Y + t, cel.Member.Size.Width + w, cel.Member.Size.Height + h)
                            NormRect(rc)
                            If Not cel.Member.Location.Equals(rc.Location) Or Not cel.Member.Size.Equals(rc.Size) Then
                                cel.Member.Location = rc.Location
                                cel.Member.Size = rc.Size

                                If mX < cel.Member.Location.X + cel.Member.Size.Width Then mX = cel.Member.Location.X + cel.Member.Size.Width
                                mY = cel.Member.Location.Y + cel.Member.Size.Height
                                If mY > cel.Member.Section.Height Then
                                    totalHeight = totalHeight + mY - cel.Member.Section.Height
                                    cel.Member.Section.Height = mY
                                    If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                                    over = True
                                End If

                                SetGrabs(cel)
                                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                                changed = True
                            End If
                            rc = Nothing
                        Next : cel = Nothing

                        If changed Then
                            If mX > mTable.Width Then
                                mTable.Width = mX
                                Refine()
                                over = True
                            End If
                            If over Then Resizing()
                            Redraw()
                            OnValueChanged()
                        End If
                    End If
                Case MousePos.ResizeTableWidth
                    If Cursor Is Cursors.SizeWE Then
                        If verting Then
                            verting = False
                            If Not canc Then DrawVertic()
                            If dX <> mTable.Width Then
                                mTable.Width = dX
                                Refine()
                                Resizing()
                                Redraw()
                                OnValueChanged()
                            End If
                        End If
                    End If
                Case MousePos.ResizeSectionHeight
                    If Cursor Is Cursors.SizeNS Then
                        If horizing Then
                            horizing = False
                            If Not canc Then DrawHoriz()
                            If dY <> selSection.Height Then
                                totalHeight = totalHeight + dY - selSection.Height
                                selSection.Height = dY
                                If Not draws.Contains(selSection) Then draws.Add(selSection)
                                Resizing()
                                Redraw()
                                OnValueChanged()
                            End If
                        End If
                    End If
            End Select
            If (MouseButtons And MouseButtons.Right) <> MouseButtons.Right Then
                Selecting(New Point(e.X, e.Y))
            End If
        End If
    End Sub

    Protected Overrides Sub OnMouseWheel(ByVal e As System.Windows.Forms.MouseEventArgs)
        MyBase.OnMouseWheel(e)

        If MouseButtons = MouseButtons.None Then
            If Not vsb Is Nothing Then
                Dim i As Integer = vsb.Value - e.Delta
                If i < vsb.Minimum Then i = vsb.Minimum
                If i > vsb.Maximum - vsb.LargeChange + 1 Then i = vsb.Maximum - vsb.LargeChange + 1
                vsb.Value = i
                Selecting(New Point(e.X, e.Y))
            End If
        End If
    End Sub
    Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
        MyBase.OnPaint(e)

        Try
            If mTable Is Nothing Then
                totalHeight = 0
            Else
                Dim y As Integer = offset.Y

                If Not mTable.PageHeader Is Nothing Then DrawSection(e.Graphics, mTable.PageHeader, GetSystemResource("PageHeader"), y)
                If Not mTable.GroupHeader Is Nothing Then DrawSection(e.Graphics, mTable.GroupHeader, GetSystemResource("GroupHeader"), y)
                If Not mTable.Detail Is Nothing Then DrawSection(e.Graphics, mTable.Detail, GetSystemResource("Detail"), y)
                If Not mTable.GroupFooter Is Nothing Then DrawSection(e.Graphics, mTable.GroupFooter, GetSystemResource("GroupFooter"), y)
                If Not mTable.PageFooter Is Nothing Then DrawSection(e.Graphics, mTable.PageFooter, GetSystemResource("PageFooter"), y)

                totalHeight = y - offset.Y
            End If
            If totalHeight = 0 Then
                e.Graphics.DrawString("There is nothing to show...", Font, SystemBrushes.HighlightText, 8, 8)
            End If
            If rubbing Then
                If Not canc Then DrawRubber(GetRect)
            End If
            If moving Then
                If Not canc Then DrawLocation()
            End If
            If verting Then
                If Not canc Then DrawVertic()
            End If
            If horizing Then
                If Not canc Then DrawHoriz()
            End If
            If sizing Then
                If Not canc Then DrawSize()
            End If
            If inserting Then
                If Not canc Then DrawInsert(GetRect)
            End If
            canc = True
        Catch ex As Exception
            Trace.WriteLine(ex.Message, "Designer/OnPaint")
        End Try
    End Sub
    Protected Overrides Sub OnResize(ByVal e As System.EventArgs)
        MyBase.OnResize(e)

        Resizing()
    End Sub
    Protected Overrides Sub OnValidating(ByVal e As System.ComponentModel.CancelEventArgs)
        MyBase.OnValidating(e)

        If rubbing Or moving Or verting Or horizing Or sizing Or inserting Then e.Cancel = True
    End Sub

    Public Function GetMembers() As Model10.Member()
        Dim mems() As Model10.Member

        Dim i As Integer
        For i = 0 To cells.Count - 1
            ReDim Preserve mems(i)
            mems(i) = CType(cells(i), Avand.Designer.Cell).Member
        Next

        Return mems
    End Function

    Public ReadOnly Property SelectedMember() As Model10.Member
        Get
            Return mSelectedMember
        End Get
    End Property
    Public ReadOnly Property SelectedMembersCount() As Integer
        Get
            Return cells.Count
        End Get
    End Property
    Public ReadOnly Property SelectedSection() As Model10.Section
        Get
            Return mSelectedSection
        End Get
    End Property
    Public Property Table() As Model10.Table
        Get
            Return mTable
        End Get
        Set(ByVal Value As Model10.Table)
            mTable = Value
        End Set
    End Property
    Public Property User() As Model10.User
        Get
            Return mUser
        End Get
        Set(ByVal Value As Model10.User)
            mUser = Value
        End Set
    End Property

    Public Sub AddPageHeader()
        mTable.PageHeader = New Model10.Pager

        Dim ext As Boolean = cells.Count > 0
        Dim sec As Model10.Section = mSelectedSection
        ClearCells()
        mSelectedSection = mTable.PageHeader
        mSelectedMember = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If ext Or sec Is Nothing Then OnSectionSelected()
        OnSelectedSectionChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub AddGroupHeader()
        mTable.GroupHeader = New Model10.Grouper

        Dim ext As Boolean = cells.Count > 0
        Dim sec As Model10.Section = mSelectedSection
        ClearCells()
        mSelectedSection = mTable.GroupHeader
        mSelectedMember = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If ext Or sec Is Nothing Then OnSectionSelected()
        OnSelectedSectionChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub AddDetail()
        mTable.Detail = New Model10.Detail

        Dim ext As Boolean = cells.Count > 0
        Dim sec As Model10.Section = mSelectedSection
        ClearCells()
        mSelectedSection = mTable.Detail
        mSelectedMember = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If ext Or sec Is Nothing Then OnSectionSelected()
        OnSelectedSectionChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub AddGroupFooter()
        mTable.GroupFooter = New Model10.Grouper

        Dim ext As Boolean = cells.Count > 0
        Dim sec As Model10.Section = mSelectedSection
        ClearCells()
        mSelectedSection = mTable.GroupFooter
        mSelectedMember = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If ext Or sec Is Nothing Then OnSectionSelected()
        OnSelectedSectionChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub AddPageFooter()
        mTable.PageFooter = New Model10.Pager

        Dim ext As Boolean = cells.Count > 0
        Dim sec As Model10.Section = mSelectedSection
        ClearCells()
        mSelectedSection = mTable.PageFooter
        mSelectedMember = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If ext Or sec Is Nothing Then OnSectionSelected()
        OnSelectedSectionChanged()
        OnValueChanged()
        sec = Nothing
    End Sub

    Public Sub RemovePageHeader()
        Dim sec As Model10.Section = mSelectedSection

        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            If cel.Member.Section Is mTable.PageHeader Then
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                cells.Remove(cel)
            End If
            cel = Nothing
        Next

        If mSelectedSection Is mTable.PageHeader Then mSelectedSection = Nothing
        If Not mSelectedMember Is Nothing Then If mSelectedMember.Section Is mTable.PageHeader Then mSelectedMember = Nothing
        mTable.PageHeader = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub RemoveGroupHeader()
        Dim sec As Model10.Section = mSelectedSection

        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            If cel.Member.Section Is mTable.GroupHeader Then
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                cells.Remove(cel)
            End If
            cel = Nothing
        Next

        If mSelectedSection Is mTable.GroupHeader Then mSelectedSection = Nothing
        If Not mSelectedMember Is Nothing Then If mSelectedMember.Section Is mTable.GroupHeader Then mSelectedMember = Nothing
        mTable.GroupHeader = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub RemoveDetail()
        Dim sec As Model10.Section = mSelectedSection

        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            If cel.Member.Section Is mTable.Detail Then
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                cells.Remove(cel)
            End If
            cel = Nothing
        Next

        If mSelectedSection Is mTable.Detail Then mSelectedSection = Nothing
        If Not mSelectedMember Is Nothing Then If mSelectedMember.Section Is mTable.Detail Then mSelectedMember = Nothing
        mTable.Detail = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub RemoveGroupFooter()
        Dim sec As Model10.Section = mSelectedSection

        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            If cel.Member.Section Is mTable.GroupFooter Then
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                cells.Remove(cel)
            End If
            cel = Nothing
        Next

        If mSelectedSection Is mTable.GroupFooter Then mSelectedSection = Nothing
        If Not mSelectedMember Is Nothing Then If mSelectedMember.Section Is mTable.GroupFooter Then mSelectedMember = Nothing
        mTable.GroupFooter = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub RemovePageFooter()
        Dim sec As Model10.Section = mSelectedSection

        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            If cel.Member.Section Is mTable.PageFooter Then
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                cells.Remove(cel)
            End If
            cel = Nothing
        Next

        If mSelectedSection Is mTable.PageFooter Then mSelectedSection = Nothing
        If Not mSelectedMember Is Nothing Then If mSelectedMember.Section Is mTable.PageFooter Then mSelectedMember = Nothing
        mTable.PageFooter = Nothing
        SetTotalHeight()
        Resizing()
        Redraw()
        Selecting(PointToClient(MousePosition))
        If Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
        OnValueChanged()
        sec = Nothing
    End Sub

    Public Sub AlignToGrid()
        Dim mX, mY As Integer
        Dim changed, over As Boolean

        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            Dim x As Integer = NormX(cel.Member.Location.X) - cel.Member.Location.X
            Dim y As Integer = NormY(cel.Member.Location.Y) - cel.Member.Location.Y
            Dim l As New Point(cel.Member.Location.X + x, cel.Member.Location.Y + y)
            If Not cel.Member.Location.Equals(l) Then
                cel.Member.Location = l

                If mX < cel.Member.Location.X + cel.Member.Size.Width Then mX = cel.Member.Location.X + cel.Member.Size.Width
                mY = cel.Member.Location.Y + cel.Member.Size.Height
                If mY > cel.Member.Section.Height Then
                    totalHeight = totalHeight + mY - cel.Member.Section.Height
                    cel.Member.Section.Height = mY
                    over = True
                End If

                SetGrabs(cel)
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                changed = True
            End If
        Next : cel = Nothing

        If changed Then
            If mX > mTable.Width Then
                mTable.Width = mX
                Refine()
                over = True
            End If
            If over Then Resizing()
            Redraw()
            OnValueChanged()
        End If
    End Sub
    Public Sub DeleteSelectedMembers()
        Dim sec As Model10.Section = mSelectedSection

        Dim i As Integer
        For i = cells.Count - 1 To 0 Step -1
            Dim cel As Avand.Designer.Cell = CType(cells(i), Avand.Designer.Cell)
            cells.Remove(cel)
            If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
            cel.Member.Section.Members.Remove(cel.Member)
            cel = Nothing
        Next

        'mSelectedSection = Nothing   
        mSelectedMember = Nothing
        Redraw()
        Selecting(PointToClient(MousePosition))
        If Not sec Is Nothing Then OnTableSelected() : OnSelectedTableChanged()
        OnValueChanged()
        sec = Nothing
    End Sub
    Public Sub Display()
        Refine()
        Redraw()
        OnValueChanged()
    End Sub
    Public Sub Fill()
        SetTotalHeight()
        Refine()
        Resizing()
        Redraw()
    End Sub
    Public Sub InsertMember(ByVal memberType As Type, Optional ByVal dataField As String = "", Optional ByVal forToolBar As Boolean = False)
        insertingMemberType = memberType
        insertingDataField = dataField
        If mSelectedSection Is Nothing Or forToolBar Then
            Me.Cursor = Cursors.Cross
        Else
            Dim mem As Model10.Member = GetMember()
            If Not mSelectedMember Is Nothing Then
                Dim l As Point = mSelectedMember.Location
                l.Offset(mTable.GridSize.Width, mTable.GridSize.Height)
                mem.Location = l
            End If
            Insert(mem)
            mem = Nothing
        End If
    End Sub
    Public Sub PasteMembers()
        Dim p1 As New Point
        If cells.Count > 0 Then
            p1.X = 10000 : p1.Y = 10000

            Dim c As Cell
            For Each c In cells
                If c.Member.Location.X < p1.X Then p1.X = c.Member.Location.X
                If c.Member.Location.Y < p1.Y Then p1.Y = c.Member.Location.Y
            Next : c = Nothing

            p1.X += mTable.GridSize.Width
            p1.Y += mTable.GridSize.Height
        End If

        Dim emt As Boolean = cells.Count = 0
        ClearCells()

        Dim p2 As New Point(10000, 10000)
        Dim mem As Model10.Member
        For Each mem In CType(Clipboard.GetDataObject.GetData(GetType(Model10.Member())), Model10.Member())
            If mem.Location.X < p2.X Then p2.X = mem.Location.X
            If mem.Location.Y < p2.Y Then p2.Y = mem.Location.Y
        Next : mem = Nothing
        Dim p3 As New Point(p1.X - p2.X, p1.Y - p2.Y)
        For Each mem In CType(Clipboard.GetDataObject.GetData(GetType(Model10.Member())), Model10.Member())
            mem.Location = New Drawing.Point(mem.Location.X + p3.X, mem.Location.Y + p3.Y)

            Dim m As Model10.Member
            For Each m In SelectedSection.Members
                If m.Name.Trim.ToLower = mem.Name.Trim.ToLower Then
                    Select Case mem.GetType.ToString
                        Case "Model10.Line"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Line"))
                        Case "Model10.Frame"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Frame"))
                        Case "Model10.Label"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Label"))
                        Case "Model10.TextBox"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.TextBox"))
                        Case "Model10.PictureBox"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.PictureBox"))
                        Case "Model10.CheckBox"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.CheckBox"))
                        Case "Model10.Select"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Select"))
                        Case "Model10.Edit"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Edit"))
                        Case "Model10.Update"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Update"))
                        Case "Model10.Cancel"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Cancel"))
                        Case "Model10.Delete"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Delete"))
                        Case "Model10.Browse"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Browse"))
                        Case "Model10.Computer"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Computer"))
                        Case "Model10.Expand"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Expand"))
                        Case "Model10.Next"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Next"))
                        Case "Model10.MultiSelect"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.MultiSelect"))
                        Case "Model10.Drilldown"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Drilldown"))
                        Case "Model10.Lookup"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Button.Lookup"))
                        Case "Model10.Sub"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Sub"))
                        Case "Model10.Chart"
                            mem.Name = SelectedSection.GetNewMemberName([Global].GetSystemResource("Member.Chart"))
                    End Select
                    Exit For
                End If
            Next : m = Nothing

            SelectedSection.Members.Add(mem, 0, 0)
            OneCell(SelectedSection, mem)
        Next : mem = Nothing

        ResetMembers()

        If emt Then OnMembersSelected()
        OnSelectedMembersChanged()
        OnValueChanged()
    End Sub
    Public Sub Reset()
        If cells.Count = 0 Then
            If mSelectedSection Is Nothing Then
                ResetTable()
            Else
                ResetSection()
            End If
        Else
            ResetMembers()
        End If
    End Sub
    Public Sub ResetTable()
        Refine()
        Resizing()
        Redraw()
        OnValueChanged()
    End Sub
    Public Sub ResetSection()
        If Not mSelectedSection Is Nothing Then
            DrawMembers(mSelectedSection)
            If Not draws.Contains(mSelectedSection) Then draws.Add(mSelectedSection)
        End If
        SetTotalHeight()
        Resizing()
        Redraw()
        OnValueChanged()
    End Sub
    Public Sub ResetMembers()
        Dim mX, mY As Integer
        Dim over As Boolean

        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            If mX < cel.Member.Location.X + cel.Member.Size.Width Then mX = cel.Member.Location.X + cel.Member.Size.Width
            mY = cel.Member.Location.Y + cel.Member.Size.Height
            If mY > cel.Member.Section.Height Then
                totalHeight = totalHeight + mY - cel.Member.Section.Height
                cel.Member.Section.Height = mY
                over = True
            End If

            SetGrabs(cel)
            If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
        Next : cel = Nothing

        If mX > mTable.Width Then
            mTable.Width = mX
            Refine()
            over = True
        End If
        If over Then Resizing()
        Redraw()
        OnValueChanged()
    End Sub
    Public Sub SelectAll()
        Dim ext As Boolean = cells.Count > 0

        ClearCells()
        Dim mem As Model10.Member
        If mSelectedSection Is Nothing Then
            If Not mTable Is Nothing Then
                If Not mTable.PageHeader Is Nothing Then
                    For Each mem In mTable.PageHeader.Members
                        OneCell(mTable.PageHeader, mem)
                    Next : mem = Nothing
                    If Not draws.Contains(mTable.PageHeader) Then draws.Add(mTable.PageHeader)
                End If
                If Not mTable.GroupHeader Is Nothing Then
                    For Each mem In mTable.GroupHeader.Members
                        OneCell(mTable.GroupHeader, mem)
                    Next : mem = Nothing
                    If Not draws.Contains(mTable.GroupHeader) Then draws.Add(mTable.GroupHeader)
                End If
                If Not mTable.Detail Is Nothing Then
                    For Each mem In mTable.Detail.Members
                        OneCell(mTable.Detail, mem)
                    Next : mem = Nothing
                    If Not draws.Contains(mTable.Detail) Then draws.Add(mTable.Detail)
                End If
                If Not mTable.GroupFooter Is Nothing Then
                    For Each mem In mTable.GroupFooter.Members
                        OneCell(mTable.GroupFooter, mem)
                    Next : mem = Nothing
                    If Not draws.Contains(mTable.GroupFooter) Then draws.Add(mTable.GroupFooter)
                End If
                If Not mTable.PageFooter Is Nothing Then
                    For Each mem In mTable.PageFooter.Members
                        OneCell(mTable.PageFooter, mem)
                    Next : mem = Nothing
                    If Not draws.Contains(mTable.PageFooter) Then draws.Add(mTable.PageFooter)
                End If
            End If
        Else
            For Each mem In mSelectedSection.Members
                OneCell(mSelectedSection, mem)
            Next : mem = Nothing
            If Not draws.Contains(mSelectedSection) Then draws.Add(mSelectedSection)
        End If
        Redraw()
        If cells.Count > 0 Then OnMembersSelected() : OnSelectedMembersChanged()
    End Sub
    Public Sub SizeToGrid()
        Dim mX, mY As Integer
        Dim changed, over As Boolean

        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            Dim s As New Size(NormX(cel.Member.Size.Width), NormY(cel.Member.Size.Height))
            If Not cel.Member.Size.Equals(s) Then
                cel.Member.Size = s

                If mX < cel.Member.Location.X + cel.Member.Size.Width Then mX = cel.Member.Location.X + cel.Member.Size.Width
                mY = cel.Member.Location.Y + cel.Member.Size.Height
                If mY > cel.Member.Section.Height Then
                    totalHeight = totalHeight + mY - cel.Member.Section.Height
                    cel.Member.Section.Height = mY
                    over = True
                End If

                SetGrabs(cel)
                If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
                changed = True
            End If
        Next : cel = Nothing

        If changed Then
            If mX > mTable.Width Then
                mTable.Width = mX
                Refine()
                over = True
            End If
            If over Then Resizing()
            Redraw()
            OnValueChanged()
        End If
    End Sub
    Public Sub ToBack()
        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            Dim mems As Model10.MemberCollection = cel.Member.Section.Members
            mems.Remove(cel.Member)
            mems.Insert(0, cel.Member)
            mems = Nothing
            If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
            cel = Nothing
        Next : cel = Nothing
        Redraw()
        OnValueChanged()
    End Sub
    Public Sub ToFront()
        Dim cel As Avand.Designer.Cell
        For Each cel In cells
            Dim mems As Model10.MemberCollection = cel.Member.Section.Members
            mems.Remove(cel.Member)
            mems.Insert(mems.Count, cel.Member)
            mems = Nothing
            If Not draws.Contains(cel.Member.Section) Then draws.Add(cel.Member.Section)
            cel = Nothing
        Next : cel = Nothing
        Redraw()
        OnValueChanged()
    End Sub
End Class
