Option Strict On
Option Explicit On 

Imports System.Runtime.Serialization
Imports Microsoft.Office.Interop

Public Enum ActionMode As Byte
    Record = 0
    View = 1
    Table = 2
    [Default] = 3
    [Function] = 4
End Enum
Public Enum AnchorMode As Byte
    None = 0
    Top = 1
    Bottom = 2
    TopAndBottom = 3
End Enum
Public Enum ButtonStyle As Byte
    Normal = 0
    Flat = 1
    Link = 2
    Image = 3
    PlusMines = 4
End Enum
Public Enum CheckStyle As Byte
    Normal = 0
    Flat = 1
End Enum
Public Enum DesignerMode As Byte
    PrintableView = 0
    FormView = 1
End Enum
Public Enum KeepMode As Byte
    No = 0
    WholeGroup = 1
    WithFirstDetail = 2
End Enum
Public Enum NewPageMode As Byte
    None = 0
    Before = 1
    After = 2
    BeforeAndAfter = 3
End Enum
Public Enum PathType As Byte
    Complete = 0
    DirectoryName = 1
    Extension = 2
    FileName = 3
    FileNameWithoutExtension = 4
    FullPath = 5
    PathRoot = 6
End Enum
Public Enum RowStatus As Byte
    First = 0
    Last = 1
    AddNew = 2
End Enum
Public Enum SizeMode As Byte
    Clip = 0
    Stretch = 1
    Zoom = 2
End Enum
Public Enum StartMode As Byte
    Undefined = 0
    Print = 1
    Preview = 2
    Form = 3
    DataSheet = 4
End Enum
Public Enum ValueType As Byte
    None = 0
    DateTime = 1
    Title = 2
    Path = 3
    PageCount = 4
    PageNumber = 5
    RecordCount = 6
    RecordNumber = 7
    UserId = 8
    UserFullName = 9
    UserName = 10
End Enum

Public Interface IDesigner
    Property DisplayGrid() As Boolean
    Property DisplayRulers() As Boolean
    Property GridSize() As Drawing.Size
    Property Locked() As Boolean
    Property SnapToGrid() As Boolean
End Interface

<Serializable()> Public MustInherit Class Member
    Implements ISerializable

    Private mAnchor As Model10.AnchorMode
    Private mBackColor As Drawing.Color
    Private mLocation As Drawing.Point
    Private mName As String
    Private mPrintable As Boolean
    Private mSection As Model10.Section
    Private mSize As Drawing.Size
    Private mVisible As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mAnchor = CType(info.GetByte("Anchor"), Model10.AnchorMode)
        mBackColor = CType(info.GetValue("BackColor", GetType(Drawing.Color)), Drawing.Color)
        mLocation = CType(info.GetValue("Location", GetType(Drawing.Point)), Drawing.Point)
        mName = info.GetString("Name")
        mPrintable = info.GetBoolean("Printable")
        mSection = CType(info.GetValue("Section", GetType(Model10.Section)), Model10.Section)
        mSize = CType(info.GetValue("Size", GetType(Drawing.Size)), Drawing.Size)
        mVisible = info.GetBoolean("Visible")
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("Anchor", mAnchor)
        info.AddValue("BackColor", mBackColor)
        info.AddValue("Location", mLocation)
        info.AddValue("Name", mName)
        info.AddValue("Printable", mPrintable)
        info.AddValue("Section", mSection)
        info.AddValue("Size", mSize)
        info.AddValue("Visible", mVisible)
    End Sub

    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New()
        MyBase.New()

        mAnchor = AnchorMode.Top
        mBackColor = Drawing.Color.Transparent
        mLocation = New Drawing.Point
        mName = ""
        mPrintable = True
        mSection = Nothing
        mSize = New Drawing.Size(100, 16)
        mVisible = True
    End Sub
    Public Sub New(ByVal member As Model10.Member)
        MyBase.New()

        mAnchor = member.Anchor
        mBackColor = member.BackColor
        mLocation = member.Location
        mName = member.Name
        mPrintable = member.Printable
        mSection = Nothing
        mSize = member.Size
        mVisible = member.Visible
    End Sub
    Public Overridable Function Clone() As Model10.Member
    End Function

    Public Property Anchor() As Model10.AnchorMode
        Get
            Return mAnchor
        End Get
        Set(ByVal Value As Model10.AnchorMode)
            mAnchor = Value
        End Set
    End Property
    Public Property BackColor() As Drawing.Color
        Get
            Return mBackColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mBackColor = Value
        End Set
    End Property
    Public Property Location() As Drawing.Point
        Get
            Return mLocation
        End Get
        Set(ByVal Value As Drawing.Point)
            mLocation = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            If Not mSection Is Nothing Then
                Dim mem As Model10.Member
                For Each mem In mSection.Members
                    If mem.Name.ToLower = Value.ToLower Then
                        If Not mem Is Me Then
                            Throw New System.Data.DuplicateNameException("A Member named '" & Value & "' already belongs to this Section.")
                        End If
                    End If
                Next : mem = Nothing
            End If
            mName = Value
        End Set
    End Property
    Public Property Printable() As Boolean
        Get
            Return mPrintable
        End Get
        Set(ByVal Value As Boolean)
            mPrintable = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Section() As Model10.Section
        Get
            Return mSection
        End Get
        Set(ByVal Value As Model10.Section)
            mSection = Value
        End Set
    End Property
    Public Property Size() As Drawing.Size
        Get
            Return mSize
        End Get
        Set(ByVal Value As Drawing.Size)
            mSize = Value
        End Set
    End Property
    Public Property Visible() As Boolean
        Get
            Return mVisible
        End Get
        Set(ByVal Value As Boolean)
            mVisible = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Line
    Inherits Model10.Member

    Private mDashStyle As Drawing.Drawing2D.DashStyle

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mDashStyle = CType(info.GetInt32("DashStyle"), Drawing.Drawing2D.DashStyle)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("DashStyle", mDashStyle)
    End Sub

    Public Sub Draw(ByVal grp As Drawing.Graphics, ByVal offsetx As Integer, ByVal offsetY As Integer)
        If MyBase.Size.Width > MyBase.Size.Height Then
            Dim p As New Drawing.Pen(MyBase.BackColor, MyBase.Size.Height)
            p.DashStyle = mDashStyle
            Dim y As Integer = CInt(offsetY + MyBase.Location.Y + MyBase.Size.Height / 2)
            Dim x1 As Integer = offsetx + MyBase.Location.X
            Dim x2 As Integer = offsetx + MyBase.Location.X + MyBase.Size.Width
            grp.DrawLine(p, x1, y, x2, y)
            p.Dispose() : p = Nothing
        Else
            Dim p As New Drawing.Pen(MyBase.BackColor, MyBase.Size.Width)
            p.DashStyle = mDashStyle
            Dim x As Integer = CInt(offsetx + MyBase.Location.X + MyBase.Size.Width / 2)
            Dim y1 As Integer = offsetY + MyBase.Location.Y
            Dim y2 As Integer = offsetY + MyBase.Location.Y + MyBase.Size.Height
            grp.DrawLine(p, x, y1, x, y2)
            p.Dispose() : p = Nothing
        End If
    End Sub
    Public Sub DrawF(ByVal grp As Drawing.Graphics, ByVal offsetx As Integer, ByVal offsetY As Integer)
        If MyBase.Size.Width > MyBase.Size.Height Then
            Dim p As New Drawing.Pen(MyBase.BackColor, MyBase.Size.Height)
            p.DashStyle = mDashStyle
            Dim y As Single = CSng(offsetY + MyBase.Location.Y + MyBase.Size.Height / 2)
            Dim x1 As Single = offsetx + MyBase.Location.X
            Dim x2 As Single = offsetx + MyBase.Location.X + MyBase.Size.Width
            grp.DrawLine(p, x1, y, x2, y)
            p.Dispose() : p = Nothing
        Else
            Dim p As New Drawing.Pen(MyBase.BackColor, MyBase.Size.Width)
            p.DashStyle = mDashStyle
            Dim x As Single = CSng(offsetx + MyBase.Location.X + MyBase.Size.Width / 2)
            Dim y1 As Single = offsetY + MyBase.Location.Y
            Dim y2 As Single = offsetY + MyBase.Location.Y + MyBase.Size.Height
            grp.DrawLine(p, x, y1, x, y2)
            p.Dispose() : p = Nothing
        End If
    End Sub

    Public Sub New()
        MyBase.New()
        MyBase.BackColor = System.Drawing.Color.Black
        MyBase.Size = New Drawing.Size(100, 1)

        mDashStyle = Drawing.Drawing2D.DashStyle.Solid
    End Sub
    Public Sub New(ByVal line As Model10.Line)
        MyBase.New(line)

        mDashStyle = line.DashStyle
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Line(Me)
    End Function

    Public Property DashStyle() As Drawing.Drawing2D.DashStyle
        Get
            Return mDashStyle
        End Get
        Set(ByVal Value As Drawing.Drawing2D.DashStyle)
            mDashStyle = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Frame
    Inherits Model10.Member

    Private mBorder3DStyle As Windows.Forms.Border3DStyle
    Private mBorderColor As Drawing.Color
    Private mBorderWidth As Single
    Private mCloseAtPageBreak As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mBorder3DStyle = CType(info.GetInt32("Border3DStyle"), Windows.Forms.Border3DStyle)
        mBorderColor = CType(info.GetValue("BorderColor", GetType(Drawing.Color)), Drawing.Color)
        mBorderWidth = info.GetSingle("BorderWidth")
        mCloseAtPageBreak = info.GetBoolean("CloseAtPageBreak")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("Border3DStyle", mBorder3DStyle)
        info.AddValue("BorderColor", mBorderColor)
        info.AddValue("BorderWidth", mBorderWidth)
        info.AddValue("CloseAtPageBreak", mCloseAtPageBreak)
    End Sub

    Public Sub Draw(ByVal grp As Drawing.Graphics, ByVal text As String, ByVal font As Drawing.Font, ByVal offsetx As Integer, ByVal offsetY As Integer)
        Dim k As Single
        Select Case mBorder3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = mBorderWidth
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select

        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetx + k / 2, MyBase.Location.Y + offsetY + k / 2, MyBase.Size.Width - k, MyBase.Size.Height - k)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, Drawing.Rectangle.Truncate(r))

        If mBorder3DStyle = Windows.Forms.Border3DStyle.Flat Then
            If k > 0 Then
                Dim p As New Drawing.Pen(mBorderColor, k)
                grp.DrawRectangle(p, Drawing.Rectangle.Truncate(New Drawing.RectangleF(r.X, r.Y, r.Width, r.Height)))
                p.Dispose() : p = Nothing
            End If
        Else
            Windows.Forms.ControlPaint.DrawBorder3D(grp, New Drawing.Rectangle(MyBase.Location.X + offsetx, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height), mBorder3DStyle)
        End If

        r.Inflate(-k / 2, -k / 2)
        b.Color = Drawing.SystemColors.ControlText
        grp.DrawString(text, font, b, r)

        b.Dispose() : b = Nothing
    End Sub
    Public Sub Draw(ByVal grp As Drawing.Graphics, ByVal offsetx As Integer, ByVal offsetY As Integer)
        Dim k As Single
        Select Case mBorder3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = mBorderWidth
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select

        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetx + k / 2, MyBase.Location.Y + offsetY + k / 2, MyBase.Size.Width - k, MyBase.Size.Height - k)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, Drawing.Rectangle.Truncate(r))
        b.Dispose() : b = Nothing

        If mBorder3DStyle = Windows.Forms.Border3DStyle.Flat Then
            If k > 0 Then
                Dim p As New Drawing.Pen(mBorderColor, k)
                grp.DrawRectangle(p, Drawing.Rectangle.Truncate(New Drawing.RectangleF(r.X, r.Y, r.Width, r.Height)))
                p.Dispose() : p = Nothing
            End If
        Else
            Windows.Forms.ControlPaint.DrawBorder3D(grp, New Drawing.Rectangle(MyBase.Location.X + offsetx, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height), mBorder3DStyle)
        End If
    End Sub
    Public Sub DrawF(ByVal grp As Drawing.Graphics, ByVal offsetx As Integer, ByVal offsetY As Integer)
        Dim k As Single = mBorderWidth

        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetx + k / 2, MyBase.Location.Y + offsetY + k / 2, MyBase.Size.Width - k, MyBase.Size.Height - k)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, r)
        b.Dispose() : b = Nothing

        If k > 0 Then
            Dim p As New Drawing.Pen(mBorderColor, k)
            grp.DrawRectangle(p, r.X, r.Y, r.Width, r.Height)
            p.Dispose() : p = Nothing
        End If
    End Sub

    Public Sub New()
        MyBase.New()
        MyBase.Size = New Drawing.Size(100, 32)

        mBorder3DStyle = Windows.Forms.Border3DStyle.Flat
        mBorderColor = Drawing.Color.Black
        mBorderWidth = 1
        mCloseAtPageBreak = True
    End Sub
    Public Sub New(ByVal frame As Model10.Frame)
        MyBase.New(frame)

        mBorder3DStyle = frame.Border3DStyle
        mBorderColor = frame.BorderColor
        mBorderWidth = frame.BorderWidth
        mCloseAtPageBreak = frame.CloseAtPageBreak
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Frame(Me)
    End Function

    Public Property Border3DStyle() As Windows.Forms.Border3DStyle
        Get
            Return mBorder3DStyle
        End Get
        Set(ByVal Value As Windows.Forms.Border3DStyle)
            mBorder3DStyle = Value
        End Set
    End Property
    Public Property BorderColor() As Drawing.Color
        Get
            Return mBorderColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mBorderColor = Value
        End Set
    End Property
    Public Property BorderWidth() As Single
        Get
            Return mBorderWidth
        End Get
        Set(ByVal Value As Single)
            mBorderWidth = Value
        End Set
    End Property
    Public Property CloseAtPageBreak() As Boolean
        Get
            Return mCloseAtPageBreak
        End Get
        Set(ByVal Value As Boolean)
            mCloseAtPageBreak = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Label
    Inherits Model10.Frame
    Implements IComparable

    <NonSerialized()> Private mDifference As Integer

    Private mAutoSize As Boolean
    Private mDataField As String
    Private mExpression As String
    Private mFilter As String
    Private mFont As Drawing.Font
    Private mForeColor As Drawing.Color
    Private mFormat As String
    Private mMaxNumberOfLines As Integer
    Private mNavigateUrl As String
    Private mParentRelation As Model10.Relation
    Private mTabIndex As Integer
    Private mTabStop As Boolean
    Private mTarget As String
    Private mText As String
    Private mToolTip As String
    Private mValue As Model10.ValueType

    Private mAlignment As Windows.Forms.HorizontalAlignment
    Private mHotkeyPrefix As Drawing.Text.HotkeyPrefix
    Private mNoWrap As Boolean
    Private mRightToLeft As Windows.Forms.RightToLeft
    Private mTrimming As Drawing.StringTrimming

    Private mBottomMargin As Integer
    Private mLeftMargin As Integer
    Private mRightMargin As Integer
    Private mTopMargin As Integer

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mAutoSize = info.GetBoolean("AutoSize")
        mDataField = info.GetString("DataField")
        mExpression = info.GetString("Expression")
        mFilter = info.GetString("Filter")
        mFont = CType(info.GetValue("Font", GetType(Drawing.Font)), Drawing.Font)
        mForeColor = CType(info.GetValue("ForeColor", GetType(Drawing.Color)), Drawing.Color)
        mFormat = info.GetString("Format")
        mMaxNumberOfLines = info.GetInt32("MaxNumberOfLines")
        mNavigateUrl = info.GetString("NavigateUrl")
        mParentRelation = CType(info.GetValue("ParentRelation", GetType(Model10.Relation)), Model10.Relation)
        mTabIndex = info.GetInt32("TabIndex")
        mTabStop = info.GetBoolean("TabStop")
        mTarget = info.GetString("Target")
        mText = info.GetString("Text")
        mToolTip = info.GetString("ToolTip")
        mValue = CType(info.GetByte("Value"), Model10.ValueType)

        mAlignment = CType(info.GetInt32("Alignment"), Windows.Forms.HorizontalAlignment)
        mHotkeyPrefix = CType(info.GetInt32("HotkeyPrefix"), Drawing.Text.HotkeyPrefix)
        mNoWrap = info.GetBoolean("NoWrap")
        mRightToLeft = CType(info.GetInt32("RightToLeft"), Windows.Forms.RightToLeft)
        mTrimming = CType(info.GetInt32("Trimming"), Drawing.StringTrimming)

        mBottomMargin = info.GetInt32("BottomMargin")
        mLeftMargin = info.GetInt32("LeftMargin")
        mRightMargin = info.GetInt32("RightMargin")
        mTopMargin = info.GetInt32("TopMargin")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("AutoSize", mAutoSize)
        info.AddValue("DataField", mDataField)
        info.AddValue("Expression", mExpression)
        info.AddValue("Filter", mFilter)
        info.AddValue("Font", mFont)
        info.AddValue("ForeColor", mForeColor)
        info.AddValue("Format", mFormat)
        info.AddValue("MaxNumberOfLines", mMaxNumberOfLines)
        info.AddValue("NavigateUrl", mNavigateUrl)
        info.AddValue("ParentRelation", mParentRelation)
        info.AddValue("TabIndex", mTabIndex)
        info.AddValue("TabStop", mTabStop)
        info.AddValue("Target", mTarget)
        info.AddValue("Text", mText)
        info.AddValue("ToolTip", mToolTip)
        info.AddValue("Value", mValue)

        info.AddValue("Alignment", mAlignment)
        info.AddValue("HotkeyPrefix", mHotkeyPrefix)
        info.AddValue("NoWrap", mNoWrap)
        info.AddValue("RightToLeft", mRightToLeft)
        info.AddValue("Trimming", mTrimming)

        info.AddValue("BottomMargin", mBottomMargin)
        info.AddValue("LeftMargin", mLeftMargin)
        info.AddValue("RightMargin", mRightMargin)
        info.AddValue("TopMargin", mTopMargin)
    End Sub

    Public Function CompareTo(ByVal obj As Object) As Integer Implements IComparable.CompareTo
        If TypeOf obj Is Model10.Label Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Label).TabIndex)
        ElseIf TypeOf obj Is Model10.PictureBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.PictureBox).TabIndex)
        ElseIf TypeOf obj Is Model10.CheckBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.CheckBox).TabIndex)
        ElseIf TypeOf obj Is Model10.Button Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Button).TabIndex)
        ElseIf TypeOf obj Is Model10.Sub Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Sub).TabIndex)
        ElseIf TypeOf obj Is Model10.Chart Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Chart).TabIndex)
        End If
        Throw New ArgumentException("Object is not an 'Tabular'")
    End Function
    Public Function GetStringFormat(ByVal rtl As Windows.Forms.RightToLeft) As Drawing.StringFormat
        Dim sf As New Drawing.StringFormat

        Dim rl As Windows.Forms.RightToLeft
        Select Case mRightToLeft
            Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
            Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
            Case Windows.Forms.RightToLeft.Inherit
                Select Case MyBase.Section.Table.Presentation.RightToLeft
                    Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
                    Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
                    Case Windows.Forms.RightToLeft.Inherit
                        Select Case rtl
                            Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
                            Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
                            Case Windows.Forms.RightToLeft.Inherit
                        End Select
                End Select
        End Select
        If rl = Windows.Forms.RightToLeft.Yes Then
            Select Case mAlignment
                Case Windows.Forms.HorizontalAlignment.Left : sf.Alignment = Drawing.StringAlignment.Far
                Case Windows.Forms.HorizontalAlignment.Right : sf.Alignment = Drawing.StringAlignment.Near
                Case Windows.Forms.HorizontalAlignment.Center : sf.Alignment = Drawing.StringAlignment.Center
            End Select
            sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.DirectionRightToLeft
        Else
            Select Case mAlignment
                Case Windows.Forms.HorizontalAlignment.Left : sf.Alignment = Drawing.StringAlignment.Near
                Case Windows.Forms.HorizontalAlignment.Right : sf.Alignment = Drawing.StringAlignment.Far
                Case Windows.Forms.HorizontalAlignment.Center : sf.Alignment = Drawing.StringAlignment.Center
            End Select
        End If
        If mNoWrap Then
            sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.NoWrap
        End If
        sf.HotkeyPrefix = mHotkeyPrefix
        sf.LineAlignment = Drawing.StringAlignment.Near
        sf.Trimming = mTrimming

        Return sf
    End Function
    Public Function GetMeasureStringFormat(ByVal rtl As Windows.Forms.RightToLeft) As Drawing.StringFormat
        Dim sf As New Drawing.StringFormat

        Dim rl As Windows.Forms.RightToLeft
        Select Case mRightToLeft
            Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
            Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
            Case Windows.Forms.RightToLeft.Inherit
                Select Case MyBase.Section.Table.Presentation.RightToLeft
                    Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
                    Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
                    Case Windows.Forms.RightToLeft.Inherit
                        Select Case rtl
                            Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
                            Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
                            Case Windows.Forms.RightToLeft.Inherit
                        End Select
                End Select
        End Select
        If rl = Windows.Forms.RightToLeft.Yes Then
            Select Case mAlignment
                Case Windows.Forms.HorizontalAlignment.Left : sf.Alignment = Drawing.StringAlignment.Far
                Case Windows.Forms.HorizontalAlignment.Right : sf.Alignment = Drawing.StringAlignment.Near
                Case Windows.Forms.HorizontalAlignment.Center : sf.Alignment = Drawing.StringAlignment.Center
            End Select
            sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.DirectionRightToLeft
        Else
            Select Case mAlignment
                Case Windows.Forms.HorizontalAlignment.Left : sf.Alignment = Drawing.StringAlignment.Near
                Case Windows.Forms.HorizontalAlignment.Right : sf.Alignment = Drawing.StringAlignment.Far
                Case Windows.Forms.HorizontalAlignment.Center : sf.Alignment = Drawing.StringAlignment.Center
            End Select
        End If
        If mNoWrap Then
            sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.NoWrap
        End If
        sf.HotkeyPrefix = mHotkeyPrefix
        sf.LineAlignment = Drawing.StringAlignment.Near
        sf.Trimming = mTrimming

        sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.MeasureTrailingSpaces '<<
        Return sf
    End Function

    Public Overloads Sub Draw(ByVal grp As Drawing.Graphics, ByVal text As String, ByVal offsetx As Integer, ByVal offsetY As Integer, ByVal rtl As Windows.Forms.RightToLeft, ByVal fColor As Drawing.Color)
        Dim k As Single
        Select Case MyBase.Border3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select

        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetx + k / 2, MyBase.Location.Y + offsetY + k / 2, MyBase.Size.Width - k, MyBase.Size.Height - k)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, Drawing.Rectangle.Truncate(r))

        If MyBase.Border3DStyle = Windows.Forms.Border3DStyle.Flat Then
            If k > 0 Then
                Dim p As New Drawing.Pen(MyBase.BorderColor, k)
                grp.DrawRectangle(p, Drawing.Rectangle.Truncate(New Drawing.RectangleF(r.X, r.Y, r.Width, r.Height)))
                p.Dispose() : p = Nothing
            End If
        Else
            Windows.Forms.ControlPaint.DrawBorder3D(grp, New Drawing.Rectangle(MyBase.Location.X + offsetx, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height), MyBase.Border3DStyle)
        End If

        r.Offset(mLeftMargin + k / 2, mTopMargin + k / 2)
        r.Width += -mLeftMargin - mRightMargin - k
        r.Height += -mTopMargin - mBottomMargin - k
        If Not fColor.Equals(Drawing.Color.Empty) And Not fColor.Equals(Drawing.Color.Transparent) And MyBase.BackColor.Equals(Drawing.Color.Transparent) Then
            b.Color = fColor
        Else
            b.Color = mForeColor
        End If
        grp.DrawString(text, mFont, b, r, GetStringFormat(rtl))

        b.Dispose() : b = Nothing
    End Sub
    Public Overloads Sub DrawF(ByVal grp As Drawing.Graphics, ByVal text As String, ByVal offsetx As Integer, ByVal offsetY As Integer, ByVal rtl As Windows.Forms.RightToLeft)
        Dim k As Single = MyBase.BorderWidth
        Dim l As Single
        Select Case MyBase.Border3DStyle
            Case Windows.Forms.Border3DStyle.Flat : l = k
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : If k < 1 Then l = 1 Else l = k
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : If k < 2 Then l = 2 Else l = k
        End Select

        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetx + k / 2, MyBase.Location.Y + offsetY + k / 2, MyBase.Size.Width - k, MyBase.Size.Height - k)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, r)

        If k > 0 Then
            Dim p As New Drawing.Pen(MyBase.BorderColor, k)
            grp.DrawRectangle(p, r.X, r.Y, r.Width, r.Height)
            p.Dispose() : p = Nothing
        End If

        r.Offset(mLeftMargin + l / 2, mTopMargin + l / 2)
        r.Width += -mLeftMargin - mRightMargin - l
        r.Height += -mTopMargin - mBottomMargin - l
        b.Color = mForeColor
        grp.DrawString(text, mFont, b, r, GetStringFormat(rtl))

        b.Dispose() : b = Nothing
    End Sub

    Public Sub New()
        MyBase.New()

        mDifference = 0

        MyBase.BorderColor = Drawing.Color.Transparent
        MyBase.BorderWidth = 0
        MyBase.Size = New Drawing.Size(100, 16)

        mAutoSize = False
        mDataField = ""
        mExpression = ""
        mFilter = ""
        mFont = New Drawing.Font("Tahoma", 8) 'Windows.Forms.Control.DefaultFont
        mForeColor = Drawing.SystemColors.ControlText 'Windows.Forms.Control.DefaultForeColor
        mFormat = ""
        mMaxNumberOfLines = 0
        mNavigateUrl = ""
        mParentRelation = Nothing
        mTabIndex = 0
        mTabStop = True
        mTarget = "_blank"
        mText = ""
        mToolTip = ""
        mValue = Model10.ValueType.None

        mBottomMargin = 0
        mLeftMargin = 0
        mRightMargin = 0
        mTopMargin = 0

        mAlignment = Windows.Forms.HorizontalAlignment.Left
        mHotkeyPrefix = Drawing.Text.HotkeyPrefix.None
        mNoWrap = False
        mRightToLeft = Windows.Forms.RightToLeft.Inherit
        mTrimming = Drawing.StringTrimming.None
    End Sub
    Public Sub New(ByVal label As Model10.Label)
        MyBase.New(label)

        mDifference = label.Difference 'Special NonSerialized Field

        mAutoSize = label.AutoSize
        mDataField = label.DataField
        mExpression = label.Expression
        mFilter = label.Filter
        mFont = CType(label.Font.Clone, Drawing.Font)
        mForeColor = label.ForeColor
        mFormat = label.Format
        mMaxNumberOfLines = label.MaxNumberOfLines
        mNavigateUrl = label.NavigateUrl
        mParentRelation = label.ParentRelation
        mTabIndex = label.TabIndex
        mTabStop = label.TabStop
        mTarget = label.Target
        mText = label.Text
        mToolTip = label.ToolTip
        mValue = label.Value

        mBottomMargin = label.BottomMargin
        mLeftMargin = label.LeftMargin
        mRightMargin = label.RightMargin
        mTopMargin = label.TopMargin

        mAlignment = label.Alignment
        mHotkeyPrefix = label.HotkeyPrefix
        mNoWrap = label.NoWrap
        mRightToLeft = label.RightToLeft
        mTrimming = label.Trimming
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Label(Me)
        CType(Clone, Model10.Label).ParentRelation = Nothing
    End Function

    <ComponentModel.Browsable(False)> Public Property Difference() As Integer
        Get
            Return mDifference
        End Get
        Set(ByVal Value As Integer)
            mDifference = Value
        End Set
    End Property

    Public Property AutoSize() As Boolean
        Get
            Return mAutoSize
        End Get
        Set(ByVal Value As Boolean)
            mAutoSize = Value
        End Set
    End Property
    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
    Public Property Expression() As String
        Get
            Return mExpression
        End Get
        Set(ByVal Value As String)
            mExpression = Value
        End Set
    End Property
    Public Property Filter() As String
        Get
            Return mFilter
        End Get
        Set(ByVal Value As String)
            mFilter = Value
        End Set
    End Property
    Public Property Font() As Drawing.Font
        Get
            Return mFont
        End Get
        Set(ByVal Value As Drawing.Font)
            mFont = Value
        End Set
    End Property
    Public Property ForeColor() As Drawing.Color
        Get
            Return mForeColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mForeColor = Value
        End Set
    End Property
    Public Property Format() As String
        Get
            Return mFormat
        End Get
        Set(ByVal Value As String)
            mFormat = Value
        End Set
    End Property
    Public Property MaxNumberOfLines() As Integer
        Get
            Return mMaxNumberOfLines
        End Get
        Set(ByVal Value As Integer)
            mMaxNumberOfLines = Value
        End Set
    End Property
    Public Property NavigateUrl() As String
        Get
            Return mNavigateUrl
        End Get
        Set(ByVal Value As String)
            mNavigateUrl = Value
        End Set
    End Property
    Public Property ParentRelation() As Model10.Relation
        Get
            Return mParentRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mParentRelation = Value
        End Set
    End Property
    Public Property TabIndex() As Integer
        Get
            Return mTabIndex
        End Get
        Set(ByVal Value As Integer)
            mTabIndex = Value
        End Set
    End Property
    Public Property TabStop() As Boolean
        Get
            Return mTabStop
        End Get
        Set(ByVal Value As Boolean)
            mTabStop = Value
        End Set
    End Property
    Public Property Target() As String
        Get
            Return mTarget
        End Get
        Set(ByVal Value As String)
            mTarget = Value
        End Set
    End Property
    Public Property Text() As String
        Get
            Return mText
        End Get
        Set(ByVal Value As String)
            mText = Value
        End Set
    End Property
    Public Property ToolTip() As String
        Get
            Return mToolTip
        End Get
        Set(ByVal Value As String)
            mToolTip = Value
        End Set
    End Property
    Public Property Value() As Model10.ValueType
        Get
            Return mValue
        End Get
        Set(ByVal Value As Model10.ValueType)
            mValue = Value
        End Set
    End Property

    Public Property BottomMargin() As Integer
        Get
            Return mBottomMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mBottomMargin = Value
        End Set
    End Property
    Public Property LeftMargin() As Integer
        Get
            Return mLeftMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mLeftMargin = Value
        End Set
    End Property
    Public Property RightMargin() As Integer
        Get
            Return mRightMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mRightMargin = Value
        End Set
    End Property
    Public Property TopMargin() As Integer
        Get
            Return mTopMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mTopMargin = Value
        End Set
    End Property

    Public Property Alignment() As Windows.Forms.HorizontalAlignment
        Get
            Return mAlignment
        End Get
        Set(ByVal Value As Windows.Forms.HorizontalAlignment)
            mAlignment = Value
        End Set
    End Property
    Public Property HotkeyPrefix() As Drawing.Text.HotkeyPrefix
        Get
            Return mHotkeyPrefix
        End Get
        Set(ByVal Value As Drawing.Text.HotkeyPrefix)
            mHotkeyPrefix = Value
        End Set
    End Property
    Public Property NoWrap() As Boolean
        Get
            Return mNoWrap
        End Get
        Set(ByVal Value As Boolean)
            mNoWrap = Value
        End Set
    End Property
    Public Property RightToLeft() As Windows.Forms.RightToLeft
        Get
            Return mRightToLeft
        End Get
        Set(ByVal Value As Windows.Forms.RightToLeft)
            mRightToLeft = Value
        End Set
    End Property
    Public Property Trimming() As Drawing.StringTrimming
        Get
            Return mTrimming
        End Get
        Set(ByVal Value As Drawing.StringTrimming)
            mTrimming = Value
        End Set
    End Property
End Class
<Serializable()> Public Class TextBox
    Inherits Model10.Label

    Private mMask As String

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mMask = info.GetString("Mask")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("Mask", mMask)
    End Sub

    Public Sub New()
        MyBase.New()
        MyBase.BackColor = Drawing.Color.White
        MyBase.Border3DStyle = Windows.Forms.Border3DStyle.Sunken
        MyBase.BorderWidth = 1
        MyBase.Size = New Drawing.Size(100, 20)

        mMask = ""
    End Sub
    Public Sub New(ByVal textBox As Model10.TextBox)
        MyBase.New(textBox)

        mMask = textBox.Mask
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.TextBox(Me)
        CType(Clone, Model10.TextBox).ParentRelation = Nothing
    End Function

    Public Property Mask() As String
        Get
            Return mMask
        End Get
        Set(ByVal Value As String)
            mMask = Value
        End Set
    End Property
End Class
<Serializable()> Public Class PictureBox
    Inherits Model10.Frame
    Implements IComparable

    <NonSerialized()> Private mKey As String

    Private mDataField As String
    Private mImage As System.MarshalByRefObject
    Private mNavigateUrl As String
    Private mParentRelation As Model10.Relation
    Private mSizeMode As Model10.SizeMode
    Private mTabIndex As Integer
    Private mTabStop As Boolean
    Private mTarget As String
    Private mToolTip As String

    Private mBottomMargin As Integer
    Private mLeftMargin As Integer
    Private mRightMargin As Integer
    Private mTopMargin As Integer

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mDataField = info.GetString("DataField")
        mImage = CType(info.GetValue("Image", GetType(System.MarshalByRefObject)), System.MarshalByRefObject)
        mNavigateUrl = info.GetString("NavigateUrl")
        mParentRelation = CType(info.GetValue("ParentRelation", GetType(Model10.Relation)), Model10.Relation)
        mSizeMode = CType(info.GetByte("SizeMode"), Model10.SizeMode)
        mTabIndex = info.GetInt32("TabIndex")
        mTabStop = info.GetBoolean("TabStop")
        mTarget = info.GetString("Target")
        mToolTip = info.GetString("ToolTip")

        mBottomMargin = info.GetInt32("BottomMargin")
        mLeftMargin = info.GetInt32("LeftMargin")
        mRightMargin = info.GetInt32("RightMargin")
        mTopMargin = info.GetInt32("TopMargin")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("DataField", mDataField)
        info.AddValue("Image", mImage)
        info.AddValue("NavigateUrl", mNavigateUrl)
        info.AddValue("ParentRelation", mParentRelation)
        info.AddValue("SizeMode", mSizeMode)
        info.AddValue("TabIndex", mTabIndex)
        info.AddValue("TabStop", mTabStop)
        info.AddValue("Target", mTarget)
        info.AddValue("ToolTip", mToolTip)

        info.AddValue("BottomMargin", mBottomMargin)
        info.AddValue("LeftMargin", mLeftMargin)
        info.AddValue("RightMargin", mRightMargin)
        info.AddValue("TopMargin", mTopMargin)
    End Sub

    Public Function CompareTo(ByVal obj As Object) As Integer Implements IComparable.CompareTo
        If TypeOf obj Is Model10.Label Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Label).TabIndex)
        ElseIf TypeOf obj Is Model10.PictureBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.PictureBox).TabIndex)
        ElseIf TypeOf obj Is Model10.CheckBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.CheckBox).TabIndex)
        ElseIf TypeOf obj Is Model10.Button Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Button).TabIndex)
        ElseIf TypeOf obj Is Model10.Sub Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Sub).TabIndex)
        ElseIf TypeOf obj Is Model10.Chart Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Chart).TabIndex)
        End If
        Throw New ArgumentException("Object is not an 'Tabular'")
    End Function

    Public Function ClipRectangle() As Drawing.Rectangle
        Select Case TypeName(mImage)
            Case "Icon"
                Dim k As Single
                Select Case MyBase.Border3DStyle
                    Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
                    Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
                    Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
                End Select

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myIcon As Drawing.Icon = CType(mImage, Drawing.Icon)
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myIcon.Width Then
                            Else
                                destRect.X += (destRect.Width - myIcon.Width) / 2
                                destRect.Width = myIcon.Width
                            End If
                            If destRect.Height < myIcon.Height Then
                            Else
                                destRect.Y += (destRect.Height - myIcon.Height) / 2
                                destRect.Height = myIcon.Height
                            End If
                        Case Model10.SizeMode.Stretch
                        Case Model10.SizeMode.Zoom
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myIcon.Width / myIcon.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    myIcon = Nothing
                End If
                Return Drawing.Rectangle.Round(destRect)
            Case Else
                Dim k As Single
                Select Case MyBase.Border3DStyle
                    Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
                    Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
                    Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
                End Select

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myImage As Drawing.Image = CType(mImage, Drawing.Image)
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myImage.Width Then
                            Else
                                destRect.X += (destRect.Width - myImage.Width) / 2
                                destRect.Width = myImage.Width
                            End If
                            If destRect.Height < myImage.Height Then
                            Else
                                destRect.Y += (destRect.Height - myImage.Height) / 2
                                destRect.Height = myImage.Height
                            End If
                        Case Model10.SizeMode.Stretch
                        Case Model10.SizeMode.Zoom
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myImage.Width / myImage.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    myImage = Nothing
                End If
                Return Drawing.Rectangle.Round(destRect)
        End Select
    End Function
    Public Function ClipRectangleF() As Drawing.Rectangle
        Select Case TypeName(mImage)
            Case "Icon"
                Dim k As Single = MyBase.BorderWidth

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myIcon As Drawing.Icon = CType(mImage, Drawing.Icon)
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myIcon.Width Then
                            Else
                                destRect.X += (destRect.Width - myIcon.Width) / 2
                                destRect.Width = myIcon.Width
                            End If
                            If destRect.Height < myIcon.Height Then
                            Else
                                destRect.Y += (destRect.Height - myIcon.Height) / 2
                                destRect.Height = myIcon.Height
                            End If
                        Case Model10.SizeMode.Stretch
                        Case Model10.SizeMode.Zoom
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myIcon.Width / myIcon.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    myIcon = Nothing
                End If
                Return Drawing.Rectangle.Round(destRect)
            Case Else
                Dim k As Single = MyBase.BorderWidth

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myImage As Drawing.Image = CType(mImage, Drawing.Image)
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myImage.Width Then
                            Else
                                destRect.X += (destRect.Width - myImage.Width) / 2
                                destRect.Width = myImage.Width
                            End If
                            If destRect.Height < myImage.Height Then
                            Else
                                destRect.Y += (destRect.Height - myImage.Height) / 2
                                destRect.Height = myImage.Height
                            End If
                        Case Model10.SizeMode.Stretch
                        Case Model10.SizeMode.Zoom
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myImage.Width / myImage.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    myImage = Nothing
                End If
                Return Drawing.Rectangle.Round(destRect)
        End Select
    End Function
    Public Function GetImage(ByVal bkColor As Drawing.Color) As Drawing.Bitmap
        Select Case TypeName(mImage)
            Case "Icon"
                Dim k As Single
                Select Case MyBase.Border3DStyle
                    Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
                    Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
                    Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
                End Select

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myIcon As Drawing.Icon = CType(mImage, Drawing.Icon)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myIcon.Width Then
                                srcRect.X = (myIcon.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width
                            Else
                                srcRect.X = 0
                                srcRect.Width = myIcon.Width

                                destRect.X += (destRect.Width - myIcon.Width) / 2
                                destRect.Width = myIcon.Width
                            End If
                            If destRect.Height < myIcon.Height Then
                                srcRect.Y = (myIcon.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myIcon.Height

                                destRect.Y += (destRect.Height - myIcon.Height) / 2
                                destRect.Height = myIcon.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myIcon.Width, myIcon.Height)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myIcon.Width, myIcon.Height)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myIcon.Width / myIcon.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.X = 0 : destRect.Y = 0
                    GetImage = New Drawing.Bitmap(CInt(destRect.Width), CInt(destRect.Height))
                    Dim grp As Drawing.Graphics = Drawing.Graphics.FromImage(GetImage)
                    grp.Clear(bkColor)
                    grp.Drawicon(myIcon, Drawing.Rectangle.Truncate(destRect))
                    grp.Dispose() : grp = Nothing
                    myIcon = Nothing
                End If
            Case Else
                Dim k As Single
                Select Case MyBase.Border3DStyle
                    Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
                    Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
                    Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
                End Select

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myImage As Drawing.Image = CType(mImage, Drawing.Image)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myImage.Width Then
                                srcRect.X = (myImage.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width
                            Else
                                srcRect.X = 0
                                srcRect.Width = myImage.Width

                                destRect.X += (destRect.Width - myImage.Width) / 2
                                destRect.Width = myImage.Width
                            End If
                            If destRect.Height < myImage.Height Then
                                srcRect.Y = (myImage.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myImage.Height

                                destRect.Y += (destRect.Height - myImage.Height) / 2
                                destRect.Height = myImage.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width, myImage.Height)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width, myImage.Height)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myImage.Width / myImage.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.X = 0 : destRect.Y = 0
                    GetImage = New Drawing.Bitmap(CInt(destRect.Width), CInt(destRect.Height))
                    Dim grp As Drawing.Graphics = Drawing.Graphics.FromImage(GetImage)
                    grp.Clear(bkColor)
                    grp.DrawImage(myImage, destRect, srcRect, Drawing.GraphicsUnit.Pixel)
                    grp.Dispose() : grp = Nothing
                    myImage = Nothing
                End If
        End Select
    End Function
    Public Function GetImageF(ByVal bkColor As Drawing.Color) As Drawing.Bitmap
        Select Case TypeName(mImage)
            Case "Icon"
                Dim k As Single = MyBase.BorderWidth

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myIcon As Drawing.Icon = CType(mImage, Drawing.Icon)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myIcon.Width Then
                                srcRect.X = (myIcon.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width
                            Else
                                srcRect.X = 0
                                srcRect.Width = myIcon.Width

                                destRect.X += (destRect.Width - myIcon.Width) / 2
                                destRect.Width = myIcon.Width
                            End If
                            If destRect.Height < myIcon.Height Then
                                srcRect.Y = (myIcon.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myIcon.Height

                                destRect.Y += (destRect.Height - myIcon.Height) / 2
                                destRect.Height = myIcon.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myIcon.Width, myIcon.Height)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myIcon.Width, myIcon.Height)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myIcon.Width / myIcon.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.X = 0 : destRect.Y = 0
                    GetImageF = New Drawing.Bitmap(CInt(destRect.Width), CInt(destRect.Height))
                    Dim grp As Drawing.Graphics = Drawing.Graphics.FromImage(GetImageF)
                    grp.Clear(bkColor)
                    grp.DrawIcon(myIcon, Drawing.Rectangle.Truncate(destRect))
                    grp.Dispose() : grp = Nothing
                    myIcon = Nothing
                End If
            Case Else
                Dim k As Single = MyBase.BorderWidth

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myImage As Drawing.Image = CType(mImage, Drawing.Image)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myImage.Width Then
                                srcRect.X = (myImage.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width
                            Else
                                srcRect.X = 0
                                srcRect.Width = myImage.Width

                                destRect.X += (destRect.Width - myImage.Width) / 2
                                destRect.Width = myImage.Width
                            End If
                            If destRect.Height < myImage.Height Then
                                srcRect.Y = (myImage.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myImage.Height

                                destRect.Y += (destRect.Height - myImage.Height) / 2
                                destRect.Height = myImage.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width, myImage.Height)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width, myImage.Height)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myImage.Width / myImage.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.X = 0 : destRect.Y = 0
                    GetImageF = New Drawing.Bitmap(CInt(destRect.Width), CInt(destRect.Height))
                    Dim grp As Drawing.Graphics = Drawing.Graphics.FromImage(GetImageF)
                    grp.Clear(bkColor)
                    grp.DrawImage(myImage, destRect, srcRect, Drawing.GraphicsUnit.Pixel)
                    grp.Dispose() : grp = Nothing
                    myImage = Nothing
                End If
        End Select
    End Function
    Public Overloads Sub Draw(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        Select Case TypeName(mImage)
            Case "Icon"
                Dim k As Single
                Select Case MyBase.Border3DStyle
                    Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
                    Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
                    Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
                End Select

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim x As Integer = MyBase.Location.X
                Dim y As Integer = MyBase.Location.Y
                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myIcon As Drawing.Icon = CType(mImage, Drawing.Icon)
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myIcon.Width Then
                            Else
                                destRect.X += (destRect.Width - myIcon.Width) / 2
                                destRect.Width = myIcon.Width
                            End If
                            If destRect.Height < myIcon.Height Then
                            Else
                                destRect.Y += (destRect.Height - myIcon.Height) / 2
                                destRect.Height = myIcon.Height
                            End If
                        Case Model10.SizeMode.Stretch
                        Case Model10.SizeMode.Zoom
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myIcon.Width / myIcon.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.Offset(offsetX + x, offsetY + y)

                    Dim r As Drawing.RectangleF = destRect
                    r.Offset(-mLeftMargin - k / 2, -mTopMargin - k / 2)
                    r.Width += mLeftMargin + mRightMargin + k
                    r.Height += mTopMargin + mBottomMargin + k
                    Dim b As New Drawing.SolidBrush(MyBase.BackColor)
                    grp.FillRectangle(b, Drawing.Rectangle.Truncate(r))
                    b.Dispose() : b = Nothing

                    Try
                        grp.DrawIcon(myIcon, Drawing.Rectangle.Truncate(destRect))
                    Catch ex As Exception
                    End Try
                    myIcon = Nothing

                    If MyBase.Border3DStyle = Windows.Forms.Border3DStyle.Flat Then
                        If k > 0 Then
                            Dim p As New Drawing.Pen(MyBase.BorderColor, k)
                            grp.DrawRectangle(p, Drawing.Rectangle.Truncate(New Drawing.RectangleF(r.X, r.Y, r.Width, r.Height)))
                            p.Dispose() : p = Nothing
                        End If
                    Else
                        Windows.Forms.ControlPaint.DrawBorder3D(grp, New Drawing.Rectangle(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height), MyBase.Border3DStyle)
                    End If
                End If
            Case Else
                Dim k As Single
                Select Case MyBase.Border3DStyle
                    Case Windows.Forms.Border3DStyle.Flat : k = MyBase.BorderWidth
                    Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
                    Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
                End Select

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim x As Integer = MyBase.Location.X
                Dim y As Integer = MyBase.Location.Y
                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myImage As Drawing.Image = CType(mImage, Drawing.Image)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myImage.Width Then
                                srcRect.X = (myImage.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width - 1
                            Else
                                srcRect.X = 0
                                srcRect.Width = myImage.Width - 1

                                destRect.X += (destRect.Width - myImage.Width) / 2
                                destRect.Width = myImage.Width
                            End If
                            If destRect.Height < myImage.Height Then
                                srcRect.Y = (myImage.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height - 1
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myImage.Height - 1

                                destRect.Y += (destRect.Height - myImage.Height) / 2
                                destRect.Height = myImage.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width - 1, myImage.Height - 1)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width - 1, myImage.Height - 1)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myImage.Width / myImage.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.Offset(offsetX + x, offsetY + y)

                    Dim r As Drawing.RectangleF = destRect
                    r.Offset(-mLeftMargin - k / 2, -mTopMargin - k / 2)
                    r.Width += mLeftMargin + mRightMargin + k
                    r.Height += mTopMargin + mBottomMargin + k
                    Dim b As New Drawing.SolidBrush(MyBase.BackColor)
                    grp.FillRectangle(b, Drawing.Rectangle.Truncate(r))
                    b.Dispose() : b = Nothing

                    Try
                        grp.DrawImage(myImage, Drawing.Rectangle.Truncate(destRect), Drawing.Rectangle.Truncate(srcRect), Drawing.GraphicsUnit.Pixel)
                    Catch ex As Exception
                    End Try
                    myImage = Nothing

                    If MyBase.Border3DStyle = Windows.Forms.Border3DStyle.Flat Then
                        If k > 0 Then
                            Dim p As New Drawing.Pen(MyBase.BorderColor, k)
                            grp.DrawRectangle(p, Drawing.Rectangle.Truncate(New Drawing.RectangleF(r.X, r.Y, r.Width, r.Height)))
                            p.Dispose() : p = Nothing
                        End If
                    Else
                        Windows.Forms.ControlPaint.DrawBorder3D(grp, New Drawing.Rectangle(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height), MyBase.Border3DStyle)
                    End If
                End If
        End Select
    End Sub
    Public Overloads Sub DrawF(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        Select Case TypeName(mImage)
            Case "Icon"
                '   !
                Exit Sub
                '---------------
                Dim k As Single = MyBase.BorderWidth

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim x As Integer = MyBase.Location.X
                Dim y As Integer = MyBase.Location.Y
                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myIcon As Drawing.Icon = CType(mImage, Drawing.Icon)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myIcon.Width Then
                                srcRect.X = (myIcon.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width - 1
                            Else
                                srcRect.X = 0
                                srcRect.Width = myIcon.Width - 1

                                destRect.X += (destRect.Width - myIcon.Width) / 2
                                destRect.Width = myIcon.Width
                            End If
                            If destRect.Height < myIcon.Height Then
                                srcRect.Y = (myIcon.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height - 1
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myIcon.Height - 1

                                destRect.Y += (destRect.Height - myIcon.Height) / 2
                                destRect.Height = myIcon.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myIcon.Width - 1, myIcon.Height - 1)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myIcon.Width - 1, myIcon.Height - 1)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myIcon.Width / myIcon.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.Offset(offsetX + x, offsetY + y)

                    Dim r As Drawing.RectangleF = destRect
                    r.Offset(-mLeftMargin - k / 2, -mTopMargin - k / 2)
                    r.Width += mLeftMargin + mRightMargin + k
                    r.Height += mTopMargin + mBottomMargin + k
                    Dim b As New Drawing.SolidBrush(MyBase.BackColor)
                    grp.FillRectangle(b, r)
                    b.Dispose() : b = Nothing

                    Try
                        grp.DrawIcon(myIcon, Drawing.Rectangle.Truncate(destRect))
                    Catch ex As Exception
                    End Try
                    myIcon = Nothing

                    If k > 0 Then
                        Dim p As New Drawing.Pen(MyBase.BorderColor, k)
                        grp.DrawRectangle(p, r.X, r.Y, r.Width, r.Height)
                        p.Dispose() : p = Nothing
                    End If
                End If
            Case Else
                Dim k As Single = MyBase.BorderWidth

                Dim bm As Integer = mBottomMargin
                Dim lm As Integer = mLeftMargin
                Dim rm As Integer = mRightMargin
                Dim tm As Integer = mTopMargin

                Dim x As Integer = MyBase.Location.X
                Dim y As Integer = MyBase.Location.Y
                Dim w As Integer = MyBase.Size.Width
                Dim h As Integer = MyBase.Size.Height

                Dim destRect As New Drawing.RectangleF(lm + k, tm + k, w - (lm + rm + 2 * k), h - (tm + bm + 2 * k))
                If destRect.Width > 0 And destRect.Height > 0 Then
                    Dim myImage As Drawing.Image = CType(mImage, Drawing.Image)
                    Dim srcRect As New Drawing.RectangleF
                    Select Case mSizeMode
                        Case Model10.SizeMode.Clip
                            If destRect.Width < myImage.Width Then
                                srcRect.X = (myImage.Width - destRect.Width) / 2
                                srcRect.Width = destRect.Width - 1
                            Else
                                srcRect.X = 0
                                srcRect.Width = myImage.Width - 1

                                destRect.X += (destRect.Width - myImage.Width) / 2
                                destRect.Width = myImage.Width
                            End If
                            If destRect.Height < myImage.Height Then
                                srcRect.Y = (myImage.Height - destRect.Height) / 2
                                srcRect.Height = destRect.Height - 1
                            Else
                                srcRect.Y = 0
                                srcRect.Height = myImage.Height - 1

                                destRect.Y += (destRect.Height - myImage.Height) / 2
                                destRect.Height = myImage.Height
                            End If
                        Case Model10.SizeMode.Stretch
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width - 1, myImage.Height - 1)
                        Case Model10.SizeMode.Zoom
                            srcRect = New Drawing.RectangleF(0, 0, myImage.Width - 1, myImage.Height - 1)
                            Dim destinationRate As Single = destRect.Width / destRect.Height
                            Dim imageRate As Single = CSng(myImage.Width / myImage.Height)
                            If destinationRate < imageRate Then
                                Dim p As Single = destRect.Width / imageRate
                                destRect.Y += (destRect.Height - p) / 2
                                destRect.Height = p
                            Else
                                Dim p As Single = destRect.Height * imageRate
                                destRect.X += (destRect.Width - p) / 2
                                destRect.Width = p
                            End If
                    End Select
                    destRect.Offset(offsetX + x, offsetY + y)

                    Dim r As Drawing.RectangleF = destRect
                    r.Offset(-mLeftMargin - k / 2, -mTopMargin - k / 2)
                    r.Width += mLeftMargin + mRightMargin + k
                    r.Height += mTopMargin + mBottomMargin + k
                    Dim b As New Drawing.SolidBrush(MyBase.BackColor)
                    grp.FillRectangle(b, r)
                    b.Dispose() : b = Nothing

                    Try
                        grp.DrawImage(myImage, destRect, srcRect, Drawing.GraphicsUnit.Pixel)
                    Catch ex As Exception
                    End Try
                    myImage = Nothing

                    If k > 0 Then
                        Dim p As New Drawing.Pen(MyBase.BorderColor, k)
                        grp.DrawRectangle(p, r.X, r.Y, r.Width, r.Height)
                        p.Dispose() : p = Nothing
                    End If
                End If
        End Select
    End Sub

    Public Sub New()
        MyBase.New()
        MyBase.BorderColor = Drawing.Color.Transparent
        MyBase.BorderWidth = 0
        MyBase.Size = New Drawing.Size(100, 100)

        mKey = ""

        mDataField = ""
        mImage = Nothing
        mNavigateUrl = ""
        mParentRelation = Nothing
        mSizeMode = Model10.SizeMode.Clip
        mTabIndex = 0
        mTabStop = True
        mTarget = ""
        mToolTip = ""

        mBottomMargin = 0
        mLeftMargin = 0
        mRightMargin = 0
        mTopMargin = 0
    End Sub
    Public Sub New(ByVal pictureBox As Model10.PictureBox)
        MyBase.New(pictureBox)

        mKey = pictureBox.Key

        mDataField = pictureBox.DataField
        mImage = pictureBox.Image
        mNavigateUrl = pictureBox.NavigateUrl
        mParentRelation = pictureBox.ParentRelation
        mSizeMode = pictureBox.SizeMode
        mTabIndex = pictureBox.TabIndex
        mTabStop = pictureBox.TabStop
        mTarget = pictureBox.Target
        mToolTip = pictureBox.ToolTip

        mBottomMargin = pictureBox.BottomMargin
        mLeftMargin = pictureBox.LeftMargin
        mRightMargin = pictureBox.RightMargin
        mTopMargin = pictureBox.TopMargin
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.PictureBox(Me)
        CType(Clone, Model10.PictureBox).ParentRelation = Nothing
    End Function

    <ComponentModel.Browsable(False)> Public Property Key() As String
        Get
            Return mKey
        End Get
        Set(ByVal Value As String)
            mKey = Value
        End Set
    End Property

    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
    Public Property Image() As System.MarshalByRefObject
        Get
            Return mImage
        End Get
        Set(ByVal Value As System.MarshalByRefObject)
            mImage = Value
        End Set
    End Property
    Public Property NavigateUrl() As String
        Get
            Return mNavigateUrl
        End Get
        Set(ByVal Value As String)
            mNavigateUrl = Value
        End Set
    End Property
    Public Property ParentRelation() As Model10.Relation
        Get
            Return mParentRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mParentRelation = Value
        End Set
    End Property
    Public Property SizeMode() As Model10.SizeMode
        Get
            Return mSizeMode
        End Get
        Set(ByVal Value As Model10.SizeMode)
            mSizeMode = Value
        End Set
    End Property
    Public Property TabIndex() As Integer
        Get
            Return mTabIndex
        End Get
        Set(ByVal Value As Integer)
            mTabIndex = Value
        End Set
    End Property
    Public Property TabStop() As Boolean
        Get
            Return mTabStop
        End Get
        Set(ByVal Value As Boolean)
            mTabStop = Value
        End Set
    End Property
    Public Property Target() As String
        Get
            Return mTarget
        End Get
        Set(ByVal Value As String)
            mTarget = Value
        End Set
    End Property
    Public Property ToolTip() As String
        Get
            Return mToolTip
        End Get
        Set(ByVal Value As String)
            mToolTip = Value
        End Set
    End Property

    Public Property BottomMargin() As Integer
        Get
            Return mBottomMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mBottomMargin = Value
        End Set
    End Property
    Public Property LeftMargin() As Integer
        Get
            Return mLeftMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mLeftMargin = Value
        End Set
    End Property
    Public Property RightMargin() As Integer
        Get
            Return mRightMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mRightMargin = Value
        End Set
    End Property
    Public Property TopMargin() As Integer
        Get
            Return mTopMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mTopMargin = Value
        End Set
    End Property
End Class
<Serializable()> Public Class CheckBase
    Inherits Model10.Member

    Private mChecked As Boolean
    Private mCheckStyle As Model10.CheckStyle
    Private mDataField As String
    Private mExpression As String
    Private mFilter As String
    Private mParentRelation As Model10.Relation
    Private mToolTip As String

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mChecked = info.GetBoolean("Checked")
        mCheckStyle = CType(info.GetByte("CheckStyle"), Model10.CheckStyle)
        mDataField = info.GetString("DataField")
        mExpression = info.GetString("Expression")
        mFilter = info.GetString("Filter")
        mParentRelation = CType(info.GetValue("ParentRelation", GetType(Model10.Relation)), Model10.Relation)
        mToolTip = info.GetString("ToolTip")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("Checked", mChecked)
        info.AddValue("CheckStyle", mCheckStyle)
        info.AddValue("DataField", mDataField)
        info.AddValue("Expression", mExpression)
        info.AddValue("Filter", mFilter)
        info.AddValue("ParentRelation", mParentRelation)
        info.AddValue("ToolTip", mToolTip)
    End Sub

    Public Function CheckPointF() As Drawing.Point
        Const CHECK_SIZE As Byte = 11

        Dim r As New Drawing.Rectangle(MyBase.Location, MyBase.Size)
        If r.Width > CHECK_SIZE Then r.X += (r.Width - CHECK_SIZE) \ 2
        If r.Height > CHECK_SIZE Then r.Y += (r.Height - CHECK_SIZE) \ 2
        Return r.Location
    End Function
    Public Shared Function GetImageF(ByVal checked As Boolean) As Drawing.Bitmap
        Const CHECK_SIZE As Byte = 11

        GetImageF = New Drawing.Bitmap(CHECK_SIZE, CHECK_SIZE)
        Dim grp As Drawing.Graphics = Drawing.Graphics.FromImage(GetImageF)
        If checked Then
            Windows.Forms.ControlPaint.DrawCheckBox(grp, New Drawing.Rectangle(-1, -1, 13, 13), Windows.Forms.ButtonState.Flat Or Windows.Forms.ButtonState.Checked)
        Else
            Windows.Forms.ControlPaint.DrawCheckBox(grp, New Drawing.Rectangle(-1, -1, 13, 13), Windows.Forms.ButtonState.Flat)
        End If
        grp.Dispose() : grp = Nothing
    End Function
    Public Sub Draw(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        Const CHECK_SIZE As Byte = 13

        Dim r As New Drawing.Rectangle(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height)
        If r.Width > CHECK_SIZE Then
            r.X += (r.Width - CHECK_SIZE) \ 2
            r.Width = CHECK_SIZE
        End If
        If r.Height > CHECK_SIZE Then
            r.Y += (r.Height - CHECK_SIZE) \ 2
            r.Height = CHECK_SIZE
        End If

        Select Case mCheckStyle
            Case CheckStyle.Normal
                If mChecked Then
                    Windows.Forms.ControlPaint.DrawCheckBox(grp, r, Windows.Forms.ButtonState.Normal Or Windows.Forms.ButtonState.Checked)
                Else
                    Windows.Forms.ControlPaint.DrawCheckBox(grp, r, Windows.Forms.ButtonState.Normal)
                End If
            Case CheckStyle.Flat
                If mChecked Then
                    Windows.Forms.ControlPaint.DrawCheckBox(grp, r, Windows.Forms.ButtonState.Flat Or Windows.Forms.ButtonState.Checked)
                Else
                    Windows.Forms.ControlPaint.DrawCheckBox(grp, r, Windows.Forms.ButtonState.Flat)
                End If
        End Select
    End Sub
    Public Sub DrawF(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        Const CHECK_SIZE As Byte = 11

        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height)
        If r.Width > CHECK_SIZE Then
            r.X += (r.Width - CHECK_SIZE) / 2
            r.Width = CHECK_SIZE
        End If
        If r.Height > CHECK_SIZE Then
            r.Y += (r.Height - CHECK_SIZE) / 2
            r.Height = CHECK_SIZE
        End If

        grp.FillRectangle(Drawing.Brushes.White, r)
        If mChecked Then
            Dim p(5) As Drawing.PointF
            p(0) = New Drawing.PointF(r.X + 2, r.Y + 5)
            p(1) = New Drawing.PointF(r.X + 4, r.Y + 7)
            p(2) = New Drawing.PointF(r.X + 9, r.Y + 2)
            p(3) = New Drawing.PointF(r.X + 9, r.Y + 5)
            p(4) = New Drawing.PointF(r.X + 4, r.Y + 10)
            p(5) = New Drawing.PointF(r.X + 2, r.Y + 8)
            grp.FillPolygon(Drawing.Brushes.Black, p)
        End If
        grp.DrawRectangle(Drawing.SystemPens.ControlDark, r.X, r.Y, r.Width, r.Height)
    End Sub

    Public Sub New()
        MyBase.New()
        MyBase.Size = New Drawing.Size(16, 16)

        mChecked = False
        mCheckStyle = CheckStyle.Normal
        mDataField = ""
        mExpression = ""
        mFilter = ""
        mParentRelation = Nothing
        mToolTip = ""
    End Sub
    Public Sub New(ByVal checkBase As Model10.CheckBase)
        MyBase.New(checkBase)

        mChecked = checkBase.Checked
        mCheckStyle = checkBase.CheckStyle
        mDataField = checkBase.DataField
        mExpression = checkBase.Expression
        mFilter = checkBase.Filter
        mParentRelation = checkBase.ParentRelation
        mToolTip = checkBase.ToolTip
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.CheckBase(Me)
        CType(Clone, Model10.CheckBase).ParentRelation = Nothing
    End Function

    Public Property Checked() As Boolean
        Get
            Return mChecked
        End Get
        Set(ByVal Value As Boolean)
            mChecked = Value
        End Set
    End Property
    Public Property CheckStyle() As Model10.CheckStyle
        Get
            Return mCheckStyle
        End Get
        Set(ByVal Value As Model10.CheckStyle)
            mCheckStyle = Value
        End Set
    End Property
    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
    Public Property Expression() As String
        Get
            Return mExpression
        End Get
        Set(ByVal Value As String)
            mExpression = Value
        End Set
    End Property
    Public Property Filter() As String
        Get
            Return mFilter
        End Get
        Set(ByVal Value As String)
            mFilter = Value
        End Set
    End Property
    Public Property ParentRelation() As Model10.Relation
        Get
            Return mParentRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mParentRelation = Value
        End Set
    End Property
    Public Property ToolTip() As String
        Get
            Return mToolTip
        End Get
        Set(ByVal Value As String)
            mToolTip = Value
        End Set
    End Property
End Class
<Serializable()> Public Class CheckBox
    Inherits Model10.CheckBase
    Implements IComparable

    Private mTabIndex As Integer
    Private mTabStop As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mTabIndex = info.GetInt32("TabIndex")
        mTabStop = info.GetBoolean("TabStop")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("TabIndex", mTabIndex)
        info.AddValue("TabStop", mTabStop)
    End Sub

    Public Function CompareTo(ByVal obj As Object) As Integer Implements IComparable.CompareTo
        If TypeOf obj Is Model10.Label Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Label).TabIndex)
        ElseIf TypeOf obj Is Model10.PictureBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.PictureBox).TabIndex)
        ElseIf TypeOf obj Is Model10.CheckBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.CheckBox).TabIndex)
        ElseIf TypeOf obj Is Model10.Button Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Button).TabIndex)
        ElseIf TypeOf obj Is Model10.Sub Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Sub).TabIndex)
        ElseIf TypeOf obj Is Model10.Chart Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Chart).TabIndex)
        End If
        Throw New ArgumentException("Object is not an 'Tabular'")
    End Function

    Public Sub New()
        MyBase.New()
        MyBase.Size = New Drawing.Size(16, 16)

        mTabIndex = 0
        mTabStop = True
    End Sub
    Public Sub New(ByVal checkBox As Model10.CheckBox)
        MyBase.New(checkBox)

        mTabIndex = checkBox.TabIndex
        mTabStop = checkBox.TabStop
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.CheckBox(Me)
        CType(Clone, Model10.CheckBase).ParentRelation = Nothing
    End Function

    Public Property TabIndex() As Integer
        Get
            Return mTabIndex
        End Get
        Set(ByVal Value As Integer)
            mTabIndex = Value
        End Set
    End Property
    Public Property TabStop() As Boolean
        Get
            Return mTabStop
        End Get
        Set(ByVal Value As Boolean)
            mTabStop = Value
        End Set
    End Property
End Class
<Serializable()> Public Class [Sub]
    Inherits Model10.Frame
    Implements IComparable

    <NonSerialized()> Private mDifference As Integer

    Private mAutoSize As Boolean
    Private mChildRelation As Model10.Relation
    Private mTabIndex As Integer
    Private mTabStop As Boolean
    Private mToolTip As String

    Private mBottomMargin As Integer
    Private mLeftMargin As Integer
    Private mRightMargin As Integer
    Private mTopMargin As Integer

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mAutoSize = info.GetBoolean("AutoSize")
        mChildRelation = CType(info.GetValue("ChildRelation", GetType(Model10.Relation)), Model10.Relation)
        mTabIndex = info.GetInt32("TabIndex")
        mTabStop = info.GetBoolean("TabStop")
        mToolTip = info.GetString("ToolTip")

        mBottomMargin = info.GetInt32("BottomMargin")
        mLeftMargin = info.GetInt32("LeftMargin")
        mRightMargin = info.GetInt32("RightMargin")
        mTopMargin = info.GetInt32("TopMargin")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("AutoSize", mAutoSize)
        info.AddValue("ChildRelation", mChildRelation)
        info.AddValue("TabIndex", mTabIndex)
        info.AddValue("TabStop", mTabStop)
        info.AddValue("ToolTip", mToolTip)

        info.AddValue("BottomMargin", mBottomMargin)
        info.AddValue("LeftMargin", mLeftMargin)
        info.AddValue("RightMargin", mRightMargin)
        info.AddValue("TopMargin", mTopMargin)
    End Sub

    Public Function CompareTo(ByVal obj As Object) As Integer Implements IComparable.CompareTo
        If TypeOf obj Is Model10.Label Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Label).TabIndex)
        ElseIf TypeOf obj Is Model10.PictureBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.PictureBox).TabIndex)
        ElseIf TypeOf obj Is Model10.CheckBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.CheckBox).TabIndex)
        ElseIf TypeOf obj Is Model10.Button Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Button).TabIndex)
        ElseIf TypeOf obj Is Model10.Sub Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Sub).TabIndex)
        ElseIf TypeOf obj Is Model10.Chart Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Chart).TabIndex)
        End If
        Throw New ArgumentException("Object is not an 'Tabular'")
    End Function

    Public Sub New()
        MyBase.New()

        mDifference = 0

        MyBase.BorderColor = Drawing.Color.Transparent
        MyBase.BorderWidth = 0
        MyBase.Size = New Drawing.Size(200, 16)

        mAutoSize = False
        mChildRelation = Nothing
        mTabIndex = 0
        mTabStop = True
        mToolTip = ""

        mBottomMargin = 0
        mLeftMargin = 0
        mRightMargin = 0
        mTopMargin = 0
    End Sub
    Public Sub New(ByVal [sub] As Model10.Sub)
        MyBase.New([sub])

        mDifference = [sub].Difference 'Special NonSerialized Field

        mAutoSize = [sub].AutoSize
        mChildRelation = [sub].ChildRelation
        mTabIndex = [sub].TabIndex
        mTabStop = [sub].TabStop
        mToolTip = [sub].ToolTip

        mBottomMargin = [sub].BottomMargin
        mLeftMargin = [sub].LeftMargin
        mRightMargin = [sub].RightMargin
        mTopMargin = [sub].TopMargin
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Sub(Me)
        CType(Clone, Model10.Sub).ChildRelation = Nothing
    End Function

    <ComponentModel.Browsable(False)> Public Property Difference() As Integer
        Get
            Return mDifference
        End Get
        Set(ByVal Value As Integer)
            mDifference = Value
        End Set
    End Property

    Public Property AutoSize() As Boolean
        Get
            Return mAutoSize
        End Get
        Set(ByVal Value As Boolean)
            mAutoSize = Value
        End Set
    End Property
    Public Property ChildRelation() As Model10.Relation
        Get
            Return mChildRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mChildRelation = Value
        End Set
    End Property
    Public Property TabIndex() As Integer
        Get
            Return mTabIndex
        End Get
        Set(ByVal Value As Integer)
            mTabIndex = Value
        End Set
    End Property
    Public Property TabStop() As Boolean
        Get
            Return mTabStop
        End Get
        Set(ByVal Value As Boolean)
            mTabStop = Value
        End Set
    End Property
    Public Property ToolTip() As String
        Get
            Return mToolTip
        End Get
        Set(ByVal Value As String)
            mToolTip = Value
        End Set
    End Property

    Public Property BottomMargin() As Integer
        Get
            Return mBottomMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mBottomMargin = Value
        End Set
    End Property
    Public Property LeftMargin() As Integer
        Get
            Return mLeftMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mLeftMargin = Value
        End Set
    End Property
    Public Property RightMargin() As Integer
        Get
            Return mRightMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mRightMargin = Value
        End Set
    End Property
    Public Property TopMargin() As Integer
        Get
            Return mTopMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mTopMargin = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Chart
    Inherits Model10.Frame
    Implements IComparable

    <NonSerialized()> Private mBitmap As Drawing.Bitmap

    Private mChartType As OWC11.ChartChartTypeEnum
    Private mChildRelation As Model10.Relation
    Private mDataField As String
    Private mPlotColor As Drawing.Color
    Private ReadOnly mSeries As Model10.SeriesCollection
    Private mTabIndex As Integer
    Private mTabStop As Boolean
    Private mTitle As String
    Private mToolTip As String

    Private mBottomMargin As Integer
    Private mLeftMargin As Integer
    Private mRightMargin As Integer
    Private mTopMargin As Integer

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mChartType = CType(info.GetInt32("ChartType"), OWC11.ChartChartTypeEnum)
        mChildRelation = CType(info.GetValue("ChildRelation", GetType(Model10.Relation)), Model10.Relation)
        mDataField = info.GetString("DataField")
        mPlotColor = CType(info.GetValue("PlotColor", GetType(Drawing.Color)), Drawing.Color)
        mSeries = CType(info.GetValue("Series", GetType(Model10.SeriesCollection)), Model10.SeriesCollection)
        mTabIndex = info.GetInt32("TabIndex")
        mTabStop = info.GetBoolean("TabStop")
        mTitle = info.GetString("Title")
        mToolTip = info.GetString("ToolTip")

        mBottomMargin = info.GetInt32("BottomMargin")
        mLeftMargin = info.GetInt32("LeftMargin")
        mRightMargin = info.GetInt32("RightMargin")
        mTopMargin = info.GetInt32("TopMargin")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("ChartType", mChartType)
        info.AddValue("ChildRelation", mChildRelation)
        info.AddValue("DataField", mDataField)
        info.AddValue("PlotColor", mPlotColor)
        info.AddValue("Series", mSeries)
        info.AddValue("TabIndex", mTabIndex)
        info.AddValue("TabStop", mTabStop)
        info.AddValue("Title", mTitle)
        info.AddValue("ToolTip", mToolTip)

        info.AddValue("BottomMargin", mBottomMargin)
        info.AddValue("LeftMargin", mLeftMargin)
        info.AddValue("RightMargin", mRightMargin)
        info.AddValue("TopMargin", mTopMargin)
    End Sub

    Public Function CompareTo(ByVal obj As Object) As Integer Implements IComparable.CompareTo
        If TypeOf obj Is Model10.Label Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Label).TabIndex)
        ElseIf TypeOf obj Is Model10.PictureBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.PictureBox).TabIndex)
        ElseIf TypeOf obj Is Model10.CheckBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.CheckBox).TabIndex)
        ElseIf TypeOf obj Is Model10.Button Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Button).TabIndex)
        ElseIf TypeOf obj Is Model10.Sub Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Sub).TabIndex)
        ElseIf TypeOf obj Is Model10.Chart Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Chart).TabIndex)
        End If
        Throw New ArgumentException("Object is not an 'Tabular'")
    End Function
    Public Function GetNewDataField(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mSeries.Count - 1
            Dim sr As Model10.Series = CType(mSeries(i), Model10.Series)
            If sr.DataField = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            sr = Nothing
        Loop
        Return str
    End Function

    Public Sub New()
        MyBase.New()
        MyBase.BorderColor = Drawing.Color.Transparent
        MyBase.BorderWidth = 0
        MyBase.Size = New Drawing.Size(200, 100)

        mChartType = OWC11.ChartChartTypeEnum.chChartTypeColumnClustered
        mChildRelation = Nothing
        mDataField = ""
        mPlotColor = Drawing.Color.FromName("lightgray")
        mSeries = New Model10.SeriesCollection(Me)
        '-----------------------------
        Dim sr As New Model10.Series
        sr.DataField = "DataField1"
        mSeries.Add(sr)
        sr = Nothing
        '-----------------------------
        mTabIndex = 0
        mTabStop = True
        mTitle = ""
        mToolTip = ""

        mBottomMargin = 0
        mLeftMargin = 0
        mRightMargin = 0
        mTopMargin = 0
    End Sub
    Public Sub New(ByVal chart As Model10.Chart)
        MyBase.New(chart)

        mChartType = CType(chart.ChartType, OWC11.ChartChartTypeEnum)
        mChildRelation = chart.ChildRelation
        mDataField = chart.DataField
        mPlotColor = chart.PlotColor
        '-------------------------------------
        mSeries = New Model10.SeriesCollection(Me)
        Dim sr As Model10.Series
        For Each sr In chart.Series
            mSeries.Add(New Model10.Series(sr))
        Next : sr = Nothing
        '-------------------------------------
        mTabIndex = chart.TabIndex
        mTabStop = chart.TabStop
        mTitle = chart.Title
        mToolTip = chart.ToolTip

        mBottomMargin = chart.BottomMargin
        mLeftMargin = chart.LeftMargin
        mRightMargin = chart.RightMargin
        mTopMargin = chart.TopMargin
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Chart(Me)
        CType(Clone, Model10.Chart).ChildRelation = Nothing
    End Function

    Public Sub DrawChartF(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        MyBase.Draw(grp, offsetX, offsetY)

        Dim k As Integer
        Select Case MyBase.Border3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = CInt(MyBase.BorderWidth)
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select
        If Not mBitmap Is Nothing Then grp.DrawImage(mBitmap, offsetX + MyBase.Location.X + k + mLeftMargin, offsetY + MyBase.Location.Y + k + mTopMargin)
    End Sub
    Public Sub DrawChart(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        MyBase.Draw(grp, offsetX, offsetY)

        Dim k As Integer
        Select Case MyBase.Border3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = CInt(MyBase.BorderWidth)
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select
        grp.DrawImage(GetChart(Me), offsetX + MyBase.Location.X + k + mLeftMargin, offsetY + MyBase.Location.Y + k + mTopMargin)
    End Sub
    Public Sub FillChart(ByVal dataRow As DataRowView, ByVal print As Boolean)
        If mChildRelation Is Nothing Then Exit Sub

        Dim dataView As DataView = dataRow.CreateChildView(mChildRelation.DataRelation)

        'Build an array for the ProductName field and an array for the ProductSales field.
        Dim aCategories(dataView.Count - 1) As Object
        Dim aSeries As New ArrayList

        Dim i, j As Integer
        Dim obj() As Object

        Dim sr As Model10.Series
        For Each sr In mSeries
            Dim a(dataView.Count - 1) As Object
            aSeries.Add(a)
            a = Nothing
        Next : sr = Nothing

        Dim row As DataRowView : i = 0
        For Each row In dataView
            aCategories(i) = row(mDataField.Trim)

            j = 0
            For Each obj In aSeries
                obj(i) = row(mSeries(j).DataField.Trim)
                j += 1
            Next : obj = Nothing

            i += 1
        Next : row = Nothing

        'Create a new bar chart.
        Dim chartSpace1 As New OWC11.ChartSpace
        Dim cht As OWC11.ChChart = chartSpace1.Charts.Add
        If mTitle = "" Then
            cht.HasTitle = False
        Else
            cht.HasTitle = True
            cht.Title.Caption = mTitle
        End If
        cht.Type = mChartType
        Dim webColConverter As New Web.UI.WebControls.WebColorConverter

        'Add the data to the chart.
        cht.SetData(OWC11.ChartDimensionsEnum.chDimCategories, OWC11.ChartSpecialDataSourcesEnum.chDataLiteral, aCategories)
        i = 0
        For Each obj In aSeries
            Dim series As OWC11.ChSeries
            If i > cht.SeriesCollection.Count - 1 Then
                series = cht.SeriesCollection.Add()
            Else
                series = cht.SeriesCollection(i)
            End If
            '--------------------------
            'Add the data to the chart.
            Dim str As String = dataView.Table.Columns(mSeries(i).DataField.Trim).Caption
            If str = "" Then str = mSeries(i).DataField.Trim
            series.Caption = str
            Try
                series.SetData(OWC11.ChartDimensionsEnum.chDimValues, OWC11.ChartSpecialDataSourcesEnum.chDataLiteral, obj)
            Catch ex As Exception
            End Try
            'Format the chart elements.
            Try
                If Not mSeries(i).Color.Equals(Drawing.Color.Transparent) Then
                    series.Interior.Color = GetColor(webColConverter, mSeries(i).Color)
                End If
            Catch ex As Exception
            End Try
            '--------------------------
            series = Nothing
            i += 1
        Next : obj = Nothing
        aCategories = Nothing
        aSeries = Nothing
        dataView = Nothing

        'Format the chart elements.
        Try
            chartSpace1.Border.Color = OWC11.ChartColorIndexEnum.chColorNone
            If MyBase.BackColor.Equals(Drawing.Color.Transparent) Then
                If MyBase.Section.BackColor.Equals(Drawing.Color.Transparent) Then
                    If print Then
                        chartSpace1.Interior.Color = "white"
                    Else
                        chartSpace1.Interior.Color = "buttonface"
                    End If
                Else
                    chartSpace1.Interior.Color = GetColor(webColConverter, MyBase.Section.BackColor)
                End If
            Else
                chartSpace1.Interior.Color = GetColor(webColConverter, MyBase.BackColor)
            End If
            If mPlotColor.Equals(Drawing.Color.Transparent) Then
                cht.PlotArea.Interior.Color = OWC11.ChartColorIndexEnum.chColorNone
            Else
                cht.PlotArea.Interior.Color = GetColor(webColConverter, mPlotColor)
            End If
        Catch ex As Exception
        End Try

        cht = Nothing
        webColConverter = Nothing

        'Return the new chart in GIF format.
        Dim k As Integer
        Select Case MyBase.Border3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = CInt(MyBase.BorderWidth)
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select
        Dim ms As New IO.MemoryStream(CType(chartSpace1.GetPicture("gif", MyBase.Size.Width - k * 2 - mLeftMargin - mRightMargin, MyBase.Size.Height - k * 2 - mTopMargin - mBottomMargin), Byte()))
        mBitmap = New Drawing.Bitmap(ms)
        ms.Close() : ms = Nothing
        chartSpace1 = Nothing
    End Sub
    Public Shared Function GetChart(ByVal chart As Model10.Chart) As Drawing.Bitmap
        'Build an array for the ProductName field and an array for the ProductSales field.
        Dim aCategories() As Object = {"name1", "name2", "name3"}
        Dim aSeries As New ArrayList

        Dim i, j As Integer
        Dim obj() As Object

        j = 0
        Dim sr As Model10.Series
        For Each sr In chart.Series
            Select Case j
                Case 0 : aSeries.Add(New Object() {34, 89, 21})
                Case 1 : aSeries.Add(New Object() {87, 56, 43})
                Case 2 : aSeries.Add(New Object() {78, 54, 12})
                Case 3 : aSeries.Add(New Object() {98, 77, 65})
                Case 4 : aSeries.Add(New Object() {12, 57, 58})
                Case 5 : aSeries.Add(New Object() {74, 25, 85})
                Case 6 : aSeries.Add(New Object() {32, 54, 54})
                Case 7 : aSeries.Add(New Object() {96, 14, 84})
                Case 8 : aSeries.Add(New Object() {25, 34, 27})
                Case 9 : aSeries.Add(New Object() {90, 65, 34})
                Case Else : aSeries.Add(New Object() {1, 1, 1})
            End Select
            j += 1
        Next : sr = Nothing

        'Create a new bar chart.
        Dim chartSpace1 As New OWC11.ChartSpace
        Dim cht As OWC11.ChChart = chartSpace1.Charts.Add
        If chart.Title = "" Then
            cht.HasTitle = False
        Else
            cht.HasTitle = True
            cht.Title.Caption = chart.Title
        End If
        cht.Type = CType(chart.ChartType, OWC11.ChartChartTypeEnum)
        Dim webColConverter As New Web.UI.WebControls.WebColorConverter

        'Add the data to the chart.
        cht.SetData(OWC11.ChartDimensionsEnum.chDimCategories, OWC11.ChartSpecialDataSourcesEnum.chDataLiteral, aCategories)
        i = 0
        For Each obj In aSeries
            Dim series As OWC11.ChSeries
            If i > cht.SeriesCollection.Count - 1 Then
                series = cht.SeriesCollection.Add()
            Else
                series = cht.SeriesCollection(i)
            End If
            '--------------------------
            'Add the data to the chart.
            series.Caption = chart.Series(i).DataField.Trim
            Try
                series.SetData(OWC11.ChartDimensionsEnum.chDimValues, OWC11.ChartSpecialDataSourcesEnum.chDataLiteral, obj)
            Catch ex As Exception
            End Try
            'Format the chart elements.
            Try
                If Not chart.Series(i).Color.Equals(Drawing.Color.Transparent) Then
                    series.Interior.Color = GetColor(webColConverter, chart.Series(i).Color)
                End If
            Catch ex As Exception
            End Try
            '--------------------------
            series = Nothing
            i += 1
        Next : obj = Nothing
        aCategories = Nothing
        aSeries = Nothing

        'Format the chart elements.
        Try
            chartSpace1.Border.Color = OWC11.ChartColorIndexEnum.chColorNone
            chartSpace1.Interior.Color = GetColor(webColConverter, chart.BackColor)
            cht.PlotArea.Interior.Color = GetColor(webColConverter, chart.PlotColor)
        Catch ex As Exception
        End Try

        cht = Nothing
        webColConverter = Nothing

        'Return the new chart in GIF format.
        Dim k As Integer
        Select Case chart.Border3DStyle
            Case Windows.Forms.Border3DStyle.Flat : k = CInt(chart.BorderWidth)
            Case Windows.Forms.Border3DStyle.RaisedInner, Windows.Forms.Border3DStyle.RaisedOuter, Windows.Forms.Border3DStyle.SunkenInner, Windows.Forms.Border3DStyle.SunkenOuter : k = 1
            Case Windows.Forms.Border3DStyle.Bump, Windows.Forms.Border3DStyle.Etched, Windows.Forms.Border3DStyle.Raised, Windows.Forms.Border3DStyle.Sunken : k = 2
        End Select
        Dim ms As New IO.MemoryStream(CType(chartSpace1.GetPicture("gif", chart.Size.Width - k * 2 - chart.LeftMargin - chart.RightMargin, chart.Size.Height - k * 2 - chart.TopMargin - chart.BottomMargin), Byte()))
        Return New Drawing.Bitmap(ms)
        ms.Close() : ms = Nothing
        chartSpace1 = Nothing
    End Function
    Private Shared Function GetColor(ByVal wcc As Web.UI.WebControls.WebColorConverter, ByVal color As Drawing.Color) As Object
        If color.Equals(Drawing.Color.Transparent) Then
            Return OWC11.ChartColorIndexEnum.chColorNone
        Else
            If color.Equals(Drawing.SystemColors.Control) Then
                Return "buttonface"
            Else
                Return wcc.ConvertToString(color)
            End If
        End If
    End Function

    <ComponentModel.Browsable(False)> Public Property Bitmap() As Drawing.Bitmap
        Get
            Return mBitmap
        End Get
        Set(ByVal Value As Drawing.Bitmap)
            mBitmap = Value
        End Set
    End Property

    Public Property ChartType() As Integer 'OWC11.ChartChartTypeEnum
        Get
            Return mChartType
        End Get
        Set(ByVal Value As Integer) 'OWC11.ChartChartTypeEnum
            mChartType = CType(Value, OWC11.ChartChartTypeEnum)
        End Set
    End Property
    Public Property ChildRelation() As Model10.Relation
        Get
            Return mChildRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mChildRelation = Value
        End Set
    End Property
    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
    Public Property PlotColor() As Drawing.Color
        Get
            Return mPlotColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mPlotColor = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public ReadOnly Property Series() As Model10.SeriesCollection
        Get
            Return mSeries
        End Get
    End Property
    Public Property TabIndex() As Integer
        Get
            Return mTabIndex
        End Get
        Set(ByVal Value As Integer)
            mTabIndex = Value
        End Set
    End Property
    Public Property TabStop() As Boolean
        Get
            Return mTabStop
        End Get
        Set(ByVal Value As Boolean)
            mTabStop = Value
        End Set
    End Property
    Public Property Title() As String
        Get
            Return mTitle
        End Get
        Set(ByVal Value As String)
            mTitle = Value
        End Set
    End Property
    Public Property ToolTip() As String
        Get
            Return mToolTip
        End Get
        Set(ByVal Value As String)
            mToolTip = Value
        End Set
    End Property

    Public Property BottomMargin() As Integer
        Get
            Return mBottomMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mBottomMargin = Value
        End Set
    End Property
    Public Property LeftMargin() As Integer
        Get
            Return mLeftMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mLeftMargin = Value
        End Set
    End Property
    Public Property RightMargin() As Integer
        Get
            Return mRightMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mRightMargin = Value
        End Set
    End Property
    Public Property TopMargin() As Integer
        Get
            Return mTopMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mTopMargin = Value
        End Set
    End Property
End Class

'<Serializable()> Public Class DateTime
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal dateTime As Model10.DateTime)
'        MyBase.New(dateTime)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.DateTime(Me)
'    End Function
'End Class
'<Serializable()> Public Class Title
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal title As Model10.Title)
'        MyBase.New(title)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.Title(Me)
'    End Function
'End Class
'<Serializable()> Public Class Path
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal path As Model10.Path)
'        MyBase.New(path)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.Path(Me)
'    End Function
'End Class
'<Serializable()> Public Class PageCount
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal pageCount As Model10.PageCount)
'        MyBase.New(pageCount)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.PageCount(Me)
'    End Function
'End Class
'<Serializable()> Public Class PageNumber
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal pageNumber As Model10.PageNumber)
'        MyBase.New(pageNumber)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.PageNumber(Me)
'    End Function
'End Class
'<Serializable()> Public Class RecordCount
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal recordCount As Model10.RecordCount)
'        MyBase.New(recordCount)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.RecordCount(Me)
'    End Function
'End Class
'<Serializable()> Public Class RecordNumber
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal recordNumber As Model10.RecordNumber)
'        MyBase.New(recordNumber)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.RecordNumber(Me)
'    End Function
'End Class
'<Serializable()> Public Class UserId
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal userId As Model10.UserId)
'        MyBase.New(userId)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.UserId(Me)
'    End Function
'End Class
'<Serializable()> Public Class UserFullName
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal userFullName As Model10.UserFullName)
'        MyBase.New(userFullName)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.UserFullName(Me)
'    End Function
'End Class
'<Serializable()> Public Class UserName
'    Inherits Model10.Label

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'    End Sub

'    Public Sub New()
'        MyBase.New()
'    End Sub
'    Public Sub New(ByVal userName As Model10.UserName)
'        MyBase.New(userName)
'    End Sub
'    Public Overrides Function Clone() As Model10.Member
'        Clone = New Model10.UserName(Me)
'    End Function
'End Class

<Serializable()> Public Class Button
    Inherits Model10.Member
    Implements IComparable

    <NonSerialized()> Private mKey As String

    Private mButtonStyle As Model10.ButtonStyle
    Private mFont As Drawing.Font
    Private mForeColor As Drawing.Color
    Private mImage As Drawing.Image
    Private mTabIndex As Integer
    Private mTabStop As Boolean
    Private mText As String
    Private mToolTip As String

    Private mAlignment As Windows.Forms.HorizontalAlignment
    Private mHotkeyPrefix As Drawing.Text.HotkeyPrefix
    Private mNoWrap As Boolean
    Private mRightToLeft As Windows.Forms.RightToLeft
    Private mTrimming As Drawing.StringTrimming

    Private mBottomMargin As Integer
    Private mLeftMargin As Integer
    Private mRightMargin As Integer
    Private mTopMargin As Integer

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mButtonStyle = CType(info.GetByte("ButtonStyle"), Model10.ButtonStyle)
        mFont = CType(info.GetValue("Font", GetType(Drawing.Font)), Drawing.Font)
        mForeColor = CType(info.GetValue("ForeColor", GetType(Drawing.Color)), Drawing.Color)
        mImage = CType(info.GetValue("Image", GetType(Drawing.Image)), Drawing.Image)
        mTabIndex = info.GetInt32("TabIndex")
        mTabStop = info.GetBoolean("TabStop")
        mText = info.GetString("Text")
        mToolTip = info.GetString("ToolTip")

        mAlignment = CType(info.GetInt32("Alignment"), Windows.Forms.HorizontalAlignment)
        mHotkeyPrefix = CType(info.GetInt32("HotkeyPrefix"), Drawing.Text.HotkeyPrefix)
        mNoWrap = info.GetBoolean("NoWrap")
        mRightToLeft = CType(info.GetInt32("RightToLeft"), Windows.Forms.RightToLeft)
        mTrimming = CType(info.GetInt32("Trimming"), Drawing.StringTrimming)

        mBottomMargin = info.GetInt32("BottomMargin")
        mLeftMargin = info.GetInt32("LeftMargin")
        mRightMargin = info.GetInt32("RightMargin")
        mTopMargin = info.GetInt32("TopMargin")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("ButtonStyle", mButtonStyle)
        info.AddValue("Font", mFont)
        info.AddValue("ForeColor", mForeColor)
        info.AddValue("Image", mImage)
        info.AddValue("TabIndex", mTabIndex)
        info.AddValue("TabStop", mTabStop)
        info.AddValue("Text", mText)
        info.AddValue("ToolTip", mToolTip)

        info.AddValue("Alignment", mAlignment)
        info.AddValue("HotkeyPrefix", mHotkeyPrefix)
        info.AddValue("NoWrap", mNoWrap)
        info.AddValue("RightToLeft", mRightToLeft)
        info.AddValue("Trimming", mTrimming)

        info.AddValue("BottomMargin", mBottomMargin)
        info.AddValue("LeftMargin", mLeftMargin)
        info.AddValue("RightMargin", mRightMargin)
        info.AddValue("TopMargin", mTopMargin)
    End Sub

    Public Function CompareTo(ByVal obj As Object) As Integer Implements IComparable.CompareTo
        If TypeOf obj Is Model10.Label Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Label).TabIndex)
        ElseIf TypeOf obj Is Model10.PictureBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.PictureBox).TabIndex)
        ElseIf TypeOf obj Is Model10.CheckBox Then
            Return mTabIndex.CompareTo(CType(obj, Model10.CheckBox).TabIndex)
        ElseIf TypeOf obj Is Model10.Button Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Button).TabIndex)
        ElseIf TypeOf obj Is Model10.Sub Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Sub).TabIndex)
        ElseIf TypeOf obj Is Model10.Chart Then
            Return mTabIndex.CompareTo(CType(obj, Model10.Chart).TabIndex)
        End If
        Throw New ArgumentException("Object is not an 'Tabular'")
    End Function
    Public Function GetStringFormat(ByVal rtl As Windows.Forms.RightToLeft) As Drawing.StringFormat
        Dim sf As New Drawing.StringFormat

        Dim rl As Windows.Forms.RightToLeft
        Select Case mRightToLeft
            Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
            Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
            Case Windows.Forms.RightToLeft.Inherit
                Select Case MyBase.Section.Table.Presentation.RightToLeft
                    Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
                    Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
                    Case Windows.Forms.RightToLeft.Inherit
                        Select Case rtl
                            Case Windows.Forms.RightToLeft.No : rl = Windows.Forms.RightToLeft.No
                            Case Windows.Forms.RightToLeft.Yes : rl = Windows.Forms.RightToLeft.Yes
                            Case Windows.Forms.RightToLeft.Inherit
                        End Select
                End Select
        End Select
        If rl = Windows.Forms.RightToLeft.Yes Then
            Select Case mAlignment
                Case Windows.Forms.HorizontalAlignment.Left : sf.Alignment = Drawing.StringAlignment.Far
                Case Windows.Forms.HorizontalAlignment.Right : sf.Alignment = Drawing.StringAlignment.Near
                Case Windows.Forms.HorizontalAlignment.Center : sf.Alignment = Drawing.StringAlignment.Center
            End Select
            sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.DirectionRightToLeft
        Else
            Select Case mAlignment
                Case Windows.Forms.HorizontalAlignment.Left : sf.Alignment = Drawing.StringAlignment.Near
                Case Windows.Forms.HorizontalAlignment.Right : sf.Alignment = Drawing.StringAlignment.Far
                Case Windows.Forms.HorizontalAlignment.Center : sf.Alignment = Drawing.StringAlignment.Center
            End Select
        End If
        If mNoWrap Then
            sf.FormatFlags = sf.FormatFlags Or Drawing.StringFormatFlags.NoWrap
        End If
        sf.HotkeyPrefix = mHotkeyPrefix
        sf.LineAlignment = Drawing.StringAlignment.Center
        sf.Trimming = mTrimming

        Return sf
    End Function

    Public Sub Draw(ByVal grp As Drawing.Graphics, ByVal text As String, ByVal offsetX As Integer, ByVal offsetY As Integer, ByVal rtl As Windows.Forms.RightToLeft, ByVal fColor As Drawing.Color)
        Dim r As New Drawing.Rectangle(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, r)

        Dim k As Integer
        Select Case mButtonStyle
            Case ButtonStyle.Normal
                k = 2
                Windows.Forms.ControlPaint.DrawBorder(grp, r, Drawing.SystemColors.Control, Windows.Forms.ButtonBorderStyle.Outset)
            Case ButtonStyle.Flat
                k = 1
                Windows.Forms.ControlPaint.DrawBorder(grp, r, Drawing.SystemColors.ControlDark, Windows.Forms.ButtonBorderStyle.Solid)
        End Select
        Select Case mButtonStyle
            Case ButtonStyle.Normal, ButtonStyle.Flat, ButtonStyle.Link
                r.Offset(mLeftMargin + k, mTopMargin + k)
                r.Width += -mLeftMargin - mRightMargin - 2 * k
                r.Height += -mTopMargin - mBottomMargin - 2 * k
                If mButtonStyle = ButtonStyle.Normal Or mButtonStyle = ButtonStyle.Flat Then r.Offset(-1, -1)
                If Not fColor.Equals(Drawing.Color.Empty) And Not fColor.Equals(Drawing.Color.Transparent) And MyBase.BackColor.Equals(Drawing.Color.Transparent) Then
                    b.Color = fColor
                Else
                    b.Color = mForeColor
                End If
                grp.DrawString(text, mFont, b, Drawing.RectangleF.op_Implicit(r), GetStringFormat(rtl))
            Case ButtonStyle.Image
                If Not mImage Is Nothing Then grp.DrawImage(mImage, r)
            Case ButtonStyle.PlusMines
                Dim x As Integer = r.X + (r.Width - 9) \ 2
                Dim y As Integer = r.Y + (r.Height - 9) \ 2
                grp.FillRectangle(Drawing.Brushes.White, New Drawing.Rectangle(x, y, 8, 8))
                grp.DrawRectangle(Drawing.SystemPens.ControlDark, New Drawing.Rectangle(x, y, 8, 8))
                grp.DrawLine(Drawing.Pens.Black, New Drawing.Point(x + 4, y + 2), New Drawing.Point(x + 4, y + 2 + 4))
                grp.DrawLine(Drawing.Pens.Black, New Drawing.Point(x + 2, y + 4), New Drawing.Point(x + 2 + 4, y + 4))
        End Select

        b.Dispose() : b = Nothing
    End Sub
    Public Sub DrawF(ByVal grp As Drawing.Graphics, ByVal offsetX As Integer, ByVal offsetY As Integer)
        Dim r As New Drawing.RectangleF(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, r)

        Select Case mButtonStyle
            Case ButtonStyle.Normal, ButtonStyle.Flat, ButtonStyle.Link
            Case ButtonStyle.Image
            Case ButtonStyle.PlusMines
                Dim x As Single = r.X + (r.Width - 9) / 2
                Dim y As Single = r.Y + (r.Height - 9) / 2
                grp.FillRectangle(Drawing.Brushes.White, x, y, 9, 9)
                grp.DrawRectangle(Drawing.SystemPens.ControlDark, x, y, 9, 9)
                grp.DrawLine(Drawing.Pens.Black, New Drawing.PointF(CSng(x + 4.5), y + 2), New Drawing.PointF(CSng(x + 4.5), y + 2 + 5))
                grp.DrawLine(Drawing.Pens.Black, New Drawing.PointF(x + 2, CSng(y + 4.5)), New Drawing.PointF(x + 2 + 5, CSng(y + 4.5)))
        End Select

        b.Dispose() : b = Nothing
    End Sub
    Public Sub DrawPressed(ByVal grp As Drawing.Graphics, ByVal text As String, ByVal offsetX As Integer, ByVal offsetY As Integer, ByVal rtl As Windows.Forms.RightToLeft, ByVal fColor As Drawing.Color)
        Dim r As New Drawing.Rectangle(MyBase.Location.X + offsetX, MyBase.Location.Y + offsetY, MyBase.Size.Width, MyBase.Size.Height)
        Dim b As New Drawing.SolidBrush(MyBase.BackColor)
        grp.FillRectangle(b, r)

        Dim k As Integer
        Select Case mButtonStyle
            Case ButtonStyle.Normal
                k = 2
                Windows.Forms.ControlPaint.DrawBorder(grp, r, Drawing.SystemColors.Control, Windows.Forms.ButtonBorderStyle.Inset)
            Case ButtonStyle.Flat
                k = 1
                Windows.Forms.ControlPaint.DrawBorder(grp, r, Drawing.SystemColors.ControlDark, Windows.Forms.ButtonBorderStyle.Solid)
        End Select
        Select Case mButtonStyle
            Case ButtonStyle.Normal, ButtonStyle.Flat, ButtonStyle.Link
                r.Offset(mLeftMargin + k, mTopMargin + k)
                r.Width += -mLeftMargin - mRightMargin - 2 * k
                r.Height += -mTopMargin - mBottomMargin - 2 * k
                'If mButtonStyle = ButtonStyle.Normal Or mButtonStyle = ButtonStyle.Flat Then r.Offset(-1, -1)
                If Not fColor.Equals(Drawing.Color.Empty) And Not fColor.Equals(Drawing.Color.Transparent) And MyBase.BackColor.Equals(Drawing.Color.Transparent) Then
                    b.Color = fColor
                Else
                    b.Color = mForeColor
                End If
                grp.DrawString(text, mFont, b, Drawing.RectangleF.op_Implicit(r), GetStringFormat(rtl))
            Case ButtonStyle.Image
                If Not mImage Is Nothing Then grp.DrawImage(mImage, r)
            Case ButtonStyle.PlusMines
                Dim x As Integer = r.X + (r.Width - 9) \ 2
                Dim y As Integer = r.Y + (r.Height - 9) \ 2
                grp.FillRectangle(Drawing.Brushes.White, New Drawing.Rectangle(x, y, 8, 8))
                grp.DrawRectangle(Drawing.SystemPens.ControlDark, New Drawing.Rectangle(x, y, 8, 8))
                'grp.DrawLine(Drawing.Pens.Black, New Drawing.Point(x + 4, y + 2), New Drawing.Point(x + 4, y + 2 + 4))
                grp.DrawLine(Drawing.Pens.Black, New Drawing.Point(x + 2, y + 4), New Drawing.Point(x + 2 + 4, y + 4))
        End Select

        b.Dispose() : b = Nothing
    End Sub

    Public Sub New()
        MyBase.New()
        MyBase.BackColor = Drawing.SystemColors.Control
        MyBase.Size = New Drawing.Size(64, 16)

        mKey = ""

        mButtonStyle = ButtonStyle.Normal
        mFont = New Drawing.Font("Tahoma", 8) 'Windows.Forms.Control.DefaultFont
        mForeColor = Drawing.SystemColors.ControlText 'Windows.Forms.Control.DefaultForeColor
        mImage = Nothing
        mTabIndex = 0
        mTabStop = True
        mText = ""
        mToolTip = ""

        mBottomMargin = 0
        mLeftMargin = 0
        mRightMargin = 0
        mTopMargin = 0

        mAlignment = Windows.Forms.HorizontalAlignment.Center
        mHotkeyPrefix = Drawing.Text.HotkeyPrefix.None
        mNoWrap = False
        mRightToLeft = Windows.Forms.RightToLeft.Inherit
        mTrimming = Drawing.StringTrimming.None
    End Sub
    Public Sub New(ByVal button As Model10.Button)
        MyBase.New(button)

        mKey = button.Key

        mButtonStyle = button.ButtonStyle
        mFont = CType(button.Font.Clone, Drawing.Font)
        mForeColor = button.ForeColor
        mImage = button.Image
        mTabIndex = button.TabIndex
        mTabStop = button.TabStop
        mText = button.Text
        mToolTip = button.ToolTip

        mBottomMargin = button.BottomMargin
        mLeftMargin = button.LeftMargin
        mRightMargin = button.RightMargin
        mTopMargin = button.TopMargin

        mAlignment = button.Alignment
        mHotkeyPrefix = button.HotkeyPrefix
        mNoWrap = button.NoWrap
        mRightToLeft = button.RightToLeft
        mTrimming = button.Trimming
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Button(Me)
    End Function

    <ComponentModel.Browsable(False)> Public Property Key() As String
        Get
            Return mKey
        End Get
        Set(ByVal Value As String)
            mKey = Value
        End Set
    End Property

    Public Property ButtonStyle() As Model10.ButtonStyle
        Get
            Return mButtonStyle
        End Get
        Set(ByVal Value As Model10.ButtonStyle)
            mButtonStyle = Value
        End Set
    End Property
    Public Property Font() As Drawing.Font
        Get
            Return mFont
        End Get
        Set(ByVal Value As Drawing.Font)
            mFont = Value
        End Set
    End Property
    Public Property ForeColor() As Drawing.Color
        Get
            Return mForeColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mForeColor = Value
        End Set
    End Property
    Public Property Image() As Drawing.Image
        Get
            Return mImage
        End Get
        Set(ByVal Value As Drawing.Image)
            mImage = Value
        End Set
    End Property
    Public Property TabIndex() As Integer
        Get
            Return mTabIndex
        End Get
        Set(ByVal Value As Integer)
            mTabIndex = Value
        End Set
    End Property
    Public Property TabStop() As Boolean
        Get
            Return mTabStop
        End Get
        Set(ByVal Value As Boolean)
            mTabStop = Value
        End Set
    End Property
    Public Property Text() As String
        Get
            Return mText
        End Get
        Set(ByVal Value As String)
            mText = Value
        End Set
    End Property
    Public Property ToolTip() As String
        Get
            Return mToolTip
        End Get
        Set(ByVal Value As String)
            mToolTip = Value
        End Set
    End Property

    Public Property BottomMargin() As Integer
        Get
            Return mBottomMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mBottomMargin = Value
        End Set
    End Property
    Public Property LeftMargin() As Integer
        Get
            Return mLeftMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mLeftMargin = Value
        End Set
    End Property
    Public Property RightMargin() As Integer
        Get
            Return mRightMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mRightMargin = Value
        End Set
    End Property
    Public Property TopMargin() As Integer
        Get
            Return mTopMargin
        End Get
        Set(ByVal Value As Integer)
            If Value < 0 Then Value = 0
            mTopMargin = Value
        End Set
    End Property

    Public Property Alignment() As Windows.Forms.HorizontalAlignment
        Get
            Return mAlignment
        End Get
        Set(ByVal Value As Windows.Forms.HorizontalAlignment)
            mAlignment = Value
        End Set
    End Property
    Public Property HotkeyPrefix() As Drawing.Text.HotkeyPrefix
        Get
            Return mHotkeyPrefix
        End Get
        Set(ByVal Value As Drawing.Text.HotkeyPrefix)
            mHotkeyPrefix = Value
        End Set
    End Property
    Public Property NoWrap() As Boolean
        Get
            Return mNoWrap
        End Get
        Set(ByVal Value As Boolean)
            mNoWrap = Value
        End Set
    End Property
    Public Property RightToLeft() As Windows.Forms.RightToLeft
        Get
            Return mRightToLeft
        End Get
        Set(ByVal Value As Windows.Forms.RightToLeft)
            mRightToLeft = Value
        End Set
    End Property
    Public Property Trimming() As Drawing.StringTrimming
        Get
            Return mTrimming
        End Get
        Set(ByVal Value As Drawing.StringTrimming)
            mTrimming = Value
        End Set
    End Property
End Class
<Serializable()> Public Class [Select]
    Inherits Model10.Button

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

    End Sub

    Public Sub New()
        MyBase.New()
    End Sub
    Public Sub New(ByVal [select] As Model10.Select)
        MyBase.New([select])
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Select(Me)
    End Function
End Class
<Serializable()> Public Class Edit
    Inherits Model10.Button

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

    End Sub

    Public Sub New()
        MyBase.New()
    End Sub
    Public Sub New(ByVal edit As Model10.Edit)
        MyBase.New(edit)
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Edit(Me)
    End Function
End Class
<Serializable()> Public Class Update
    Inherits Model10.Button

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

    End Sub

    Public Sub New()
        MyBase.New()
    End Sub
    Public Sub New(ByVal update As Model10.Update)
        MyBase.New(update)
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Update(Me)
    End Function
End Class
<Serializable()> Public Class Cancel
    Inherits Model10.Button

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

    End Sub

    Public Sub New()
        MyBase.New()
    End Sub
    Public Sub New(ByVal cancel As Model10.Cancel)
        MyBase.New(cancel)
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Cancel(Me)
    End Function
End Class
<Serializable()> Public Class Delete
    Inherits Model10.Button

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

    End Sub

    Public Sub New()
        MyBase.New()
    End Sub
    Public Sub New(ByVal delete As Model10.Delete)
        MyBase.New(delete)
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Delete(Me)
    End Function
End Class
<Serializable()> Public Class Browse
    Inherits Model10.Button

    Private mContentField As String
    Private mFilter As String
    Private mPathField As String
    Private mPathType As Model10.PathType
    Private mShowEditor As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mContentField = info.GetString("ContentField")
        mFilter = info.GetString("Filter")
        mPathField = info.GetString("PathField")
        mPathType = CType(info.GetByte("PathType"), Model10.PathType)
        mShowEditor = info.GetBoolean("ShowEditor")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("ContentField", mContentField)
        info.AddValue("Filter", mFilter)
        info.AddValue("PathField", mPathField)
        info.AddValue("PathType", mPathType)
        info.AddValue("ShowEditor", mShowEditor)
    End Sub

    Public Sub New()
        MyBase.New()

        mContentField = ""
        Dim s1 As String = "*.bmp;*.gif;*.jpg;*.jpeg;*.png;*.ico"
        Dim s2 As String = "*.emf;*.wmf"
        mFilter = "All image files(" & s1 & ";" & s2 & ")|" & s1 & ";" & s2 & "|Bitmap files(" & s1 & ")|" & s1 & "|Metafiles(" & s2 & ")|" & s2 & ""
        mPathField = ""
        mPathType = PathType.Complete
        mShowEditor = True
    End Sub
    Public Sub New(ByVal browse As Model10.Browse)
        MyBase.New(browse)

        mContentField = browse.ContentField
        mFilter = browse.Filter
        mPathField = browse.PathField
        mPathType = browse.PathType
        mShowEditor = browse.ShowEditor
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Browse(Me)
    End Function

    Public Property ContentField() As String
        Get
            Return mContentField
        End Get
        Set(ByVal Value As String)
            mContentField = Value
        End Set
    End Property
    Public Property Filter() As String
        Get
            Return mFilter
        End Get
        Set(ByVal Value As String)
            mFilter = Value
        End Set
    End Property
    Public Property PathField() As String
        Get
            Return mPathField
        End Get
        Set(ByVal Value As String)
            mPathField = Value
        End Set
    End Property
    Public Property PathType() As Model10.PathType
        Get
            Return mPathType
        End Get
        Set(ByVal Value As Model10.PathType)
            mPathType = Value
        End Set
    End Property
    Public Property ShowEditor() As Boolean
        Get
            Return mShowEditor
        End Get
        Set(ByVal Value As Boolean)
            mShowEditor = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Computer
    Inherits Model10.Button

    Private mCompute As Model10.Compute

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mCompute = CType(info.GetValue("Compute", GetType(Model10.Compute)), Model10.Compute)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("Compute", mCompute)
    End Sub

    Public Sub New()
        MyBase.New()

        mCompute = Nothing
    End Sub
    Public Sub New(ByVal computer As Model10.Computer)
        MyBase.New(computer)

        mCompute = computer.Compute
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Computer(Me)
        CType(Clone, Model10.Computer).Compute = Nothing
    End Function

    Public Property Compute() As Model10.Compute
        Get
            Return mCompute
        End Get
        Set(ByVal Value As Model10.Compute)
            mCompute = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Expand
    Inherits Model10.Button

    Private mChildRelation As Model10.Relation
    Private mParentRelation As Model10.Relation

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mChildRelation = CType(info.GetValue("ChildRelation", GetType(Model10.Relation)), Model10.Relation)
        mParentRelation = CType(info.GetValue("ParentRelation", GetType(Model10.Relation)), Model10.Relation)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("ChildRelation", mChildRelation)
        info.AddValue("ParentRelation", mParentRelation)
    End Sub

    Public Sub New()
        MyBase.New()

        mChildRelation = Nothing
        mParentRelation = Nothing
    End Sub
    Public Sub New(ByVal expand As Model10.Expand)
        MyBase.New(expand)

        mChildRelation = expand.ChildRelation
        mParentRelation = expand.ParentRelation
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Expand(Me)
        CType(Clone, Model10.Expand).ChildRelation = Nothing
        CType(Clone, Model10.Expand).ParentRelation = Nothing
    End Function

    Public Property ChildRelation() As Model10.Relation
        Get
            Return mChildRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mChildRelation = Value
        End Set
    End Property
    Public Property ParentRelation() As Model10.Relation
        Get
            Return mParentRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mParentRelation = Value
        End Set
    End Property
End Class
<Serializable()> Public Class [Next]
    Inherits Model10.Button

    Private mTable As Model10.Table

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mTable = CType(info.GetValue("Table", GetType(Model10.Table)), Model10.Table)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("Table", mTable)
    End Sub

    Public Sub New()
        MyBase.New()

        mTable = Nothing
    End Sub
    Public Sub New(ByVal [next] As Model10.Next)
        MyBase.New([next])

        mTable = [next].Table
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Next(Me)
        CType(Clone, Model10.Next).Table = Nothing
    End Function

    Public Property Table() As Model10.Table
        Get
            Return mTable
        End Get
        Set(ByVal Value As Model10.Table)
            mTable = Value
        End Set
    End Property
End Class
<Serializable()> Public Class MultiSelect
    Inherits Model10.Button

    Private mDataTable As Model10.Table
    Private mDestinationField As String
    Private mSelectionField As String
    Private mSourceField As String
    Private mStartupTable As Model10.Table

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mDataTable = CType(info.GetValue("DataTable", GetType(Model10.Table)), Model10.Table)
        mDestinationField = info.GetString("DestinationField")
        mSelectionField = info.GetString("SelectionField")
        mSourceField = info.GetString("SourceField")
        mStartupTable = CType(info.GetValue("StartupTable", GetType(Model10.Table)), Model10.Table)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("DataTable", mDataTable)
        info.AddValue("DestinationField", mDestinationField)
        info.AddValue("SelectionField", mSelectionField)
        info.AddValue("SourceField", mSourceField)
        info.AddValue("StartupTable", mStartupTable)
    End Sub

    Public Sub New()
        MyBase.New()

        mDataTable = Nothing
        mDestinationField = ""
        mSelectionField = ""
        mSourceField = ""
        mStartupTable = Nothing
    End Sub
    Public Sub New(ByVal multiSelect As Model10.MultiSelect)
        MyBase.New(multiSelect)

        mDataTable = multiSelect.DataTable
        mDestinationField = multiSelect.DestinationField
        mSelectionField = multiSelect.SelectionField
        mSourceField = multiSelect.SourceField
        mStartupTable = multiSelect.StartupTable
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.MultiSelect(Me)
        CType(Clone, Model10.MultiSelect).DataTable = Nothing
        CType(Clone, Model10.MultiSelect).StartupTable = Nothing
    End Function

    Public Property DataTable() As Model10.Table
        Get
            Return mDataTable
        End Get
        Set(ByVal Value As Model10.Table)
            mDataTable = Value
        End Set
    End Property
    Public Property DestinationField() As String
        Get
            Return mDestinationField
        End Get
        Set(ByVal Value As String)
            mDestinationField = Value
        End Set
    End Property
    Public Property SelectionField() As String
        Get
            Return mSelectionField
        End Get
        Set(ByVal Value As String)
            mSelectionField = Value
        End Set
    End Property
    Public Property SourceField() As String
        Get
            Return mSourceField
        End Get
        Set(ByVal Value As String)
            mSourceField = Value
        End Set
    End Property
    Public Property StartupTable() As Model10.Table
        Get
            Return mStartupTable
        End Get
        Set(ByVal Value As Model10.Table)
            mStartupTable = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Drilldown
    Inherits Model10.Button

    Private mChildRelation As Model10.Relation

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mChildRelation = CType(info.GetValue("ChildRelation", GetType(Model10.Relation)), Model10.Relation)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("ChildRelation", mChildRelation)
    End Sub

    Public Sub New()
        MyBase.New()

        mChildRelation = Nothing
    End Sub
    Public Sub New(ByVal drilldown As Model10.Drilldown)
        MyBase.New(drilldown)

        mChildRelation = drilldown.ChildRelation
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Drilldown(Me)
        CType(Clone, Model10.Drilldown).ChildRelation = Nothing
    End Function

    Public Property ChildRelation() As Model10.Relation
        Get
            Return mChildRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mChildRelation = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Lookup
    Inherits Model10.Button

    Private mParentRelation As Model10.Relation
    Private mStartupTable As Model10.Table

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mParentRelation = CType(info.GetValue("ParentRelation", GetType(Model10.Relation)), Model10.Relation)
        mStartupTable = CType(info.GetValue("StartupTable", GetType(Model10.Table)), Model10.Table)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("ParentRelation", mParentRelation)
        info.AddValue("StartupTable", mStartupTable)
    End Sub

    Public Sub New()
        MyBase.New()

        mParentRelation = Nothing
        mStartupTable = Nothing
    End Sub
    Public Sub New(ByVal lookup As Model10.Lookup)
        MyBase.New(lookup)

        mParentRelation = lookup.ParentRelation
        mStartupTable = lookup.StartupTable
    End Sub
    Public Overrides Function Clone() As Model10.Member
        Clone = New Model10.Lookup(Me)
        CType(Clone, Model10.Lookup).ParentRelation = Nothing
        CType(Clone, Model10.Lookup).StartupTable = Nothing
    End Function

    Public Property ParentRelation() As Model10.Relation
        Get
            Return mParentRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mParentRelation = Value
        End Set
    End Property
    Public Property StartupTable() As Model10.Table
        Get
            Return mStartupTable
        End Get
        Set(ByVal Value As Model10.Table)
            mStartupTable = Value
        End Set
    End Property
End Class

<Serializable()> Public MustInherit Class Section
    Implements ISerializable

    <NonSerialized()> Private mBitmap As Drawing.Bitmap

    Private mBackColor As Drawing.Color
    Private mHeight As Integer
    Private mLocation As Drawing.Point
    Private ReadOnly mMembers As Model10.MemberCollection
    Private mTable As Model10.Table

    Private Const MIN_SIZE As Byte = 1

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mBackColor = CType(info.GetValue("BackColor", GetType(Drawing.Color)), Drawing.Color)
        mHeight = info.GetInt32("Height")
        mLocation = CType(info.GetValue("Location", GetType(Drawing.Point)), Drawing.Point)
        mMembers = CType(info.GetValue("Members", GetType(Model10.MemberCollection)), Model10.MemberCollection)
        mTable = CType(info.GetValue("Table", GetType(Model10.Table)), Model10.Table)
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("BackColor", mBackColor)
        info.AddValue("Height", mHeight)
        info.AddValue("Location", mLocation)
        info.AddValue("Members", mMembers)
        info.AddValue("Table", mTable)
    End Sub

    Public Function GetMinHeight() As Integer
        GetMinHeight = MIN_SIZE

        Dim mem As Model10.Member
        For Each mem In mMembers
            If GetMinHeight < mem.Location.Y + mem.Size.Height Then GetMinHeight = mem.Location.Y + mem.Size.Height
        Next : mem = Nothing
    End Function
    Public Function GetName() As String
        If Not mTable Is Nothing Then
            If Me Is mTable.PageHeader Then Return "PageHeader"
            If Me Is mTable.GroupHeader Then Return "GroupHeader"
            If Me Is mTable.Detail Then Return "Detail"
            If Me Is mTable.GroupFooter Then Return "GroupFooter"
            If Me Is mTable.PageFooter Then Return "PageFooter"
        End If
    End Function
    Public Function GetNewMemberName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mMembers.Count - 1
            Dim mem As Model10.Member = CType(mMembers(i), Model10.Member)
            If mem.Name = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            mem = Nothing
        Loop
        Return str
    End Function
    Public Function GetMembers() As Model10.Member()
        Dim mems() As Model10.Member

        Dim i As Integer

        Dim mem As Model10.Member
        For Each mem In mMembers
            Dim t As Type = mem.GetType
            If t Is GetType(Model10.Label) Then
                Dim lbl As Model10.Label = CType(mem, Model10.Label)
                If lbl.NavigateUrl.Trim <> "" Then
                    ReDim Preserve mems(i)
                    mems(i) = mem
                    i += 1
                End If
                lbl = Nothing
            ElseIf t Is GetType(Model10.TextBox) Then
                ReDim Preserve mems(i)
                mems(i) = mem
                i += 1
            ElseIf t Is GetType(Model10.PictureBox) Then
                Dim pic As Model10.PictureBox = CType(mem, Model10.PictureBox)
                If pic.NavigateUrl.Trim <> "" Then
                    ReDim Preserve mems(i)
                    mems(i) = mem
                    i += 1
                End If
                pic = Nothing
            ElseIf t Is GetType(Model10.CheckBox) Then
                ReDim Preserve mems(i)
                mems(i) = mem
                i += 1
            ElseIf TypeOf mem Is Model10.Button Then
                ReDim Preserve mems(i)
                mems(i) = mem
                i += 1
            End If
            t = Nothing
        Next : mem = Nothing

        If Not mems Is Nothing Then Array.Sort(mems)
        Return mems
    End Function

    Public Sub New()
        MyBase.New()

        mBitmap = Nothing

        mBackColor = Drawing.SystemColors.Control
        mHeight = 32
        mLocation = New Drawing.Point
        mMembers = New Model10.MemberCollection(Me)
        mTable = Nothing
    End Sub
    Public Sub New(ByVal section As Model10.Section, ByVal bkColor As Drawing.Color)
        MyBase.New()

        mBitmap = Nothing

        mBackColor = bkColor 'section.BackColor
        mHeight = section.Height
        mLocation = New Drawing.Point
        mMembers = New Model10.MemberCollection(Me)
        mTable = section.Table
    End Sub

    <ComponentModel.Browsable(False)> Public Property Bitmap() As Drawing.Bitmap
        Get
            Return mBitmap
        End Get
        Set(ByVal Value As Drawing.Bitmap)
            mBitmap = Value
        End Set
    End Property

    Public Property BackColor() As Drawing.Color
        Get
            Return mBackColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mBackColor = Value
        End Set
    End Property
    Public Property Height() As Integer
        Get
            Return mHeight
        End Get
        Set(ByVal Value As Integer)
            Dim h As Integer = GetMinHeight()
            If Value < h Then Value = h
            mHeight = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Location() As Drawing.Point
        Get
            Return mLocation
        End Get
        Set(ByVal Value As Drawing.Point)
            mLocation = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public ReadOnly Property Members() As Model10.MemberCollection
        Get
            Return mMembers
        End Get
    End Property
    <ComponentModel.Browsable(False)> Public Property Table() As Model10.Table
        Get
            Return mTable
        End Get
        Set(ByVal Value As Model10.Table)
            mTable = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Pager
    Inherits Section

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

    End Sub

    Public Sub New()
        MyBase.New()

    End Sub
    Public Sub New(ByVal pager As Model10.Pager, ByVal bkColor As Drawing.Color)
        MyBase.New(pager, bkColor)

    End Sub
    Public Function Clone() As Model10.Pager
        Clone = New Model10.Pager(Me, MyBase.BackColor)

        Dim mem As Model10.Member
        For Each mem In Me.Members
            Clone.Members.Add(mem.Clone, 0, 0)
        Next : mem = Nothing
    End Function
End Class
<Serializable()> Public Class Grouper
    Inherits Section

    Private mKeepTogether As Boolean
    Private mNewPage As Model10.NewPageMode
    Private mPrintBackground As Boolean
    Private mRepeatOnEveryPage As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mKeepTogether = info.GetBoolean("KeepTogether")
        mNewPage = CType(info.GetByte("NewPage"), Model10.NewPageMode)
        mPrintBackground = info.GetBoolean("PrintBackground")
        mRepeatOnEveryPage = info.GetBoolean("RepeatOnEveryPage")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("KeepTogether", mKeepTogether)
        info.AddValue("NewPage", mNewPage)
        info.AddValue("PrintBackground", mPrintBackground)
        info.AddValue("RepeatOnEveryPage", mRepeatOnEveryPage)
    End Sub

    Public Sub New()
        MyBase.New()

        mKeepTogether = False
        mNewPage = NewPageMode.None
        mPrintBackground = False
        mRepeatOnEveryPage = False
    End Sub
    Public Sub New(ByVal grouper As Model10.Grouper, ByVal bkColor As Drawing.Color)
        MyBase.New(grouper, bkColor)

        mKeepTogether = grouper.KeepTogether
        mNewPage = grouper.NewPage
        mPrintBackground = grouper.PrintBackground
        mRepeatOnEveryPage = grouper.RepeatOnEveryPage
    End Sub
    Public Function Clone() As Model10.Grouper
        Clone = New Model10.Grouper(Me, MyBase.BackColor)

        Dim mem As Model10.Member
        For Each mem In Me.Members
            Clone.Members.Add(mem.Clone, 0, 0)
        Next : mem = Nothing
    End Function

    Public Property KeepTogether() As Boolean
        Get
            Return mKeepTogether
        End Get
        Set(ByVal Value As Boolean)
            mKeepTogether = Value
        End Set
    End Property
    Public Property NewPage() As Model10.NewPageMode
        Get
            Return mNewPage
        End Get
        Set(ByVal Value As Model10.NewPageMode)
            mNewPage = Value
        End Set
    End Property
    Public Property PrintBackground() As Boolean
        Get
            Return mPrintBackground
        End Get
        Set(ByVal Value As Boolean)
            mPrintBackground = Value
        End Set
    End Property
    Public Property RepeatOnEveryPage() As Boolean
        Get
            Return mRepeatOnEveryPage
        End Get
        Set(ByVal Value As Boolean)
            mRepeatOnEveryPage = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Detail
    Inherits Section

    Private mAlternatingBackColor As Drawing.Color
    Private mKeepTogether As Boolean
    Private mNewPage As Model10.NewPageMode
    Private mPrintBackground As Boolean
    Private mSelectionBackColor As Drawing.Color
    Private mSelectionForeColor As Drawing.Color

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mAlternatingBackColor = CType(info.GetValue("AlternatingBackColor", GetType(Drawing.Color)), Drawing.Color)
        mKeepTogether = info.GetBoolean("KeepTogether")
        mNewPage = CType(info.GetByte("NewPage"), Model10.NewPageMode)
        mPrintBackground = info.GetBoolean("PrintBackground")
        mSelectionBackColor = CType(info.GetValue("SelectionBackColor", GetType(Drawing.Color)), Drawing.Color)
        mSelectionForeColor = CType(info.GetValue("SelectionForeColor", GetType(Drawing.Color)), Drawing.Color)
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("AlternatingBackColor", mAlternatingBackColor)
        info.AddValue("KeepTogether", mKeepTogether)
        info.AddValue("NewPage", mNewPage)
        info.AddValue("PrintBackground", mPrintBackground)
        info.AddValue("SelectionBackColor", mSelectionBackColor)
        info.AddValue("SelectionForeColor", mSelectionForeColor)
    End Sub

    Public Sub New()
        MyBase.New()

        mAlternatingBackColor = Drawing.SystemColors.Control
        mKeepTogether = False
        mNewPage = NewPageMode.None
        mPrintBackground = False
        mSelectionBackColor = Drawing.SystemColors.Control
        mSelectionForeColor = Drawing.Color.Transparent
    End Sub
    Public Sub New(ByVal detail As Model10.Detail, ByVal bkColor As Drawing.Color)
        MyBase.New(detail, bkColor)

        mAlternatingBackColor = detail.AlternatingBackColor
        mKeepTogether = detail.KeepTogether
        mNewPage = detail.NewPage
        mPrintBackground = detail.PrintBackground
        mSelectionBackColor = detail.SelectionBackColor
        mSelectionForeColor = detail.SelectionForeColor
    End Sub
    Public Function Clone() As Model10.Detail
        Clone = New Model10.Detail(Me, MyBase.BackColor)

        Dim mem As Model10.Member
        For Each mem In Me.Members
            Clone.Members.Add(mem.Clone, 0, 0)
        Next : mem = Nothing
    End Function

    Public Property AlternatingBackColor() As Drawing.Color
        Get
            Return mAlternatingBackColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mAlternatingBackColor = Value
        End Set
    End Property
    Public Property KeepTogether() As Boolean
        Get
            Return mKeepTogether
        End Get
        Set(ByVal Value As Boolean)
            mKeepTogether = Value
        End Set
    End Property
    Public Property NewPage() As Model10.NewPageMode
        Get
            Return mNewPage
        End Get
        Set(ByVal Value As Model10.NewPageMode)
            mNewPage = Value
        End Set
    End Property
    Public Property PrintBackground() As Boolean
        Get
            Return mPrintBackground
        End Get
        Set(ByVal Value As Boolean)
            mPrintBackground = Value
        End Set
    End Property
    Public Property SelectionBackColor() As Drawing.Color
        Get
            Return mSelectionBackColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mSelectionBackColor = Value
        End Set
    End Property
    Public Property SelectionForeColor() As Drawing.Color
        Get
            Return mSelectionForeColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mSelectionForeColor = Value
        End Set
    End Property
End Class

<Serializable()> Public Class Series
    Implements ISerializable

    Private mChart As Model10.Chart
    Private mColor As Drawing.Color
    Private mDataField As String

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mChart = CType(info.GetValue("Chart", GetType(Model10.Chart)), Model10.Chart)
        mColor = CType(info.GetValue("Color", GetType(Drawing.Color)), Drawing.Color)
        mDataField = info.GetString("DataField")
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("Chart", mChart)
        info.AddValue("Color", mColor)
        info.AddValue("DataField", mDataField)
    End Sub

    Public Overrides Function ToString() As String
        Return mDataField
    End Function
    Public Sub New()
        MyBase.New()

        mChart = Nothing
        mColor = Drawing.Color.Transparent
        mDataField = ""
    End Sub
    Public Sub New(ByVal series As Model10.Series)
        MyBase.New()

        mChart = Nothing
        mColor = series.Color
        mDataField = series.DataField
    End Sub

    <ComponentModel.Browsable(False)> Public Property Chart() As Model10.Chart
        Get
            Return mChart
        End Get
        Set(ByVal Value As Model10.Chart)
            mChart = Value
        End Set
    End Property
    Public Property Color() As Drawing.Color
        Get
            Return mColor
        End Get
        Set(ByVal Value As Drawing.Color)
            mColor = Value
        End Set
    End Property
    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
End Class
<Serializable()> Public Class DataBinding
    Implements ISerializable

    Private mDataField As String
    Private mName As String
    Private mObject As Object
    Private mPropertyName As String
    Private mTable As Model10.Table

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mDataField = info.GetString("DataField")
        mName = info.GetString("Name")
        mObject = info.GetValue("Object", GetType(Object))
        mPropertyName = info.GetString("PropertyName")
        mTable = CType(info.GetValue("Table", GetType(Model10.Table)), Model10.Table)
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("DataField", mDataField)
        info.AddValue("Name", mName)
        info.AddValue("Object", mObject)
        info.AddValue("PropertyName", mPropertyName)
        info.AddValue("Table", mTable)
    End Sub

    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New()
        MyBase.New()

        mDataField = ""
        mName = ""
        mObject = Nothing
        mPropertyName = ""
        mTable = Nothing
    End Sub

    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            If Not mTable Is Nothing Then
                Dim db As Model10.DataBinding
                For Each db In mTable.DataBindings
                    If db.Name.ToLower = Value.ToLower Then
                        If Not db Is Me Then
                            Throw New System.Data.DuplicateNameException("A Databinding named '" & Value & "' already belongs to this Table.")
                        End If
                    End If
                Next : db = Nothing
            End If
            mName = Value
        End Set
    End Property
    Public Property [Object]() As Object
        Get
            Return mObject
        End Get
        Set(ByVal Value As Object)
            mObject = Value
        End Set
    End Property
    Public Property PropertyName() As String
        Get
            Return mPropertyName
        End Get
        Set(ByVal Value As String)
            mPropertyName = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Table() As Model10.Table
        Get
            Return mTable
        End Get
        Set(ByVal Value As Model10.Table)
            mTable = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Relation
    Implements ISerializable

    <NonSerialized()> Private mDataRelation As DataRelation

    Private mChildTable As Model10.Table
    Private mName As String
    Private mParentTable As Model10.Table
    Private mPresentation As Model10.Presentation
    Private mSideMargin As Integer
    Private mVisible As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mChildTable = CType(info.GetValue("ChildTable", GetType(Model10.Table)), Model10.Table)
        mName = info.GetString("Name")
        mParentTable = CType(info.GetValue("ParentTable", GetType(Model10.Table)), Model10.Table)
        mPresentation = CType(info.GetValue("Presentation", GetType(Model10.Presentation)), Model10.Presentation)
        mSideMargin = info.GetInt32("SideMargin")
        mVisible = info.GetBoolean("Visible")
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("ChildTable", mChildTable)
        info.AddValue("Name", mName)
        info.AddValue("ParentTable", mParentTable)
        info.AddValue("Presentation", mPresentation)
        info.AddValue("SideMargin", mSideMargin)
        info.AddValue("Visible", mVisible)
    End Sub

    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New(ByVal relationName As String, ByVal parentTable As Model10.Table, ByVal childTable As Model10.Table, ByVal parentColumn() As DataColumn, ByVal childColumn() As DataColumn)
        MyBase.New()

        mDataRelation = New DataRelation(relationName, parentColumn, childColumn)

        mChildTable = childTable
        mName = mDataRelation.RelationName
        mParentTable = parentTable
        mPresentation = Nothing
        mSideMargin = 0
        mVisible = True
    End Sub

    <ComponentModel.Browsable(False)> Public Property DataRelation() As DataRelation
        Get
            If mDataRelation Is Nothing Then
                If Not mPresentation Is Nothing Then Return mPresentation.DataSet.Relations(mName)
            Else
                Return mDataRelation
            End If
        End Get
        Set(ByVal Value As DataRelation)
            mDataRelation = Value
        End Set
    End Property

    Public Property ChildTable() As Model10.Table
        Get
            Return mChildTable
        End Get
        Set(ByVal Value As Model10.Table)
            mChildTable = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            If Not mPresentation Is Nothing Then
                Dim dr As Model10.Relation
                For Each dr In mPresentation.Relations
                    If dr.Name.ToLower = Value.ToLower Then
                        If Not dr Is Me Then
                            Throw New System.Data.DuplicateNameException("A Relation named '" & Value & "' already belongs to this Presentation.")
                        End If
                    End If
                Next : dr = Nothing
            End If
            DataRelation.RelationName = Value
            mName = Value
        End Set
    End Property
    Public Property ParentTable() As Model10.Table
        Get
            Return mParentTable
        End Get
        Set(ByVal Value As Model10.Table)
            mParentTable = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Presentation() As Model10.Presentation
        Get
            Return mPresentation
        End Get
        Set(ByVal Value As Model10.Presentation)
            mPresentation = Value
        End Set
    End Property
    Public Property SideMargin() As Integer
        Get
            Return mSideMargin
        End Get
        Set(ByVal Value As Integer)
            mSideMargin = Value
        End Set
    End Property
    Public Property Visible() As Boolean
        Get
            Return mVisible
        End Get
        Set(ByVal Value As Boolean)
            mVisible = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Parameter
    Implements ISerializable

    <NonSerialized()> Private mOleDbParameter As OleDb.OleDbParameter

    Private mCommand As Model10.Command
    Private mDbType As DbType
    Private mDirection As ParameterDirection
    Private mIsNullable As Boolean
    Private mName As String
    Private mPrecision As Byte
    Private mScale As Byte
    Private mSize As Integer
    Private mSourceColumn As String
    Private mSourceVersion As DataRowVersion
    Private mValue As Object

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mCommand = CType(info.GetValue("Command", GetType(Model10.Command)), Model10.Command)
        mDbType = CType(info.GetByte("DbType"), DbType)
        mDirection = CType(info.GetByte("Direction"), ParameterDirection)
        mIsNullable = info.GetBoolean("IsNullable")
        mName = info.GetString("Name")
        mPrecision = info.GetByte("Precision")
        mScale = info.GetByte("Scale")
        mSize = info.GetInt32("Size")
        mSourceColumn = info.GetString("SourceColumn")
        mSourceVersion = CType(info.GetInt32("SourceVersion"), DataRowVersion)
        mValue = info.GetValue("Value", GetType(Object))
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("Command", mCommand)
        info.AddValue("DbType", mDbType)
        info.AddValue("Direction", mDirection)
        info.AddValue("IsNullable", mIsNullable)
        info.AddValue("Name", mName)
        info.AddValue("Precision", mPrecision)
        info.AddValue("Scale", mScale)
        info.AddValue("Size", mSize)
        info.AddValue("SourceColumn", mSourceColumn)
        info.AddValue("SourceVersion", mSourceVersion)
        info.AddValue("Value", mValue)
    End Sub

    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New()
        MyBase.New()

        mOleDbParameter = Nothing

        Dim myParam As New OleDb.OleDbParameter
        '------------
        mCommand = Nothing
        mDbType = myParam.DbType
        mDirection = myParam.Direction
        mIsNullable = myParam.IsNullable
        mName = myParam.ParameterName
        mPrecision = myParam.Precision
        mScale = myParam.Scale
        mSize = myParam.Size
        mSourceColumn = myParam.SourceColumn
        mSourceVersion = myParam.SourceVersion
        mValue = myParam.Value
        '------------
        myParam = Nothing
    End Sub
    Public Sub New(ByVal parameter As Model10.Parameter)
        MyBase.New()

        mOleDbParameter = Nothing

        mCommand = Nothing
        mDbType = parameter.DbType
        mDirection = parameter.Direction
        mIsNullable = parameter.IsNullable
        mName = parameter.Name
        mPrecision = parameter.Precision
        mScale = parameter.Scale
        mSize = parameter.Size
        mSourceColumn = parameter.SourceColumn
        mSourceVersion = parameter.SourceVersion
        mValue = parameter.Value
    End Sub

    <ComponentModel.Browsable(False)> Public Property OleDbParameter() As OleDb.OleDbParameter
        Get
            If mOleDbParameter Is Nothing Then
                mOleDbParameter = New OleDb.OleDbParameter(mName, mValue)
                mOleDbParameter.DbType = mDbType
                mOleDbParameter.Size = mSize
                mOleDbParameter.Direction = mDirection
                mOleDbParameter.IsNullable = mIsNullable
                mOleDbParameter.Precision = mPrecision
                mOleDbParameter.Scale = mScale
                mOleDbParameter.SourceColumn = mSourceColumn
                mOleDbParameter.SourceVersion = mSourceVersion
            End If
            Return mOleDbParameter
        End Get
        Set(ByVal Value As OleDb.OleDbParameter)
            mOleDbParameter = Value
        End Set
    End Property

    <ComponentModel.Browsable(False)> Public Property Command() As Model10.Command
        Get
            Return mCommand
        End Get
        Set(ByVal Value As Model10.Command)
            mCommand = Value
        End Set
    End Property
    Public Property DbType() As DbType
        Get
            Return mDbType
        End Get
        Set(ByVal Value As DbType)
            OleDbParameter.DbType = Value
            mDbType = Value
        End Set
    End Property
    Public Property Direction() As ParameterDirection
        Get
            Return mDirection
        End Get
        Set(ByVal Value As ParameterDirection)
            OleDbParameter.Direction = Value
            mDirection = Value
        End Set
    End Property
    Public Property IsNullable() As Boolean
        Get
            Return mIsNullable
        End Get
        Set(ByVal Value As Boolean)
            OleDbParameter.IsNullable = Value
            mIsNullable = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            If Not mCommand Is Nothing Then
                Dim prm As Model10.Parameter
                For Each prm In mCommand.Parameters
                    If prm.Name.ToLower = Value.ToLower Then
                        If Not prm Is Me Then
                            Throw New System.Data.DuplicateNameException("A Parameter named '" & Value & "' already belongs to this Command.")
                        End If
                    End If
                Next : prm = Nothing
            End If
            OleDbParameter.ParameterName = Value
            mName = Value
        End Set
    End Property
    Public Property Precision() As Byte
        Get
            Return mPrecision
        End Get
        Set(ByVal Value As Byte)
            OleDbParameter.Precision = Value
            mPrecision = Value
        End Set
    End Property
    Public Property Scale() As Byte
        Get
            Return mScale
        End Get
        Set(ByVal Value As Byte)
            OleDbParameter.Scale = Value
            mScale = Value
        End Set
    End Property
    Public Property Size() As Integer
        Get
            Return mSize
        End Get
        Set(ByVal Value As Integer)
            OleDbParameter.Size = Value
            mSize = Value
        End Set
    End Property
    Public Property SourceColumn() As String
        Get
            Return mSourceColumn
        End Get
        Set(ByVal Value As String)
            OleDbParameter.SourceColumn = Value
            mSourceColumn = Value
        End Set
    End Property
    Public Property SourceVersion() As DataRowVersion
        Get
            Return mSourceVersion
        End Get
        Set(ByVal Value As DataRowVersion)
            OleDbParameter.SourceVersion = Value
            mSourceVersion = Value
        End Set
    End Property
    Public Property Value() As Object
        Get
            Return mValue
        End Get
        Set(ByVal Value As Object)
            OleDbParameter.Value = Value
            mValue = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Command
    Implements ISerializable

    <NonSerialized()> Private mOleDbCommand As OleDb.OleDbCommand
    <NonSerialized()> Private mPrepared As Boolean

    Private mCommandText As String
    Private mCommandTimeout As Integer
    Private mCommandType As CommandType
    Private mConnection As Model10.Connection
    Private ReadOnly mParameters As Model10.ParameterCollection
    Private mTable As Model10.Table
    Private mUpdatedRowSource As UpdateRowSource

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mCommandText = info.GetString("CommandText")
        mCommandTimeout = info.GetInt32("CommandTimeout")
        mCommandType = CType(info.GetInt32("CommandType"), CommandType)
        mConnection = CType(info.GetValue("Connection", GetType(Model10.Connection)), Model10.Connection)
        mParameters = CType(info.GetValue("Parameters", GetType(Model10.ParameterCollection)), Model10.ParameterCollection)
        mTable = CType(info.GetValue("Table", GetType(Model10.Table)), Model10.Table)
        mUpdatedRowSource = CType(info.GetInt32("UpdatedRowSource"), UpdateRowSource)
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("CommandText", mCommandText)
        info.AddValue("CommandTimeout", mCommandTimeout)
        info.AddValue("CommandType", mCommandType)
        info.AddValue("Connection", mConnection)
        info.AddValue("Parameters", mParameters)
        info.AddValue("Table", mTable)
        info.AddValue("UpdatedRowSource", mUpdatedRowSource)
    End Sub

    Public Function GetName() As String
        If Not mTable Is Nothing Then
            If Me Is mTable.SelectCommand Then Return "Select"
            If Me Is mTable.InsertCommand Then Return "Insert"
            If Me Is mTable.UpdateCommand Then Return "Update"
            If Me Is mTable.DeleteCommand Then Return "Delete"
        End If
    End Function
    Public Function GetNewParameterName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mParameters.Count - 1
            Dim prm As Model10.Parameter = CType(mParameters(i), Model10.Parameter)
            If prm.Name = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            prm = Nothing
        Loop
        Return str
    End Function
    Public Sub New()
        MyBase.New()

        mOleDbCommand = Nothing
        mPrepared = False

        mCommandText = ""
        mCommandTimeout = 30
        mCommandType = CommandType.Text
        mConnection = Nothing
        mParameters = New Model10.ParameterCollection(Me)
        mTable = Nothing
        mUpdatedRowSource = UpdateRowSource.Both
    End Sub
    Public Sub New(ByVal command As Model10.Command)
        MyBase.New()

        mOleDbCommand = Nothing
        mPrepared = False

        mCommandText = command.CommandText
        mCommandTimeout = command.CommandTimeout
        mCommandType = command.CommandType
        mConnection = Nothing
        mParameters = New Model10.ParameterCollection(Me)
        mTable = Nothing
        mUpdatedRowSource = command.UpdatedRowSource

        Dim prm As Model10.Parameter
        For Each prm In command.Parameters
            Me.Parameters.Add(New Model10.Parameter(prm))
        Next : prm = Nothing
    End Sub

    <ComponentModel.Browsable(False)> Public Property OleDbCommand() As OleDb.OleDbCommand
        Get
            If mOleDbCommand Is Nothing Then
                mOleDbCommand = New OleDb.OleDbCommand
                mOleDbCommand.CommandText = mCommandText
                mOleDbCommand.CommandTimeout = mCommandTimeout
                mOleDbCommand.CommandType = mCommandType
                If Not mConnection Is Nothing Then mOleDbCommand.Connection = mConnection.OleDbConnection

                Dim prm As Model10.Parameter
                For Each prm In mParameters
                    mOleDbCommand.Parameters.Add(prm.OleDbParameter)
                Next : prm = Nothing

                mOleDbCommand.UpdatedRowSource = mUpdatedRowSource
            End If
            Return mOleDbCommand
        End Get
        Set(ByVal Value As OleDb.OleDbCommand)
            mOleDbCommand = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Prepared() As Boolean
        Get
            Return mPrepared
        End Get
        Set(ByVal Value As Boolean)
            mPrepared = Value
        End Set
    End Property

    Public Property CommandText() As String
        Get
            Return mCommandText
        End Get
        Set(ByVal Value As String)
            OleDbCommand.CommandText = Value
            mCommandText = Value
        End Set
    End Property
    Public Property CommandTimeout() As Integer
        Get
            Return mCommandTimeout
        End Get
        Set(ByVal Value As Integer)
            OleDbCommand.CommandTimeout = Value
            mCommandTimeout = Value
        End Set
    End Property
    Public Property CommandType() As CommandType
        Get
            Return mCommandType
        End Get
        Set(ByVal Value As CommandType)
            OleDbCommand.CommandType = Value
            mCommandType = Value
        End Set
    End Property
    Public Property Connection() As Model10.Connection
        Get
            Return mConnection
        End Get
        Set(ByVal Value As Model10.Connection)
            If Value Is Nothing Then
                OleDbCommand.Connection = Nothing
            Else
                OleDbCommand.Connection = Value.OleDbConnection
            End If
            mConnection = Value
        End Set
    End Property
    Public ReadOnly Property Parameters() As Model10.ParameterCollection
        Get
            Return mParameters
        End Get
    End Property
    <ComponentModel.Browsable(False)> Public Property Table() As Model10.Table
        Get
            Return mTable
        End Get
        Set(ByVal Value As Model10.Table)
            mTable = Value
        End Set
    End Property
    Public Property UpdatedRowSource() As UpdateRowSource
        Get
            Return mUpdatedRowSource
        End Get
        Set(ByVal Value As UpdateRowSource)
            OleDbCommand.UpdatedRowSource = Value
            mUpdatedRowSource = Value
        End Set
    End Property
End Class
'<Serializable()> Public Class [Default]
'    Inherits Model10.Command

'    Private mDataField As String
'    Private mName As String

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next

'        mDataField = info.GetString("DataField")
'        mName = info.GetString("Name")
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'        info.AddValue("DataField", mDataField)
'        info.AddValue("Name", mName)
'    End Sub

'    Public Overrides Function ToString() As String
'        Return mName
'    End Function
'    Public Sub New()
'        MyBase.New()

'        mDataField = ""
'        mName = ""
'    End Sub

'    Public Property DataField() As String
'        Get
'            Return mDataField
'        End Get
'        Set(ByVal Value As String)
'            mDataField = Value
'        End Set
'    End Property
'    Public Property Name() As String
'        Get
'            Return mName
'        End Get
'        Set(ByVal Value As String)
'            Value = Value.Trim
'            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
'            If Not MyBase.Table Is Nothing Then
'                Dim def As Model10.Default
'                For Each def In MyBase.Table.Defaults
'                    If Not def Is Me Then
'                        If def.Name.ToLower = Value.ToLower Then
'                            Throw New System.Data.DuplicateNameException("A Compute named '" & Value & "' already belongs to this Table.")
'                        End If
'                    End If
'                Next : def = Nothing
'            End If
'            mName = Value
'        End Set
'    End Property
'End Class
<Serializable()> Public Class Compute
    Inherits Model10.Command

    Private mAction As Model10.ActionMode
    Private mName As String
    Private mPromptMessage As String
    Private mRemove As Boolean
    Private mSelectionField As String

    Private mPresentation As String = ""
    Private mRelationName As String = ""

    <NonSerialized()> Private mForm As Windows.Forms.Form

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.New(info, context)
        On Error Resume Next

        mAction = CType(info.GetByte("Action"), Model10.ActionMode)
        mName = info.GetString("Name")
        mPromptMessage = info.GetString("PromptMessage")
        mRemove = info.GetBoolean("Remove")
        mSelectionField = info.GetString("SelectionField")

        mPresentation = info.GetString("Presentation")
        mRelationName = info.GetString("RelationName")
    End Sub
    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        MyBase.GetObjectData(info, context)

        info.AddValue("Action", mAction)
        info.AddValue("Name", mName)
        info.AddValue("PromptMessage", mPromptMessage)
        info.AddValue("Remove", mRemove)
        info.AddValue("SelectionField", mSelectionField)

        info.AddValue("Presentation", mPresentation)
        info.AddValue("RelationName", mRelationName)
    End Sub

    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New()
        MyBase.New()

        mAction = ActionMode.Record
        mName = ""
        mPromptMessage = ""
        mRemove = False
        mSelectionField = ""

        mPresentation = ""
        mRelationName = ""
    End Sub
    Public Sub New(ByVal compute As Model10.Compute)
        MyBase.New(compute)

        mAction = compute.Action
        mName = compute.Name
        mPromptMessage = compute.PromptMessage
        mRemove = compute.Remove
        mSelectionField = compute.SelectionField

        mPresentation = compute.Presentation
        mRelationName = compute.RelationName
    End Sub

    Public Property Action() As Model10.ActionMode
        Get
            Return mAction
        End Get
        Set(ByVal Value As Model10.ActionMode)
            mAction = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            If Not MyBase.Table Is Nothing Then
                Dim cmp As Model10.Compute
                For Each cmp In MyBase.Table.Computes
                    If Not cmp Is Me Then
                        If cmp.Name.ToLower = Value.ToLower Then
                            Throw New System.Data.DuplicateNameException("A Compute named '" & Value & "' already belongs to this Table.")
                        End If
                    End If
                Next : cmp = Nothing
            End If
            mName = Value
        End Set
    End Property
    Public Property PromptMessage() As String
        Get
            Return mPromptMessage
        End Get
        Set(ByVal Value As String)
            mPromptMessage = Value
        End Set
    End Property
    Public Property Remove() As Boolean
        Get
            Return mRemove
        End Get
        Set(ByVal Value As Boolean)
            mRemove = Value
        End Set
    End Property
    Public Property SelectionField() As String
        Get
            Return mSelectionField
        End Get
        Set(ByVal Value As String)
            mSelectionField = Value
        End Set
    End Property

    Public Property Presentation() As String
        Get
            Return mPresentation
        End Get
        Set(ByVal Value As String)
            mPresentation = Value
        End Set
    End Property
    Public Property RelationName() As String
        Get
            Return mRelationName
        End Get
        Set(ByVal Value As String)
            mRelationName = Value
        End Set
    End Property

    <ComponentModel.Browsable(False)> Public Property Form() As Windows.Forms.Form
        Get
            Return mForm
        End Get
        Set(ByVal Value As Windows.Forms.Form)
            mForm = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Table
    Implements ISerializable
    Implements IDesigner

    <NonSerialized()> Private mDataTable As DataTable
    <NonSerialized()> Private mField As String
    <NonSerialized()> Private mFilled As Boolean

    Private mAutoFill As Boolean = True
    Private mAutoSave As Boolean
    Private mBand As Boolean
    Private ReadOnly mChildRelations As Model10.RelationCollection
    Private ReadOnly mComputes As Model10.ComputeCollection
    Private mDataBindings As Model10.DataBindingCollection
    Private mDataField As String
    'Private ReadOnly mDefaults As Model10.DefaultCollection
    Private mDesignerMode As Model10.DesignerMode
    Private mDetail As Model10.Detail
    Private mErrorField As String
    Private mFillOnce As Boolean = True
    Private mFilter As String
    Private mFoxProField As String
    Private mGroupFooter As Model10.Grouper
    Private mGroupHeader As Model10.Grouper
    Private mImage As Drawing.Image
    Private mInitialPageNumber As Integer
    Private mKeepTogether As Model10.KeepMode
    Private mName As String
    Private mNavigator As Boolean = True
    Private mPageFooter As Model10.Pager
    Private mPageHeader As Model10.Pager
    Private mPageSize As Integer
    Private ReadOnly mParentRelations As Model10.RelationCollection
    Private mPresentation As Model10.Presentation
    Private mPrintable As Boolean
    Private mRetrieveIdentity As Boolean
    Private mRowPosition As Model10.RowStatus
    Private mSize As Drawing.Size
    Private mSort As String
    Private mWidth As Integer

    Private mDisplayGrid As Boolean
    Private mDisplayRulers As Boolean
    Private mGridSize As Drawing.Size
    Private mLocked As Boolean
    Private mSnapToGrid As Boolean

    Private mLandscape As Boolean
    Private mPaperKind As Drawing.Printing.PaperKind
    Private mPrinterName As String
    Private mSourceKind As Drawing.Printing.PaperSourceKind

    Private mBottomMargin As Integer
    Private mLeftMargin As Integer
    Private mRightMargin As Integer
    Private mTopMargin As Integer

    <NonSerialized()> Private WithEvents mOleDbDataAdapter As OleDb.OleDbDataAdapter

    Private mDeleteCommand As Model10.Command
    Private mInsertCommand As Model10.Command
    Private mSelectCommand As Model10.Command
    Private mUpdateCommand As Model10.Command

    Private Const MIN_SIZE As Byte = 1

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mAutoFill = info.GetBoolean("AutoFill")
        mAutoSave = info.GetBoolean("AutoSave")
        mBand = info.GetBoolean("Band")
        mChildRelations = CType(info.GetValue("ChildRelations", GetType(Model10.RelationCollection)), Model10.RelationCollection)
        mComputes = CType(info.GetValue("Computes", GetType(Model10.ComputeCollection)), Model10.ComputeCollection)
        mDataBindings = CType(info.GetValue("DataBindings", GetType(Model10.DataBindingCollection)), Model10.DataBindingCollection)
        mDataField = info.GetString("DataField")
        'mDefaults = CType(info.GetValue("Defaults", GetType(Model10.DefaultCollection)), Model10.DefaultCollection)
        mDesignerMode = CType(info.GetByte("DesignerMode"), Model10.DesignerMode)
        mDetail = CType(info.GetValue("Detail", GetType(Model10.Detail)), Model10.Detail)
        mErrorField = info.GetString("ErrorField")
        mFillOnce = info.GetBoolean("FillOnce")
        mFilter = info.GetString("Filter")
        mFoxProField = info.GetString("FoxProField")
        mGroupFooter = CType(info.GetValue("GroupFooter", GetType(Model10.Grouper)), Model10.Grouper)
        mGroupHeader = CType(info.GetValue("GroupHeader", GetType(Model10.Grouper)), Model10.Grouper)
        mImage = CType(info.GetValue("Image", GetType(Drawing.Image)), Drawing.Image)
        mInitialPageNumber = info.GetInt32("InitialPageNumber")
        mKeepTogether = CType(info.GetByte("KeepTogether"), Model10.KeepMode)
        mName = info.GetString("Name")
        mNavigator = info.GetBoolean("Navigator")
        mPageFooter = CType(info.GetValue("PageFooter", GetType(Model10.Pager)), Model10.Pager)
        mPageHeader = CType(info.GetValue("PageHeader", GetType(Model10.Pager)), Model10.Pager)
        mPageSize = info.GetInt32("PageSize")
        mParentRelations = CType(info.GetValue("ParentRelations", GetType(Model10.RelationCollection)), Model10.RelationCollection)
        mPresentation = CType(info.GetValue("Presentation", GetType(Model10.Presentation)), Model10.Presentation)
        mPrintable = info.GetBoolean("Printable")
        mRetrieveIdentity = info.GetBoolean("RetrieveIdentity")
        mRowPosition = CType(info.GetByte("RowPosition"), Model10.RowStatus)
        mSize = CType(info.GetValue("Size", GetType(Drawing.Size)), Drawing.Size)
        mSort = info.GetString("Sort")
        mWidth = info.GetInt32("Width")

        mDisplayGrid = info.GetBoolean("DisplayGrid")
        mDisplayRulers = info.GetBoolean("DisplayRulers")
        mGridSize = CType(info.GetValue("GridSize", GetType(Drawing.Size)), Drawing.Size)
        mLocked = info.GetBoolean("Locked")
        mSnapToGrid = info.GetBoolean("SnapToGrid")

        mLandscape = info.GetBoolean("Landscape")
        mPaperKind = CType(info.GetInt32("PaperKind"), Drawing.Printing.PaperKind)
        mPrinterName = info.GetString("PrinterName")
        mSourceKind = CType(info.GetInt32("SourceKind"), Drawing.Printing.PaperSourceKind)

        mBottomMargin = info.GetInt32("BottomMargin")
        mLeftMargin = info.GetInt32("LeftMargin")
        mRightMargin = info.GetInt32("RightMargin")
        mTopMargin = info.GetInt32("TopMargin")

        mDeleteCommand = CType(info.GetValue("DeleteCommand", GetType(Model10.Command)), Model10.Command)
        mInsertCommand = CType(info.GetValue("InsertCommand", GetType(Model10.Command)), Model10.Command)
        mSelectCommand = CType(info.GetValue("SelectCommand", GetType(Model10.Command)), Model10.Command)
        mUpdateCommand = CType(info.GetValue("UpdateCommand", GetType(Model10.Command)), Model10.Command)

        If mPresentation Is Nothing Then mDataTable = CType(info.GetValue("DataTable", GetType(DataTable)), DataTable)
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("AutoFill", mAutoFill)
        info.AddValue("AutoSave", mAutoSave)
        info.AddValue("Band", mBand)
        info.AddValue("ChildRelations", mChildRelations)
        info.AddValue("Computes", mComputes)
        info.AddValue("DataBindings", mDataBindings)
        info.AddValue("DataField", mDataField)
        'info.AddValue("Defaults", mDefaults)
        info.AddValue("DesignerMode", mDesignerMode)
        info.AddValue("Detail", mDetail)
        info.AddValue("ErrorField", mErrorField)
        info.AddValue("FillOnce", mFillOnce)
        info.AddValue("Filter", mFilter)
        info.AddValue("FoxProField", mFoxProField)
        info.AddValue("GroupFooter", mGroupFooter)
        info.AddValue("GroupHeader", mGroupHeader)
        info.AddValue("Image", mImage)
        info.AddValue("InitialPageNumber", mInitialPageNumber)
        info.AddValue("KeepTogether", mKeepTogether)
        info.AddValue("Name", mName)
        info.AddValue("Navigator", mNavigator)
        info.AddValue("PageFooter", mPageFooter)
        info.AddValue("PageHeader", mPageHeader)
        info.AddValue("PageSize", mPageSize)
        info.AddValue("ParentRelations", mParentRelations)
        info.AddValue("Presentation", mPresentation)
        info.AddValue("Printable", mPrintable)
        info.AddValue("RetrieveIdentity", mRetrieveIdentity)
        info.AddValue("RowPosition", mRowPosition)
        info.AddValue("Size", mSize)
        info.AddValue("Sort", mSort)
        info.AddValue("Width", mWidth)

        info.AddValue("DisplayGrid", mDisplayGrid)
        info.AddValue("DisplayRulers", mDisplayRulers)
        info.AddValue("GridSize", mGridSize)
        info.AddValue("Locked", mLocked)
        info.AddValue("SnapToGrid", mSnapToGrid)

        info.AddValue("Landscape", mLandscape)
        info.AddValue("PaperKind", mPaperKind)
        info.AddValue("PrinterName", mPrinterName)
        info.AddValue("SourceKind", mSourceKind)

        info.AddValue("BottomMargin", mBottomMargin)
        info.AddValue("LeftMargin", mLeftMargin)
        info.AddValue("RightMargin", mRightMargin)
        info.AddValue("TopMargin", mTopMargin)

        info.AddValue("DeleteCommand", mDeleteCommand)
        info.AddValue("InsertCommand", mInsertCommand)
        info.AddValue("SelectCommand", mSelectCommand)
        info.AddValue("UpdateCommand", mUpdateCommand)

        If mPresentation Is Nothing Then info.AddValue("DataTable", mDataTable)
    End Sub

    Public Function GetMinWidth() As Integer
        GetMinWidth = MIN_SIZE

        If Not mPageHeader Is Nothing Then
            Dim mem As Model10.Member
            For Each mem In mPageHeader.Members
                If GetMinWidth < mem.Location.X + mem.Size.Width Then GetMinWidth = mem.Location.X + mem.Size.Width
            Next : mem = Nothing
        End If
        If Not mGroupHeader Is Nothing Then
            Dim mem As Model10.Member
            For Each mem In mGroupHeader.Members
                If GetMinWidth < mem.Location.x + mem.Size.Width Then GetMinWidth = mem.Location.x + mem.Size.Width
            Next : mem = Nothing
        End If
        If Not mDetail Is Nothing Then
            Dim mem As Model10.Member
            For Each mem In mDetail.Members
                If GetMinWidth < mem.Location.x + mem.Size.Width Then GetMinWidth = mem.Location.x + mem.Size.Width
            Next : mem = Nothing
        End If
        If Not mGroupFooter Is Nothing Then
            Dim mem As Model10.Member
            For Each mem In mGroupFooter.Members
                If GetMinWidth < mem.Location.x + mem.Size.Width Then GetMinWidth = mem.Location.x + mem.Size.Width
            Next : mem = Nothing
        End If
        If Not mPageFooter Is Nothing Then
            Dim mem As Model10.Member
            For Each mem In mPageFooter.Members
                If GetMinWidth < mem.Location.x + mem.Size.Width Then GetMinWidth = mem.Location.x + mem.Size.Width
            Next : mem = Nothing
        End If
    End Function
    Public Function GetNewColumnName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > DataTable.Columns.Count - 1
            Dim dc As DataColumn = DataTable.Columns(i)
            If dc.ColumnName = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            dc = Nothing
        Loop
        Return str
    End Function
    Public Function GetNewComputeName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mComputes.Count - 1
            Dim cmp As Model10.Compute = CType(mComputes(i), Model10.Compute)
            If cmp.Name = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            cmp = Nothing
        Loop
        Return str
    End Function
    Public Function GetNewConstraintName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > DataTable.Constraints.Count - 1
            Dim con As Constraint = DataTable.Constraints(i)
            If con.ConstraintName = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            con = Nothing
        Loop
        Return str
    End Function
    Public Function GetNewDataBindingName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mDataBindings.Count - 1
            Dim db As Model10.DataBinding = mDataBindings(i)
            If db.Name = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            db = Nothing
        Loop
        Return str
    End Function
    'Public Function GetNewDefaultName(ByVal name As String) As String
    '    Dim str As String
    '    Dim i, j As Integer
    '    j = 1 : str = name & j.ToString
    '    Do Until i > mDefaults.Count - 1
    '        Dim def As Model10.Default = CType(mDefaults(i), Model10.Default)
    '        If def.Name = str Then
    '            j += 1
    '            str = name & j.ToString

    '            i = 0
    '        Else
    '            i += 1
    '        End If
    '        def = Nothing
    '    Loop
    '    Return str
    'End Function
    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New()
        MyBase.New()

        mDataTable = New DataTable
        mField = ""
        mFilled = False

        mAutoFill = True
        mAutoSave = True
        mBand = True
        mChildRelations = New Model10.RelationCollection(Nothing)
        mComputes = New Model10.ComputeCollection(Me)
        mDataBindings = New Model10.DataBindingCollection(Me)
        mDataField = ""
        'mDefaults = New Model10.DefaultCollection(Me)
        mDesignerMode = DesignerMode.FormView
        mDetail = New Model10.Detail : mDetail.Table = Me
        mErrorField = ""
        mFillOnce = True
        mFilter = ""
        mFoxProField = ""
        mGroupFooter = New Model10.Grouper : mGroupFooter.Table = Me
        mGroupHeader = New Model10.Grouper : mGroupHeader.Table = Me
        mImage = Nothing
        mInitialPageNumber = 1
        mKeepTogether = KeepMode.No
        mName = mDataTable.TableName
        mNavigator = False
        mPageFooter = Nothing
        mPageHeader = Nothing
        mPageSize = 10
        mParentRelations = New Model10.RelationCollection(Nothing)
        mPresentation = Nothing
        mPrintable = True
        mRetrieveIdentity = False
        mRowPosition = RowStatus.First
        mSize = New Drawing.Size(400, 300)
        mSort = ""
        mWidth = 280

        mDisplayGrid = True
        mDisplayRulers = True
        mGridSize = New Drawing.Size(4, 4)
        mLocked = False
        mSnapToGrid = True

        mLandscape = False
        mPaperKind = Drawing.Printing.PaperKind.A4
        mPrinterName = ""
        mSourceKind = Drawing.Printing.PaperSourceKind.AutomaticFeed

        mBottomMargin = 100
        mLeftMargin = 100
        mRightMargin = 100
        mTopMargin = 100

        mOleDbDataAdapter = Nothing

        mDeleteCommand = Nothing
        mInsertCommand = Nothing
        mSelectCommand = Nothing
        mUpdateCommand = Nothing
    End Sub
    Public Sub New(ByVal table As Model10.Table)
        MyBase.New()

        mDataTable = table.DataTable.Clone
        mField = ""
        mFilled = False

        mAutoFill = table.AutoFill
        mAutoSave = table.AutoSave
        mBand = table.Band
        mChildRelations = New Model10.RelationCollection(Nothing)
        mComputes = New Model10.ComputeCollection(Me)
        '----------------------------------------
        Dim cmp As Model10.Compute
        For Each cmp In table.Computes
            mComputes.Add(New Model10.Compute(cmp))
        Next : cmp = Nothing
        '----------------------------------------
        mDataBindings = New Model10.DataBindingCollection(Me)
        '----------------------------------------
        '----------------------------------------
        mDataField = table.DataField
        'mDefaults = New Model10.DefaultCollection(Me)
        mDesignerMode = table.DesignerMode
        If table.Detail Is Nothing Then
            mDetail = Nothing
        Else
            mDetail = table.Detail.Clone
            mDetail.Table = Me
        End If
        mErrorField = table.ErrorField
        mFillOnce = table.FillOnce
        mFilter = table.Filter
        mFoxProField = table.FoxProField
        If table.GroupFooter Is Nothing Then
            mGroupFooter = Nothing
        Else
            mGroupFooter = table.GroupFooter.Clone
            mGroupFooter.Table = Me
        End If
        If table.GroupHeader Is Nothing Then
            mGroupHeader = Nothing
        Else
            mGroupHeader = table.GroupHeader.Clone
            mGroupHeader.Table = Me
        End If
        mImage = table.Image
        mInitialPageNumber = table.InitialPageNumber
        mKeepTogether = table.KeepTogether
        mName = mDataTable.TableName
        mNavigator = table.Navigator
        If table.PageFooter Is Nothing Then
            mPageFooter = Nothing
        Else
            mPageFooter = table.PageFooter.Clone
            mPageFooter.Table = Me
        End If
        If table.PageHeader Is Nothing Then
            mPageHeader = Nothing
        Else
            mPageHeader = table.PageHeader.Clone
            mPageHeader.Table = Me
        End If
        mPageSize = table.PageSize
        mParentRelations = New Model10.RelationCollection(Nothing)
        mPresentation = Nothing
        mPrintable = table.Printable
        mRetrieveIdentity = table.RetrieveIdentity
        mRowPosition = table.RowPosition
        mSize = table.Size
        mSort = table.Sort
        mWidth = table.Width

        mDisplayGrid = table.DisplayGrid
        mDisplayRulers = table.DisplayRulers
        mGridSize = table.GridSize
        mLocked = table.Locked
        mSnapToGrid = table.SnapToGrid

        mLandscape = table.Landscape
        mPaperKind = table.PaperKind
        mPrinterName = table.PrinterName
        mSourceKind = table.SourceKind

        mBottomMargin = table.BottomMargin
        mLeftMargin = table.LeftMargin
        mRightMargin = table.RightMargin
        mTopMargin = table.TopMargin

        mOleDbDataAdapter = Nothing

        If table.DeleteCommand Is Nothing Then
            mDeleteCommand = Nothing
        Else
            mDeleteCommand = New Model10.Command(table.DeleteCommand)
            mDeleteCommand.Table = Me
        End If
        If table.InsertCommand Is Nothing Then
            mInsertCommand = Nothing
        Else
            mInsertCommand = New Model10.Command(table.InsertCommand)
            mInsertCommand.Table = Me
        End If
        If table.SelectCommand Is Nothing Then
            mSelectCommand = Nothing
        Else
            mSelectCommand = New Model10.Command(table.SelectCommand)
            mSelectCommand.Table = Me
        End If
        If table.UpdateCommand Is Nothing Then
            mUpdateCommand = Nothing
        Else
            mUpdateCommand = New Model10.Command(table.UpdateCommand)
            mUpdateCommand.Table = Me
        End If
    End Sub

    <ComponentModel.Browsable(False)> Public Property DataTable() As DataTable
        Get
            If mDataTable Is Nothing Then
                If Not mPresentation Is Nothing Then Return mPresentation.DataSet.Tables(mName)
            Else
                Return mDataTable
            End If
        End Get
        Set(ByVal Value As DataTable)
            mDataTable = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Field() As String
        Get
            Return mField
        End Get
        Set(ByVal Value As String)
            mField = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Filled() As Boolean
        Get
            Return mFilled
        End Get
        Set(ByVal Value As Boolean)
            mFilled = Value
        End Set
    End Property

    Public Property AutoFill() As Boolean
        Get
            Return mAutoFill
        End Get
        Set(ByVal Value As Boolean)
            mAutoFill = Value
        End Set
    End Property
    Public Property AutoSave() As Boolean
        Get
            Return mAutoSave
        End Get
        Set(ByVal Value As Boolean)
            mAutoSave = Value
        End Set
    End Property
    Public Property Band() As Boolean
        Get
            Return mBand
        End Get
        Set(ByVal Value As Boolean)
            mBand = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public ReadOnly Property ChildRelations() As Model10.RelationCollection
        Get
            Return mChildRelations
        End Get
    End Property
    <ComponentModel.Browsable(False)> Public ReadOnly Property Computes() As Model10.ComputeCollection
        Get
            Return mComputes
        End Get
    End Property
    <ComponentModel.Browsable(False)> Public Property DataBindings() As Model10.DataBindingCollection
        Get
            Return mDataBindings
        End Get
        Set(ByVal Value As Model10.DataBindingCollection)
            mDataBindings = Value
        End Set
    End Property
    Public Property DataField() As String
        Get
            Return mDataField
        End Get
        Set(ByVal Value As String)
            mDataField = Value
        End Set
    End Property
    '<ComponentModel.Browsable(False)> Public ReadOnly Property Defaults() As Model10.DefaultCollection
    '    Get
    '        Return mDefaults
    '    End Get
    'End Property
    Public Property DesignerMode() As Model10.DesignerMode
        Get
            Return mDesignerMode
        End Get
        Set(ByVal Value As Model10.DesignerMode)
            mDesignerMode = Value
        End Set
    End Property
    Public Property ErrorField() As String
        Get
            Return mErrorField
        End Get
        Set(ByVal Value As String)
            mErrorField = Value
        End Set
    End Property
    Public Property FillOnce() As Boolean
        Get
            Return mFillOnce
        End Get
        Set(ByVal Value As Boolean)
            mFillOnce = Value
        End Set
    End Property
    Public Property Filter() As String
        Get
            Return mFilter
        End Get
        Set(ByVal Value As String)
            mFilter = Value
        End Set
    End Property
    Public Property FoxProField() As String
        Get
            Return mFoxProField
        End Get
        Set(ByVal Value As String)
            mFoxProField = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Detail() As Model10.Detail
        Get
            Return mDetail
        End Get
        Set(ByVal Value As Model10.Detail)
            mDetail = Value
            If Not Value Is Nothing Then mDetail.Table = Me
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property GroupFooter() As Model10.Grouper
        Get
            Return mGroupFooter
        End Get
        Set(ByVal Value As Model10.Grouper)
            mGroupFooter = Value
            If Not Value Is Nothing Then mGroupFooter.Table = Me
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property GroupHeader() As Model10.Grouper
        Get
            Return mGroupHeader
        End Get
        Set(ByVal Value As Model10.Grouper)
            mGroupHeader = Value
            If Not Value Is Nothing Then mGroupHeader.Table = Me
        End Set
    End Property
    Public Property Image() As Drawing.Image
        Get
            Return mImage
        End Get
        Set(ByVal Value As Drawing.Image)
            mImage = Value
        End Set
    End Property
    Public Property InitialPageNumber() As Integer
        Get
            Return mInitialPageNumber
        End Get
        Set(ByVal Value As Integer)
            mInitialPageNumber = Value
        End Set
    End Property
    Public Property KeepTogether() As Model10.KeepMode
        Get
            Return mKeepTogether
        End Get
        Set(ByVal Value As Model10.KeepMode)
            mKeepTogether = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            DataTable.TableName = Value
            mName = Value
        End Set
    End Property
    Public Property Navigator() As Boolean
        Get
            Return mNavigator
        End Get
        Set(ByVal Value As Boolean)
            mNavigator = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property PageFooter() As Model10.Pager
        Get
            Return mPageFooter
        End Get
        Set(ByVal Value As Model10.Pager)
            mPageFooter = Value
            If Not Value Is Nothing Then mPageFooter.Table = Me
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property PageHeader() As Model10.Pager
        Get
            Return mPageHeader
        End Get
        Set(ByVal Value As Model10.Pager)
            mPageHeader = Value
            If Not Value Is Nothing Then mPageHeader.Table = Me
        End Set
    End Property
    Public Property PageSize() As Integer
        Get
            Return mPageSize
        End Get
        Set(ByVal Value As Integer)
            mPageSize = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public ReadOnly Property ParentRelations() As Model10.RelationCollection
        Get
            Return mParentRelations
        End Get
    End Property
    <ComponentModel.Browsable(False)> Public Property Presentation() As Model10.Presentation
        Get
            Return mPresentation
        End Get
        Set(ByVal Value As Model10.Presentation)
            mPresentation = Value
        End Set
    End Property
    Public Property Printable() As Boolean
        Get
            Return mPrintable
        End Get
        Set(ByVal Value As Boolean)
            mPrintable = Value
        End Set
    End Property
    Public Property RetrieveIdentity() As Boolean
        Get
            Return mRetrieveIdentity
        End Get
        Set(ByVal Value As Boolean)
            mRetrieveIdentity = Value
        End Set
    End Property
    Public Property RowPosition() As Model10.RowStatus
        Get
            Return mRowPosition
        End Get
        Set(ByVal Value As Model10.RowStatus)
            mRowPosition = Value
        End Set
    End Property
    Public Property Size() As Drawing.Size
        Get
            Return mSize
        End Get
        Set(ByVal Value As Drawing.Size)
            mSize = Value
        End Set
    End Property
    Public Property Sort() As String
        Get
            Return mSort
        End Get
        Set(ByVal Value As String)
            mSort = Value
        End Set
    End Property
    Public Property Width() As Integer
        Get
            Return mWidth
        End Get
        Set(ByVal Value As Integer)
            Dim w As Integer = GetMinWidth()
            If Value < w Then Value = w
            mWidth = Value
        End Set
    End Property

    Public Property DisplayGrid() As Boolean Implements IDesigner.DisplayGrid
        Get
            Return mDisplayGrid
        End Get
        Set(ByVal Value As Boolean)
            mDisplayGrid = Value
        End Set
    End Property
    Public Property DisplayRulers() As Boolean Implements IDesigner.DisplayRulers
        Get
            Return mDisplayRulers
        End Get
        Set(ByVal Value As Boolean)
            mDisplayRulers = Value
        End Set
    End Property
    Public Property GridSize() As Drawing.Size Implements IDesigner.GridSize
        Get
            Return mGridSize
        End Get
        Set(ByVal Value As Drawing.Size)
            If Value.Width < 2 Or Value.Height < 2 Then Throw New ArgumentOutOfRangeException
            mGridSize = Value
        End Set
    End Property
    Public Property Locked() As Boolean Implements IDesigner.Locked
        Get
            Return mLocked
        End Get
        Set(ByVal Value As Boolean)
            mLocked = Value
        End Set
    End Property
    Public Property SnapToGrid() As Boolean Implements IDesigner.SnapToGrid
        Get
            Return mSnapToGrid
        End Get
        Set(ByVal Value As Boolean)
            mSnapToGrid = Value
        End Set
    End Property

    Public Property Landscape() As Boolean
        Get
            Return mLandscape
        End Get
        Set(ByVal Value As Boolean)
            mLandscape = Value
        End Set
    End Property
    Public Property PaperKind() As Drawing.Printing.PaperKind
        Get
            Return mPaperKind
        End Get
        Set(ByVal Value As Drawing.Printing.PaperKind)
            mPaperKind = Value
        End Set
    End Property
    Public Property PrinterName() As String
        Get
            Return mPrinterName
        End Get
        Set(ByVal Value As String)
            mPrinterName = Value
        End Set
    End Property
    Public Property SourceKind() As Drawing.Printing.PaperSourceKind
        Get
            Return mSourceKind
        End Get
        Set(ByVal Value As Drawing.Printing.PaperSourceKind)
            mSourceKind = Value
        End Set
    End Property

    Public Property LeftMargin() As Integer
        Get
            Return mLeftMargin
        End Get
        Set(ByVal Value As Integer)
            mLeftMargin = Value
        End Set
    End Property
    Public Property TopMargin() As Integer
        Get
            Return mTopMargin
        End Get
        Set(ByVal Value As Integer)
            mTopMargin = Value
        End Set
    End Property
    Public Property RightMargin() As Integer
        Get
            Return mRightMargin
        End Get
        Set(ByVal Value As Integer)
            mRightMargin = Value
        End Set
    End Property
    Public Property BottomMargin() As Integer
        Get
            Return mBottomMargin
        End Get
        Set(ByVal Value As Integer)
            mBottomMargin = Value
        End Set
    End Property

    Public Sub ReadPageSettings(ByVal pageSettings As Drawing.Printing.PageSettings)
        mLandscape = pageSettings.Landscape
        mPrinterName = pageSettings.PrinterSettings.PrinterName

        mBottomMargin = pageSettings.Margins.Bottom
        mLeftMargin = pageSettings.Margins.Left
        mRightMargin = pageSettings.Margins.Right
        mTopMargin = pageSettings.Margins.Top

        If pageSettings.PrinterSettings.IsValid Then
            mPaperKind = pageSettings.PaperSize.Kind
            mSourceKind = pageSettings.PaperSource.Kind
        End If
    End Sub
    Public Sub WritePageSettings(ByVal pageSettings As Drawing.Printing.PageSettings)
        pageSettings.Landscape = mLandscape

        pageSettings.Margins.Bottom = mBottomMargin
        pageSettings.Margins.Left = mLeftMargin
        pageSettings.Margins.Right = mRightMargin
        pageSettings.Margins.Top = mTopMargin

        Dim pName As String = pageSettings.PrinterSettings.PrinterName
        If Not pageSettings.PrinterSettings.IsValid Then pageSettings.PrinterSettings.PrinterName = pName
        If pageSettings.PrinterSettings.IsValid Then
            Dim j As Integer
            For j = 0 To pageSettings.PrinterSettings.PaperSizes.Count - 1
                If pageSettings.PrinterSettings.PaperSizes(j).Kind.Equals(mPaperKind) Then
                    pageSettings.PaperSize = pageSettings.PrinterSettings.PaperSizes(j)
                    Exit For
                End If
            Next
            For j = 0 To pageSettings.PrinterSettings.PaperSources.Count - 1
                If pageSettings.PrinterSettings.PaperSources(j).Kind.Equals(mSourceKind) Then
                    pageSettings.PaperSource = pageSettings.PrinterSettings.PaperSources(j)
                    Exit For
                End If
            Next
        End If
    End Sub

    Public Sub UpdateForward()
        OleDbDataAdapter.Update(DataTable) 'mDataTable

        Dim rel As Model10.Relation
        For Each rel In mChildRelations
            If Not rel.ChildTable Is Me Then rel.ChildTable.UpdateForward()
        Next : rel = Nothing
    End Sub
    Public Sub UpdateBackward()
        Dim rel As Model10.Relation
        For Each rel In mChildRelations
            If Not rel.ChildTable Is Me Then rel.ChildTable.UpdateBackward()
        Next : rel = Nothing

        OleDbDataAdapter.Update(DataTable) 'mDataTable
    End Sub


    <ComponentModel.Browsable(False)> Public Property OleDbDataAdapter() As OleDb.OleDbDataAdapter
        Get
            If mOleDbDataAdapter Is Nothing Then
                mOleDbDataAdapter = New OleDb.OleDbDataAdapter
                If Not mSelectCommand Is Nothing Then mOleDbDataAdapter.SelectCommand = mSelectCommand.OleDbCommand
                If Not mInsertCommand Is Nothing Then mOleDbDataAdapter.InsertCommand = mInsertCommand.OleDbCommand
                If Not mUpdateCommand Is Nothing Then mOleDbDataAdapter.UpdateCommand = mUpdateCommand.OleDbCommand
                If Not mDeleteCommand Is Nothing Then mOleDbDataAdapter.DeleteCommand = mDeleteCommand.OleDbCommand
            End If
            Return mOleDbDataAdapter
        End Get
        Set(ByVal Value As OleDb.OleDbDataAdapter)
            mOleDbDataAdapter = Value
        End Set
    End Property

    <ComponentModel.Browsable(False)> Public Property DeleteCommand() As Model10.Command
        Get
            Return mDeleteCommand
        End Get
        Set(ByVal Value As Model10.Command)
            If Value Is Nothing Then
                OleDbDataAdapter.DeleteCommand = Nothing
            Else
                Value.Table = Me
                OleDbDataAdapter.DeleteCommand = Value.OleDbCommand
            End If
            mDeleteCommand = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property InsertCommand() As Model10.Command
        Get
            Return mInsertCommand
        End Get
        Set(ByVal Value As Model10.Command)
            If Value Is Nothing Then
                OleDbDataAdapter.InsertCommand = Nothing
            Else
                Value.Table = Me
                OleDbDataAdapter.InsertCommand = Value.OleDbCommand
            End If
            mInsertCommand = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property SelectCommand() As Model10.Command
        Get
            Return mSelectCommand
        End Get
        Set(ByVal Value As Model10.Command)
            If Value Is Nothing Then
                OleDbDataAdapter.SelectCommand = Nothing
            Else
                Value.Table = Me
                OleDbDataAdapter.SelectCommand = Value.OleDbCommand
            End If
            mSelectCommand = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property UpdateCommand() As Model10.Command
        Get
            Return mUpdateCommand
        End Get
        Set(ByVal Value As Model10.Command)
            If Value Is Nothing Then
                OleDbDataAdapter.UpdateCommand = Nothing
            Else
                Value.Table = Me
                OleDbDataAdapter.UpdateCommand = Value.OleDbCommand
            End If
            mUpdateCommand = Value
        End Set
    End Property

    Private Sub mOleDbDataAdapter_RowUpdated(ByVal sender As Object, ByVal e As System.Data.OleDb.OleDbRowUpdatedEventArgs) Handles mOleDbDataAdapter.RowUpdated
        If Not mInsertCommand Is Nothing And mRetrieveIdentity Then
            If e.StatementType = StatementType.Insert And e.RecordsAffected = 1 Then
                Static idCMD As OleDb.OleDbCommand
                If idCMD Is Nothing Then
                    idCMD = New OleDb.OleDbCommand("SELECT @@IDENTITY", mInsertCommand.Connection.OleDbConnection)
                    idCMD.Prepare()
                Else
                    If Not idCMD.Connection Is mInsertCommand.Connection.OleDbConnection Then
                        idCMD.Connection = mInsertCommand.Connection.OleDbConnection
                    End If
                End If

                Dim col As DataColumn
                For Each col In e.Row.Table.Columns
                    If col.AutoIncrement Then
                        e.Row(col) = CInt(idCMD.ExecuteScalar())
                        Exit For
                    End If
                Next : col = Nothing
            End If
        End If

        Dim dField As String = mErrorField.Trim
        If dField <> "" Then
            Try
                If e.Errors Is Nothing Then
                    e.Row(dField) = Nothing
                Else
                    e.Row(dField) = e.Errors.Message
                End If
            Catch ex As Exception
            End Try
        End If
    End Sub
    Private Sub mOleDbDataAdapter_FillError(ByVal sender As Object, ByVal e As System.Data.FillErrorEventArgs) Handles mOleDbDataAdapter.FillError
        e.[Continue] = True
    End Sub
End Class
<Serializable()> Public Class Connection
    Implements ISerializable

    <NonSerialized()> Private mOleDbConnection As OleDb.OleDbConnection

    Private mConnectionString As String
    Private mName As String
    Private mPresentation As Model10.Presentation

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mConnectionString = info.GetString("ConnectionString")
        mName = info.GetString("Name")
        mPresentation = CType(info.GetValue("Presentation", GetType(Model10.Presentation)), Model10.Presentation)
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("ConnectionString", mConnectionString)
        info.AddValue("Name", mName)
        info.AddValue("Presentation", mPresentation)
    End Sub

    Public Overrides Function ToString() As String
        Return mName
    End Function
    Public Sub New()
        MyBase.New()

        mOleDbConnection = Nothing

        mConnectionString = ""
        mName = ""
        mPresentation = Nothing
    End Sub
    Public Sub New(ByVal connection As Model10.Connection)
        MyBase.New()

        mOleDbConnection = Nothing

        mConnectionString = connection.ConnectionString
        mName = connection.Name
        mPresentation = Nothing
    End Sub

    <ComponentModel.Browsable(False)> Public Property OleDbConnection() As OleDb.OleDbConnection
        Get
            If mOleDbConnection Is Nothing Then
                mOleDbConnection = New OleDb.OleDbConnection(mConnectionString)
            End If
            Return mOleDbConnection
        End Get
        Set(ByVal Value As OleDb.OleDbConnection)
            mOleDbConnection = Value
        End Set
    End Property

    Public Property ConnectionString() As String
        Get
            Return mConnectionString
        End Get
        Set(ByVal Value As String)
            OleDbConnection.ConnectionString = Value
            mConnectionString = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            Value = Value.Trim
            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
            If Not mPresentation Is Nothing Then
                Dim cnn As Model10.Connection
                For Each cnn In mPresentation.Connections
                    If Not cnn Is Me Then
                        If cnn.Name.ToLower = Value.ToLower Then
                            Throw New System.Data.DuplicateNameException("A Connection named '" & Value & "' already belongs to this Presentation.")
                        End If
                    End If
                Next : cnn = Nothing
            End If
            mName = Value
        End Set
    End Property
    <ComponentModel.Browsable(False)> Public Property Presentation() As Model10.Presentation
        Get
            Return mPresentation
        End Get
        Set(ByVal Value As Model10.Presentation)
            mPresentation = Value
        End Set
    End Property
End Class
<Serializable()> Public Class Presentation
    Implements ISerializable
    Implements IDesigner

    Private ReadOnly mConnections As Model10.ConnectionCollection
    Private mDataSet As DataSet
    <NonSerialized()> Private mDataViewManager As DataViewManager
    <NonSerialized()> Private mId As Integer
    <NonSerialized()> Private mName As String
    <NonSerialized()> Private mPAth As String
    Private ReadOnly mRelations As Model10.RelationCollection
    Private mRightToLeft As Windows.Forms.RightToLeft
    Private mStartMode As Model10.StartMode
    Private mStartupTable As Model10.Table
    Private ReadOnly mTables As Model10.TableCollection

    Private mDisplayGrid As Boolean
    Private mDisplayRulers As Boolean
    Private mGridSize As Drawing.Size
    Private mLocked As Boolean
    Private mSnapToGrid As Boolean

    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
        On Error Resume Next

        mConnections = CType(info.GetValue("Connections", GetType(Model10.ConnectionCollection)), Model10.ConnectionCollection)
        mDataSet = CType(info.GetValue("DataSet", GetType(DataSet)), DataSet)
        mRelations = CType(info.GetValue("Relations", GetType(Model10.RelationCollection)), Model10.RelationCollection)
        mRightToLeft = CType(info.GetInt32("RightToLeft"), Windows.Forms.RightToLeft)
        mStartMode = CType(info.GetByte("StartMode"), Model10.StartMode)
        mStartupTable = CType(info.GetValue("StartupTable", GetType(Model10.Table)), Model10.Table)
        mTables = CType(info.GetValue("Tables", GetType(Model10.TableCollection)), Model10.TableCollection)

        mDisplayGrid = info.GetBoolean("DisplayGrid")
        mDisplayRulers = info.GetBoolean("DisplayRulers")
        mGridSize = CType(info.GetValue("GridSize", GetType(Drawing.Size)), Drawing.Size)
        mLocked = info.GetBoolean("Locked")
        mSnapToGrid = info.GetBoolean("SnapToGrid")
    End Sub
    Protected Overridable Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext) Implements ISerializable.GetObjectData
        info.AddValue("Connections", mConnections)
        info.AddValue("DataSet", mDataSet)
        info.AddValue("Relations", mRelations)
        info.AddValue("RightToLeft", mRightToLeft)
        info.AddValue("StartMode", mStartMode)
        info.AddValue("StartupTable", mStartupTable)
        info.AddValue("Tables", mTables)

        info.AddValue("DisplayGrid", mDisplayGrid)
        info.AddValue("DisplayRulers", mDisplayRulers)
        info.AddValue("GridSize", mGridSize)
        info.AddValue("Locked", mLocked)
        info.AddValue("SnapToGrid", mSnapToGrid)
    End Sub

    Public Function GetNewConnectionName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mConnections.Count - 1
            Dim cnn As Model10.Connection = CType(mConnections(i), Model10.Connection)
            If cnn.Name = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            cnn = Nothing
        Loop
        Return str
    End Function
    Public Function GetNewRelationName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mRelations.Count - 1
            Dim rlt As Model10.Relation = mRelations(i)
            If rlt.DataRelation.RelationName = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            rlt = Nothing
        Loop
        Return str
    End Function
    Public Function GetNewTableName(ByVal name As String) As String
        Dim str As String
        Dim i, j As Integer
        j = 1 : str = name & j.ToString
        Do Until i > mTables.Count - 1
            Dim tbl As Model10.Table = mTables(i)
            If tbl.DataTable.TableName = str Then
                j += 1
                str = name & j.ToString

                i = 0
            Else
                i += 1
            End If
            tbl = Nothing
        Loop
        Return str
    End Function
    Public Overrides Function ToString() As String
        Return mDataSet.DataSetName
    End Function
    Public Sub New()
        MyBase.New()

        mConnections = New Model10.ConnectionCollection(Me)
        mDataSet = New DataSet
        mRelations = New Model10.RelationCollection(Me)
        mRightToLeft = Windows.Forms.RightToLeft.Inherit
        mStartMode = StartMode.Undefined
        mStartupTable = Nothing
        mTables = New Model10.TableCollection(Me)

        mDisplayGrid = True
        mDisplayRulers = True
        mGridSize = New Drawing.Size(6, 6)
        mLocked = False
        mSnapToGrid = True
    End Sub

    Public Sub OpenConnections()
        Dim cn As Model10.Connection
        For Each cn In mConnections
            cn.OleDbConnection.Open()
        Next : cn = Nothing
    End Sub
    Public Sub CloseConnections()
        Dim cn As Model10.Connection
        For Each cn In mConnections
            cn.OleDbConnection.Close()
        Next : cn = Nothing
    End Sub

    Public Sub FillConnection(ByVal connections As Model10.ConnectionCollection)
        Dim tbl As Model10.Table
        For Each tbl In mTables
            FillCommand(tbl.SelectCommand, connections)
            FillCommand(tbl.InsertCommand, connections)
            FillCommand(tbl.UpdateCommand, connections)
            FillCommand(tbl.DeleteCommand, connections)
        Next : tbl = Nothing
    End Sub
    Private Sub FillCommand(ByVal command As Model10.Command, ByVal connections As Model10.ConnectionCollection)
        If Not command Is Nothing Then
            If Not command.Connection Is Nothing Then
                Dim cnn As Model10.Connection
                For Each cnn In connections
                    If cnn.Name.Trim.ToLower = command.Connection.Name.Trim.ToLower Then
                        command.Connection.ConnectionString = cnn.ConnectionString
                    End If
                Next : cnn = Nothing
            End If
        End If
    End Sub

    Public Sub UpdateForward()
        Dim tbl As Model10.Table
        For Each tbl In mTables
            Dim foo As Boolean = False

            Dim rel As Model10.Relation
            For Each rel In mRelations
                If rel.ChildTable Is tbl Then
                    foo = True
                    Exit For
                End If
            Next : rel = Nothing

            If Not foo Then tbl.UpdateForward()
        Next : tbl = Nothing
    End Sub
    Public Sub UpdateBackward()
        Dim tbl As Model10.Table
        For Each tbl In mTables
            Dim foo As Boolean = False

            Dim rel As Model10.Relation
            For Each rel In mRelations
                If rel.ChildTable Is tbl Then
                    foo = True
                    Exit For
                End If
            Next : rel = Nothing

            If Not foo Then tbl.UpdateBackward()
        Next : tbl = Nothing
    End Sub

    Public ReadOnly Property Connections() As Model10.ConnectionCollection
        Get
            Return mConnections
        End Get
    End Property
    <ComponentModel.Browsable(False)> Public Property DataSet() As DataSet
        Get
            Return mDataSet
        End Get
        Set(ByVal Value As DataSet)
            mDataSet = Value
        End Set
    End Property
    Public Property DataViewManager() As DataViewManager
        Get
            Return mDataViewManager
        End Get
        Set(ByVal Value As DataViewManager)
            mDataViewManager = Value
        End Set
    End Property
    Public Property Id() As Integer
        Get
            Return mId
        End Get
        Set(ByVal Value As Integer)
            mId = Value
        End Set
    End Property
    Public Property Name() As String
        Get
            Return mName
        End Get
        Set(ByVal Value As String)
            mName = Value
        End Set
    End Property
    Public Property Path() As String
        Get
            Return mPAth
        End Get
        Set(ByVal Value As String)
            mPAth = Value
        End Set
    End Property
    Public ReadOnly Property Relations() As Model10.RelationCollection
        Get
            Return mRelations
        End Get
    End Property
    Public Property RightToLeft() As Windows.Forms.RightToLeft
        Get
            Return mRightToLeft
        End Get
        Set(ByVal Value As Windows.Forms.RightToLeft)
            mRightToLeft = Value
        End Set
    End Property
    Public Property StartMode() As Model10.StartMode
        Get
            Return mStartMode
        End Get
        Set(ByVal Value As Model10.StartMode)
            mStartMode = Value
        End Set
    End Property
    Public Property StartupTable() As Model10.Table
        Get
            Return mStartupTable
        End Get
        Set(ByVal Value As Model10.Table)
            mStartupTable = Value
        End Set
    End Property
    Public ReadOnly Property Tables() As Model10.TableCollection
        Get
            Return mTables
        End Get
    End Property

    Public Property DisplayGrid() As Boolean Implements IDesigner.DisplayGrid
        Get
            Return mDisplayGrid
        End Get
        Set(ByVal Value As Boolean)
            mDisplayGrid = Value
        End Set
    End Property
    Public Property DisplayRulers() As Boolean Implements IDesigner.DisplayRulers
        Get
            Return mDisplayRulers
        End Get
        Set(ByVal Value As Boolean)
            mDisplayRulers = Value
        End Set
    End Property
    Public Property GridSize() As Drawing.Size Implements IDesigner.GridSize
        Get
            Return mGridSize
        End Get
        Set(ByVal Value As Drawing.Size)
            mGridSize = Value
        End Set
    End Property
    Public Property Locked() As Boolean Implements IDesigner.Locked
        Get
            Return mLocked
        End Get
        Set(ByVal Value As Boolean)
            mLocked = Value
        End Set
    End Property
    Public Property SnapToGrid() As Boolean Implements IDesigner.SnapToGrid
        Get
            Return mSnapToGrid
        End Get
        Set(ByVal Value As Boolean)
            mSnapToGrid = Value
        End Set
    End Property
End Class

<Serializable()> Public Class MemberCollection
    Inherits CollectionBase

    Private mSection As Model10.Section

    Public Function Contains(ByVal value As Model10.Member) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal section As Model10.Section)
        MyBase.New()

        mSection = section
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Member
        Get
            Return CType(MyBase.List.Item(index), Model10.Member)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Member
        Get
            Dim str As String = name.Trim.ToLower

            Dim mem As Model10.Member
            For Each mem In list
                If mem.Name.Trim.ToLower = str Then
                    Return mem
                End If
            Next : mem = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Member, ByVal x As Integer, ByVal y As Integer)
        If Not mSection Is Nothing Then
            If value.Name = "" Then
                value.Name = mSection.GetNewMemberName(TypeName(value))
            Else
                Dim mem As Model10.Member
                For Each mem In mSection.Members
                    If mem.Name.ToLower = value.Name.ToLower Then value.Name = mSection.GetNewMemberName("Member") 'Throw New System.Data.DuplicateNameException("A Member named '" & value.Name & "' already belongs to this Section.")
                Next : mem = Nothing
            End If
            If Not value.Section Is Nothing Then value.Section.Members.Remove(value)
            value.Section = mSection
        End If
        MyBase.List.Add(value)
        Dim p As Drawing.Point = value.Location
        p.Offset(x, y)
        value.Location = p
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Member)
        If Not mSection Is Nothing Then
            If value.Name = "" Then
                value.Name = mSection.GetNewMemberName(TypeName(value))
            Else
                Dim mem As Model10.Member
                For Each mem In mSection.Members
                    If mem.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Member named '" & value.Name & "' already belongs to this Section.")
                Next : mem = Nothing
            End If
            If Not value.Section Is Nothing Then value.Section.Members.Remove(value)
            value.Section = mSection
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Member)
        If Not mSection Is Nothing Then
            value.Section = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
<Serializable()> Public Class SeriesCollection
    Inherits CollectionBase

    Private mChart As Model10.Chart

    Public Function Contains(ByVal value As Model10.Series) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal chart As Model10.Chart)
        MyBase.New()

        mChart = chart
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Series
        Get
            Return CType(MyBase.List.Item(index), Model10.Series)
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Series)
        If Not mChart Is Nothing Then
            If Not value.Chart Is Nothing Then value.Chart.Series.Remove(value)
            value.Chart = mChart
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Series)
        If Not mChart Is Nothing Then
            If Not value.Chart Is Nothing Then value.Chart.Series.Remove(value)
            value.Chart = mChart
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Series)
        If Not mChart Is Nothing Then
            value.Chart = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
<Serializable()> Public Class DataBindingCollection
    Inherits CollectionBase

    Private mTable As Model10.Table

    Public Function Contains(ByVal value As Model10.DataBinding) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal table As Model10.Table)
        MyBase.New()

        mTable = table
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.DataBinding
        Get
            Return CType(MyBase.List.Item(index), Model10.DataBinding)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.DataBinding
        Get
            Dim str As String = name.Trim.ToLower

            Dim db As Model10.DataBinding
            For Each db In list
                If db.Name.Trim.ToLower = str Then
                    Return db
                End If
            Next : db = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.DataBinding)
        If Not mTable Is Nothing Then
            If value.Name = "" Then
                Throw New System.Data.DuplicateNameException("Name can not be empty.")
            Else
                Dim db As Model10.DataBinding
                For Each db In mTable.DataBindings
                    If db.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A DataBinding named '" & value.Name & "' already belongs to this Table.")
                Next : db = Nothing
            End If
            If Not value.Table Is Nothing Then value.Table.DataBindings.Remove(value)
            value.Table = mTable
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.DataBinding)
        If Not mTable Is Nothing Then
            If value.Name = "" Then
                Throw New System.Data.DuplicateNameException("Name can not be empty.")
            Else
                Dim db As Model10.DataBinding
                For Each db In mTable.DataBindings
                    If db.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A DataBinding named '" & value.Name & "' already belongs to this Table.")
                Next : db = Nothing
            End If
            If Not value.Table Is Nothing Then value.Table.DataBindings.Remove(value)
            value.Table = mTable
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.DataBinding)
        If Not mTable Is Nothing Then
            value.Table = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
<Serializable()> Public Class RelationCollection
    Inherits CollectionBase

    Private mPresentation As Model10.Presentation

    Private Sub Strip(ByVal members As Model10.MemberCollection, ByVal relation As Model10.Relation)
        Dim mem As Model10.Member
        For Each mem In members
            If TypeOf mem Is Model10.Drilldown Then
                Dim drl As Model10.Drilldown = CType(mem, Model10.Drilldown)
                If drl.ChildRelation Is relation Then drl.ChildRelation = Nothing
                drl = Nothing
            ElseIf TypeOf mem Is Model10.Lookup Then
                Dim lkp As Model10.Lookup = CType(mem, Model10.Lookup)
                If lkp.ParentRelation Is relation Then lkp.ParentRelation = Nothing
                lkp = Nothing
            ElseIf TypeOf mem Is Model10.Label Then
                Dim lbl As Model10.Label = CType(mem, Model10.Label)
                If lbl.ParentRelation Is relation Then lbl.ParentRelation = Nothing
                lbl = Nothing
            ElseIf TypeOf mem Is Model10.CheckBox Then
                Dim chk As Model10.CheckBox = CType(mem, Model10.CheckBox)
                If chk.ParentRelation Is relation Then chk.ParentRelation = Nothing
                chk = Nothing
            ElseIf TypeOf mem Is Model10.PictureBox Then
                Dim pic As Model10.PictureBox = CType(mem, Model10.PictureBox)
                If pic.ParentRelation Is relation Then pic.ParentRelation = Nothing
                pic = Nothing
            ElseIf TypeOf mem Is Model10.Expand Then
                Dim exp As Model10.Expand = CType(mem, Model10.Expand)
                If exp.ChildRelation Is relation Then exp.ChildRelation = Nothing
                If exp.ParentRelation Is relation Then exp.ParentRelation = Nothing
                exp = Nothing
            ElseIf TypeOf mem Is Model10.Sub Then
                Dim sb As Model10.Sub = CType(mem, Model10.Sub)
                If sb.ChildRelation Is relation Then sb.ChildRelation = Nothing
                sb = Nothing
            ElseIf TypeOf mem Is Model10.Chart Then
                Dim ch As Model10.Chart = CType(mem, Model10.Chart)
                If ch.ChildRelation Is relation Then ch.ChildRelation = Nothing
                ch = Nothing
            End If
        Next : mem = Nothing
    End Sub

    Public Function Contains(ByVal value As Model10.Relation) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal presentation As Model10.Presentation)
        MyBase.New()

        mPresentation = presentation
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Relation
        Get
            Return CType(MyBase.List.Item(index), Model10.Relation)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Relation
        Get
            Dim str As String = name.Trim.ToLower

            Dim rlt As Model10.Relation
            For Each rlt In list
                If rlt.Name.Trim.ToLower = str Then
                    Return rlt
                End If
            Next : rlt = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Relation)
        If Not mPresentation Is Nothing Then
            mPresentation.DataSet.Relations.Add(value.DataRelation) : value.Name = value.DataRelation.RelationName
            If Not value.Presentation Is Nothing Then value.Presentation.Relations.Remove(value)
            value.Presentation = mPresentation
            value.ChildTable.ParentRelations.Add(value)
            value.ParentTable.ChildRelations.Add(value)
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Relation)
        If Not mPresentation Is Nothing Then
            mPresentation.DataSet.Relations.Add(value.DataRelation) : value.Name = value.DataRelation.RelationName
            If Not value.Presentation Is Nothing Then value.Presentation.Relations.Remove(value)
            value.Presentation = mPresentation
            value.ChildTable.ParentRelations.Add(value)
            value.ParentTable.ChildRelations.Add(value)
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Relation)
        If Not mPresentation Is Nothing Then
            mPresentation.DataSet.Relations.Remove(value.DataRelation)
            value.Presentation = Nothing
            value.ChildTable.ParentRelations.Remove(value)
            value.ParentTable.ChildRelations.Remove(value)

            Dim tbl As Model10.Table
            For Each tbl In mPresentation.Tables
                If Not tbl.PageHeader Is Nothing Then Strip(tbl.PageHeader.Members, value)
                If Not tbl.GroupHeader Is Nothing Then Strip(tbl.GroupHeader.Members, value)
                If Not tbl.Detail Is Nothing Then Strip(tbl.Detail.Members, value)
                If Not tbl.GroupFooter Is Nothing Then Strip(tbl.GroupFooter.Members, value)
                If Not tbl.PageFooter Is Nothing Then Strip(tbl.PageFooter.Members, value)
            Next : tbl = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
<Serializable()> Public Class ParameterCollection
    Inherits CollectionBase

    Private mCommand As Model10.Command

    Public Function Contains(ByVal value As Model10.Parameter) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal command As Model10.Command)
        MyBase.New()

        mCommand = command
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Parameter
        Get
            Return CType(MyBase.List.Item(index), Model10.Parameter)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Parameter
        Get
            Dim str As String = name.Trim.ToLower

            Dim prm As Model10.Parameter
            For Each prm In list
                If prm.Name.Trim.ToLower = str Then
                    Return prm
                End If
            Next : prm = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Parameter)
        If Not mCommand Is Nothing Then
            If value.Name = "" Then
                value.Name = mCommand.GetNewParameterName(TypeName(value))
            Else
                Dim prm As Model10.Parameter
                For Each prm In mCommand.Parameters
                    If prm.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Member named '" & value.Name & "' already belongs to this Section.")
                Next : prm = Nothing
            End If
            mCommand.OleDbCommand.Parameters.Add(value.OleDbParameter)
            If Not value.Command Is Nothing Then value.Command.Parameters.Remove(value)
            value.Command = mCommand
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Parameter)
        If Not mCommand Is Nothing Then
            If value.Name = "" Then
                value.Name = mCommand.GetNewParameterName(TypeName(value))
            Else
                Dim prm As Model10.Parameter
                For Each prm In mCommand.Parameters
                    If prm.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Member named '" & value.Name & "' already belongs to this Section.")
                Next : prm = Nothing
            End If
            mCommand.OleDbCommand.Parameters.Add(value.OleDbParameter)
            If Not value.Command Is Nothing Then value.Command.Parameters.Remove(value)
            value.Command = mCommand
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Parameter)
        If Not mCommand Is Nothing Then
            mCommand.OleDbCommand.Parameters.Remove(value.OleDbParameter)
            value.Command = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
'<Serializable()> Public Class DefaultCollection
'    Inherits CollectionBase

'    Private mTable As Model10.Table

'    Public Function Contains(ByVal value As Model10.Default) As Boolean
'        Return MyBase.List.Contains(value)
'    End Function
'    Public Sub New(ByVal table As Model10.Table)
'        MyBase.New()

'        mTable = table
'    End Sub

'    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Default
'        Get
'            Return CType(MyBase.List.Item(index), Model10.Default)
'        End Get
'    End Property
'    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Default
'        Get
'            Dim str As String = name.Trim.ToLower

'            Dim ds As Model10.Default
'            For Each ds In list
'                If ds.Name.Trim.ToLower = str Then
'                    Return ds
'                End If
'            Next : ds = Nothing
'        End Get
'    End Property
'    Public Sub Add(ByVal value As Model10.Default)
'        If Not mTable Is Nothing Then
'            If value.Name = "" Then
'                Throw New System.Data.DuplicateNameException("Name can not be empty.")
'            Else
'                Dim prt As Model10.Default
'                For Each prt In mTable.Defaults
'                    If prt.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Default named '" & value.Name & "' already belongs to this Table.")
'                Next : prt = Nothing
'            End If
'            If Not value.Table Is Nothing Then value.Table.Defaults.Remove(value)
'            value.Table = mTable
'        End If
'        MyBase.List.Add(value)
'    End Sub
'    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Default)
'        If Not mTable Is Nothing Then
'            If value.Name = "" Then
'                Throw New System.Data.DuplicateNameException("Name can not be empty.")
'            Else
'                Dim prt As Model10.Default
'                For Each prt In mTable.Defaults
'                    If prt.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Default named '" & value.Name & "' already belongs to this Table.")
'                Next : prt = Nothing
'            End If
'            If Not value.Table Is Nothing Then value.Table.Defaults.Remove(value)
'            value.Table = mTable
'        End If
'        MyBase.List.Insert(index, value)
'    End Sub
'    Public Sub Remove(ByVal value As Model10.Default)
'        If Not mTable Is Nothing Then
'            value.Table = Nothing
'        End If
'        MyBase.List.Remove(value)
'    End Sub
'End Class
<Serializable()> Public Class ComputeCollection
    Inherits CollectionBase

    Private mTable As Model10.Table

    Private Sub Strip(ByVal members As Model10.MemberCollection, ByVal compute As Model10.Compute)
        Dim mem As Model10.Member
        For Each mem In members
            If TypeOf mem Is Model10.Computer Then
                Dim ctr As Model10.Computer = CType(mem, Model10.Computer)
                If ctr.Compute Is compute Then ctr.Compute = Nothing
                ctr = Nothing
            End If
        Next : mem = Nothing
    End Sub

    Public Function Contains(ByVal value As Model10.Compute) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal table As Model10.Table)
        MyBase.New()

        mTable = table
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Compute
        Get
            Return CType(MyBase.List.Item(index), Model10.Compute)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Compute
        Get
            Dim str As String = name.Trim.ToLower

            Dim cmp As Model10.Compute
            For Each cmp In list
                If cmp.Name.Trim.ToLower = str Then
                    Return cmp
                End If
            Next : cmp = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Compute)
        If Not mTable Is Nothing Then
            If value.Name = "" Then
                value.Name = mTable.GetNewComputeName(TypeName(value))
            Else
                Dim cmp As Model10.Compute
                For Each cmp In mTable.Computes
                    If cmp.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Compute named '" & value.Name & "' already belongs to this Table.")
                Next : cmp = Nothing
            End If
            If Not value.Table Is Nothing Then value.Table.Computes.Remove(value)
            value.Table = mTable
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Compute)
        If Not mTable Is Nothing Then
            If value.Name = "" Then
                value.Name = mTable.GetNewComputeName(TypeName(value))
            Else
                Dim cmp As Model10.Compute
                For Each cmp In mTable.Computes
                    If cmp.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Compute named '" & value.Name & "' already belongs to this Table.")
                Next : cmp = Nothing
            End If
            If Not value.Table Is Nothing Then value.Table.Computes.Remove(value)
            value.Table = mTable
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Compute)
        If Not mTable Is Nothing Then
            value.Table = Nothing

            If Not mTable.PageHeader Is Nothing Then Strip(mTable.PageHeader.Members, value)
            If Not mTable.GroupHeader Is Nothing Then Strip(mTable.GroupHeader.Members, value)
            If Not mTable.Detail Is Nothing Then Strip(mTable.Detail.Members, value)
            If Not mTable.GroupFooter Is Nothing Then Strip(mTable.GroupFooter.Members, value)
            If Not mTable.PageFooter Is Nothing Then Strip(mTable.PageFooter.Members, value)
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
<Serializable()> Public Class TableCollection
    Inherits CollectionBase

    Private mPresentation As Model10.Presentation

    Private Sub Strip(ByVal members As Model10.MemberCollection, ByVal table As Model10.Table)
        Dim mem As Model10.Member
        For Each mem In members
            If TypeOf mem Is Model10.Next Then
                Dim nxt As Model10.Next = CType(mem, Model10.Next)
                If nxt.Table Is table Then nxt.Table = Nothing
                nxt = Nothing
            End If
        Next : mem = Nothing
    End Sub

    Public Function Contains(ByVal value As Model10.Table) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New(ByVal presentation As Model10.Presentation)
        MyBase.New()

        mPresentation = presentation
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Table
        Get
            Return CType(MyBase.List.Item(index), Model10.Table)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Table
        Get
            Dim str As String = name.Trim.ToLower

            Dim tbl As Model10.Table
            For Each tbl In list
                If tbl.Name.Trim.ToLower = str Then
                    Return tbl
                End If
            Next : tbl = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Table)
        If Not mPresentation Is Nothing Then
            mPresentation.DataSet.Tables.Add(value.DataTable) : value.Name = value.DataTable.TableName
            If Not value.Presentation Is Nothing Then value.Presentation.Tables.Remove(value)
            value.Presentation = mPresentation

            If MyBase.List.Count = 0 Then mPresentation.StartupTable = value
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Table)
        If Not mPresentation Is Nothing Then
            mPresentation.DataSet.Tables.Add(value.DataTable) : value.Name = value.DataTable.TableName
            If Not value.Presentation Is Nothing Then value.Presentation.Tables.Remove(value)
            value.Presentation = mPresentation

            If MyBase.List.Count = 0 Then mPresentation.StartupTable = value
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Table)
        If Not mPresentation Is Nothing Then
            mPresentation.DataSet.Tables.Remove(value.DataTable)
            value.Presentation = Nothing
            If mPresentation.StartupTable Is value Then mPresentation.StartupTable = Nothing

            Dim tbl As Model10.Table
            For Each tbl In mPresentation.Tables
                If Not tbl.PageHeader Is Nothing Then Strip(tbl.PageHeader.Members, value)
                If Not tbl.GroupHeader Is Nothing Then Strip(tbl.GroupHeader.Members, value)
                If Not tbl.Detail Is Nothing Then Strip(tbl.Detail.Members, value)
                If Not tbl.GroupFooter Is Nothing Then Strip(tbl.GroupFooter.Members, value)
                If Not tbl.PageFooter Is Nothing Then Strip(tbl.PageFooter.Members, value)
            Next : tbl = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class
<Serializable()> Public Class ConnectionCollection
    Inherits CollectionBase

    Private mPresentation As Model10.Presentation

    Public Function Contains(ByVal value As Model10.Connection) As Boolean
        Return MyBase.List.Contains(value)
    End Function
    Public Sub New()
        MyBase.New()

    End Sub
    Public Sub New(ByVal presentation As Model10.Presentation)
        MyBase.New()

        mPresentation = presentation
    End Sub

    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Connection
        Get
            Return CType(MyBase.List.Item(index), Model10.Connection)
        End Get
    End Property
    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Connection
        Get
            Dim str As String = name.Trim.ToLower

            Dim cnn As Model10.Connection
            For Each cnn In list
                If cnn.Name.Trim.ToLower = str Then
                    Return cnn
                End If
            Next : cnn = Nothing
        End Get
    End Property
    Public Sub Add(ByVal value As Model10.Connection)
        If Not mPresentation Is Nothing Then
            If value.Name = "" Then
                value.Name = mPresentation.GetNewConnectionName(TypeName(value))
            Else
                Dim cnn As Model10.Connection
                For Each cnn In mPresentation.Connections
                    If cnn.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Connection named '" & value.Name & "' already belongs to this Presentation.")
                Next : cnn = Nothing
            End If
            If Not value.Presentation Is Nothing Then value.Presentation.Connections.Remove(value)
            value.Presentation = mPresentation
        End If
        MyBase.List.Add(value)
    End Sub
    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Connection)
        If Not mPresentation Is Nothing Then
            If value.Name = "" Then
                value.Name = mPresentation.GetNewConnectionName(TypeName(value))
            Else
                Dim cnn As Model10.Connection
                For Each cnn In mPresentation.Connections
                    If cnn.Name.ToLower = value.Name.ToLower Then Throw New System.Data.DuplicateNameException("A Connection named '" & value.Name & "' already belongs to this Presentation.")
                Next : cnn = Nothing
            End If
            If Not value.Presentation Is Nothing Then value.Presentation.Connections.Remove(value)
            value.Presentation = mPresentation
        End If
        MyBase.List.Insert(index, value)
    End Sub
    Public Sub Remove(ByVal value As Model10.Connection)
        If Not mPresentation Is Nothing Then
            Dim tbl As Model10.Table
            For Each tbl In mPresentation.Tables
                If Not tbl.SelectCommand Is Nothing Then If tbl.SelectCommand.Connection Is value Then tbl.SelectCommand.Connection = Nothing
                If Not tbl.InsertCommand Is Nothing Then If tbl.InsertCommand.Connection Is value Then tbl.InsertCommand.Connection = Nothing
                If Not tbl.UpdateCommand Is Nothing Then If tbl.UpdateCommand.Connection Is value Then tbl.UpdateCommand.Connection = Nothing
                If Not tbl.DeleteCommand Is Nothing Then If tbl.DeleteCommand.Connection Is value Then tbl.DeleteCommand.Connection = Nothing

                Dim cmp As Model10.Compute
                For Each cmp In tbl.Computes
                    If cmp.Connection Is value Then cmp.Connection = Nothing
                Next : cmp = Nothing
            Next : tbl = Nothing
            value.Presentation = Nothing
        End If
        MyBase.List.Remove(value)
    End Sub
End Class

Public Class Current
    Private mDataRow As DataRowView
    Private mDataView As DataView
    Private mParaRow As DataRowView
    Private mRecordCount As Integer
    Private mRecordNumber As Integer
    Private mRelation As Model10.Relation
    Private mRowData As DataRowView
    Private mTable As Model10.Table

    Private mForm As Model10.Form
    Private mMemberName As String
    Private mReport As Model10.Report

    Public Sub New(ByVal table As Model10.Table, ByVal relation As Model10.Relation, ByVal dataView As DataView, ByVal dataRow As DataRowView, ByVal paraRow As DataRowView, ByVal recordCount As Integer, ByVal recordNumber As Integer)
        mDataRow = dataRow
        mDataView = dataView
        mParaRow = paraRow
        mRecordCount = recordCount
        mRecordNumber = recordNumber
        mRelation = relation
        'mRowData = rowData
        mTable = table

        mForm = Nothing
        mMemberName = ""
        mReport = Nothing
    End Sub

    Public Property DataRow() As DataRowView
        Get
            Return mDataRow
        End Get
        Set(ByVal Value As DataRowView)
            mDataRow = Value
        End Set
    End Property
    Public Property DataView() As DataView
        Get
            Return mDataView
        End Get
        Set(ByVal Value As DataView)
            mDataView = Value
        End Set
    End Property
    Public Property ParaRow() As DataRowView
        Get
            Return mParaRow
        End Get
        Set(ByVal Value As DataRowView)
            mParaRow = Value
        End Set
    End Property
    Public Property RecordCount() As Integer
        Get
            Return mRecordCount
        End Get
        Set(ByVal Value As Integer)
            mRecordCount = Value
        End Set
    End Property
    Public Property RecordNumber() As Integer
        Get
            Return mRecordNumber
        End Get
        Set(ByVal Value As Integer)
            mRecordNumber = Value
        End Set
    End Property
    Public Property Relation() As Model10.Relation
        Get
            Return mRelation
        End Get
        Set(ByVal Value As Model10.Relation)
            mRelation = Value
        End Set
    End Property
    Public Property RowData() As DataRowView
        Get
            Return mRowData
        End Get
        Set(ByVal Value As DataRowView)
            mRowData = Value
        End Set
    End Property
    Public Property Table() As Model10.Table
        Get
            Return mTable
        End Get
        Set(ByVal Value As Model10.Table)
            mTable = Value
        End Set
    End Property

    Public Property Form() As Model10.Form
        Get
            Return mForm
        End Get
        Set(ByVal Value As Model10.Form)
            mForm = Value
        End Set
    End Property
    Public Property MemberName() As String
        Get
            Return mMemberName
        End Get
        Set(ByVal Value As String)
            mMemberName = Value
        End Set
    End Property
    Public Property Report() As Model10.Report
        Get
            Return mReport
        End Get
        Set(ByVal Value As Model10.Report)
            mReport = Value
        End Set
    End Property
End Class








'<Serializable()> Public Class Partner
'    Inherits Model10.Command

'    Private mDataField As String
'    Private mName As String

'    Protected Sub New(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.New(info, context)
'        On Error Resume Next

'        mDataField = info.GetString("DataField")
'        mName = info.GetString("Name")
'    End Sub
'    Protected Overrides Sub GetObjectData(ByVal info As SerializationInfo, ByVal context As StreamingContext)
'        MyBase.GetObjectData(info, context)

'        info.AddValue("DataField", mDataField)
'        info.AddValue("Name", mName)
'    End Sub

'    Public Overrides Function ToString() As String
'        Return mName
'    End Function
'    Public Sub New()
'        MyBase.New()

'        mDataField = ""
'        mName = ""
'    End Sub

'    Public Property DataField() As String
'        Get
'            Return mDataField
'        End Get
'        Set(ByVal Value As String)
'            mDataField = Value
'        End Set
'    End Property
'    Public Property Name() As String
'        Get
'            Return mName
'        End Get
'        Set(ByVal Value As String)
'            Value = Value.Trim
'            If Value = "" Then Throw New System.Data.DuplicateNameException("Name can not be empty.")
'            If Not MyBase.Table Is Nothing Then
'                Dim def As Model10.Default
'                For Each def In MyBase.Table.Defaults
'                    If Not def Is Me Then
'                        If def.Name.ToLower = Value.ToLower Then
'                            Throw New System.Data.DuplicateNameException("A Compute named '" & Value & "' already belongs to this Table.")
'                        End If
'                    End If
'                Next : def = Nothing
'            End If
'            mName = Value
'        End Set
'    End Property
'End Class
'<Serializable()> Public Class PartnerCollection
'    Inherits CollectionBase

'    Private mTable As Model10.Table

'    Public Function Contains(ByVal value As Model10.Partner) As Boolean
'        Return MyBase.List.Contains(value)
'    End Function
'    Public Sub New(ByVal table As Model10.Table)
'        MyBase.New()

'        mTable = table
'    End Sub

'    Default Public ReadOnly Property Item(ByVal index As Integer) As Model10.Partner
'        Get
'            Return CType(MyBase.List.Item(index), Model10.Partner)
'        End Get
'    End Property
'    Default Public ReadOnly Property Item(ByVal name As String) As Model10.Partner
'        Get
'            Dim str As String = name.Trim.ToLower

'            Dim ds As Model10.Partner
'            For Each ds In list
'                If ds.Name.Trim.ToLower = str Then
'                    Return ds
'                End If
'            Next : ds = Nothing
'        End Get
'    End Property
'    Public Sub Add(ByVal value As Model10.Partner)
'        If Not mTable Is Nothing Then
'            If value.Name = "" Then
'                Throw New System.Data.DuplicateNameException("Name can not be empty.")
'            Else
'            End If
'            value.Table = mTable
'        End If
'        MyBase.List.Add(value)
'    End Sub
'    Public Sub Insert(ByVal index As Integer, ByVal value As Model10.Partner)
'        If Not mTable Is Nothing Then
'            If value.Name = "" Then
'                Throw New System.Data.DuplicateNameException("Name can not be empty.")
'            Else
'            End If
'            value.Table = mTable
'        End If
'        MyBase.List.Insert(index, value)
'    End Sub
'    Public Sub Remove(ByVal value As Model10.Partner)
'        If Not mTable Is Nothing Then
'            value.Table = Nothing
'        End If
'        MyBase.List.Remove(value)
'    End Sub
'End Class
